#-*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

# RHINO SPECIFIC IMPORTS | FROM PYTHON 3.5 to 2.7
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

# Rhino Scripting
import rhinoscriptsyntax as rs
import System
import Rhino as r
import scriptcontext as sc

# Python built-ins
from builtins import range
import os
import warnings

# API support code imports
from .convert.convert_agnostic_to_sortal import convertA
from .convert.convert_rhino_to_agnostic import convertR
from .convert.convert_sortal_to_agnostic import convertS
from ..rhino_specific.visualAgnostic import visualA

# Sortal library imports
from sortal import form, sort

# Class initializations
convertA = convertA()
convertR = convertR()
convertS = convertS()

"""
=============
FUNCTION INFO
=============
'maximalize' returns a maximalized sortal shape/list of Rhino GUIDs.

INPUT
shape: Shape data in the form of a list of Rhino GUIDs (target disjunctive sort for maximalized must be provided) 
       or the shape name as a text string
target: Target disjunctive sort for the shape to be maximalized (text string), this is necessary if the input is a list of Rhino GUIDs
rhino: Boolean value (True/False); True - returns list of Rhino GUIDs corresponding to maximalized shape; False (default) - returns agnostic form
       Note that if Rhino GUIDs are given as input, then Rhino GUIDs will be returned
hide: Boolean value (True/False); True - hides Rhino geometry output from viewport; False (default) - leaves Rhino geometry output visible in viewport
delete: Boolean value (True/False); if the inputs are a list of Rhino GUIDs, then True deletes the inputs, 
        and False (default) leaves the inputs still in the Rhino viewport

OUTPUT
maxShape: If a shape name is inputted, then it returns the maximalized sortal shape unless the variable rhino is set to True
          If a list of Rhino GUIDs is inputted, then it returns a list of Rhino GUIDs corresponding to the maximalized form by default

WARNINGS / ERRORS
TypeError: If type of input is invalid (i.e. not a list of Rhino GUIDs or a shape name text string)
ValueError: If input is an empty list or if shape name doest not exist in the form register
KeyError: If 'rhino_shapes_c' is not present in the sort register, i.e. sort types in sortal library has not yet been set up
    
Data Flow: GUIDs/text string of shape name -> sortal shape (-> GUIDs, if rhino is set to true or if input was GUIDs)
"""

def maximalize (shape, target = None, rhino = False, hide = False, delete = False):
    if 'rhino_shapes_c' in sort.register:
        # Maximalizing list of Rhino GUIDs
        if type(shape) == type([]):
            try:
                rs.ObjectType(shape[0])
            except:
                if len(shape) > 0:
                    raise TypeError('Shape input is not composed of Rhino GUIDs.')
                else:
                    raise ValueError('Shape input is empty.')
                    
            shapeAgs = convertR.toAgnostic(shape, target)[0]
            shapeSortal = convertA.toSortal(shapeAgs)
            shapeSortal.maximalize()
            maxShape = convertS.toAgnostic(shapeSortal)
            rhino = True
        
        elif type(shape) == type('a'):
            if shape in form.formRegister:
                shapeSortal = form.formRegister[shape]
                shapeSortal.maximalize()
                maxShape = shapeSortal
            else:
                raise ValueError('Shape with name ' + shape +' is not available in the sort register.')
       
        else:
            raise TypeError('Invalid input for maximalization')
                
        if rhino:
            draw_shape = visualA(maxShape)
            maxShape = draw_shape.plotAgnostic()
            
            if delete and type(shape) == type([]):
                rs.DeleteObjects(shape)
                
            if hide:
                rs.HideObjects(maxShape)
            
        return maxShape
    else:
        raise KeyError('Compound sort type "rhino_shapes_c" not present in the sort register. Please set the highest tier sort in your SDL file to a compound sort named "rhino_shapes_c" to ensure compatibility with the API.', 0, '"rhino_shapes_c" not present"')
                
        
