#-*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

# RHINO SPECIFIC IMPORTS | FROM PYTHON 3.5 to 2.7
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

# Rhino Scripting
import rhinoscriptsyntax as rs
import System
import Rhino as r
import scriptcontext as sc

# Python built-ins
from builtins import range
import os
import warnings

# API support code imports
from .convert.convert_rhino_to_agnostic import convertR
from ..rhino_specific.visualAgnostic import visualA

# Sortal library imports
from sortal import form, sort

# Class initializations
convertR = convertR()

"""
=============
FUNCTION INFO
=============
'redraw' redraws a inputted list of Rhino GUIDs with their corresponding tags, predicates, directives, labels and descriptions.
The original Rhino GUIDs inputted into the function are deleted, along with any text dots that serve only as tag data and predicate/directive information holders.

SUPPORT FUNCTIONS
convertR: converts Rhino geometry to an agnostic dictionary
visualA: draws the agnostic dictionary data as Rhino geometry in the viewport

INPUT
rhinoObject: List of Rhino GUIDs to redraw
targetSort: Name of target disjunctive sort type to classify geometry under (necessary, default is non-parametric 3D disjunctive sort 'N3D')

OUTPUT
objectIds: List of Rhino GUIDs corresponding to redrawn geometry
Deletes inputted list of Rhino GUIDs, generates new list of GUIDs as outputs

WARNINGS / ERRORS
TypeError: If any elements inside rhinoObject are not Rhino GUIDs, or if the input for rhinoObject is not a list of Rhino GUIDs
KeyError: If the sortal library has not yet been set up, i.e. if the compound sort 'rhino_shapes_c' does not yet exist in the sort register

Data Flow: Rhino GUIDs -> new list of Rhino GUIDs (inclusive of information holder text dots)
"""

def redraw(rhinoObject, targetSort = 'N3D'):
    if 'rhino_shapes_c' in sort.register:
        if type(rhinoObject) == type([]):
            try:
                rs.ObjectType(rhinoObject[0])
            except:
                raise TypeError('Rhino object data is not a list of Rhino GUIDs.')

            if targetSort not in sort.register:
                for sortType in sort.register['rhino_shapes_c'].component:
                    if sortType.type == 'sort-disjunctive':
                        for subSortType in sortType.component:
                            if subSortType.type == 'sort-attribute' or (subSortType.type == 'sort-primitive' and 'description' not in subSortType.baseSort().characteristic.type):
                                targetSort = sortType.name
                            
            # Converts Rhino Objects to agnostic form
            shapeAg, shapeIDs, preds, dirs  = convertR.toAgnostic(rhinoObject, targetSort)
    
            # Deletion of existing Rhino Objects excludes elliptical arcs because of issues with handling them after redraw
            for sortType in shapeIDs:
                rs.DeleteObjects(shapeIDs[sortType])
                
            # Deletes any text dots that serve only as visual information holders
            for guid in rhinoObject:
                if rs.IsTextDot(guid):
                    if 'visual' in rs.GetUserText(guid):
                        rs.DeleteObject(guid)

            # Draws shape anew complete with labels, label descriptions, predicates, and directives
            newShape = visualA(shapeAg, (0,0,0), 'Default', preds, dirs)
            objectIDs = newShape.plotAgnostic()
    
            # Returns Rhino Object GUIDs
            return objectIDs
        else:
            raise TypeError ('Input is not a list of rhino objects')
    else:
        raise KeyError('The sortal library has not yet been set up; the compound sort "rhino_shapes_c" is not yet present in the sort register.')