#-*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

# RHINO SPECIFIC IMPORTS | FROM PYTHON 3.5 to 2.7
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

# Rhino Scripting
import rhinoscriptsyntax as rs
import System
import Rhino as r

# Python library imports
import warnings

# Sortal library imports
from sortal import sort, flow, rule, form

"""
=============
FUNCTION INFO
=============
'create_flow' creates a flow from the inputs flowName (name of the flow object), flowDesc (description of the flow object),
and flowSeq (text string sequence of rule names and ordering, looping and execution instructions)

SUPPORT FUNCTIONS
iterative_pop: Pops any characters corresponding to instructions for rule ordering, looping and execution within a flow sequence
name_check: Obtains  the name of a rule minus any ordering, looping and execution instructions

INPUT
flowName: Flow name (text string); this will be used to retrieve the flow object from the flow register
flowDesc: Flow description (text string, optional); this is used to describe the flow object
flowSeq: Flow sequence (text string); this text string contains the sequence of rules as well as their ordering, looping and execution instructions

OUTPUT
newFlow: The flow sortal data structure is returned if the flow is successfully created
    
WARNINGS/ERRORS
TypeError: If the input for flowName, flowDesc, or flowSeq is not a text string
ValueError: If the input for flowName or flowSeq is empty, or a rule name specified in flowSeq is not present in the rule register
KeyError: If a flow object with the same data as flowName already exists in the flow register
          or if 'rhino_shapes_c' is not present in the sort register, i.e. sort types in sortal library has not yet been set up

Data Flow: flow name, flow description (optional), flow sequence
           -> checking existence of any flow objects with the same name as flow name inside the flow register (overwrite/rewrite/cancel)
           -> checking of validity of all rules present inside flow sequence
           -> creation of flow object
           -> return flow object
"""

def create_flow(flowName, flowDesc = '', flowSeq = ''):
    if 'rhino_shapes_c' in sort.register:
        if type(flowName) != type('a'):
            raise TypeError('Input for flow name is not a text string.')
        elif len(flowName) == 0:
            raise ValueError('Empty flow sequence string')
        
        if flowName in flow.register:
            go_ahead = rs.MessageBox("A rule with the same name as " + flowName + " already exists in the rule register. Overwrite pre-existing rule?",  3, "Rule Overwrite")
        
            if go_ahead == 6:
                del flow.register[flowName]
                
            elif go_ahead == 7:
                newName = raw_input('Please input new rule name for the rule to be created')
                flowName = newName
                
            elif go_ahead == 2:
                warnings.warn('Exiting create_rule function')
                return None
        
        if type(flowSeq) != type('a'):
            raise TypeError('Input for flow sequence is not a text string.')
        elif len(flowSeq) == 0:
            raise ValueError('Empty flow sequence string')
        
        if type(flowDesc) != type('a'):
            raise TypeError('Input for flow description is not a text string.')
            
        flowSeqCheck = flowSeq.split(' ')
    
        for rule_item in flowSeqCheck:
            if name_check(rule_item) not in rule.register:
                raise ValueError(name_check(rule_item) + ' is not present in the rule register.')
        
        if flowName not in flow.register:
            newFlow = flow.createFlow(flowName, flowDesc, flowSeq)
            return newFlow
        else:
            raise KeyError(flowName + ' is already present in the rule register. Please use a different name or delete the previous flow sequence first.')

    else:
        raise KeyError('Compound sort type "rhino_shapes_c" not present in the sort register. Please set the highest tier sort in your SDL file to a compound sort named "rhino_shapes_c" to ensure compatibility with the API.', 0, '"rhino_shapes_c" not present"')
                

"""
=====================
SUPPORT FUNCTION INFO
=====================
'iterative_pop' removes any looping or ordering instructions within a section of the flow sequence

INPUT
sequence: Part of flow sequence to clean
character: Character to search for within flow sequence part

OUTPUT
Rule name as present in rule register

Data Flow: Part of flow sequence, character -> rule name
"""
    
def iterative_pop(sequence, character):
    while True:
        if character in sequence:
            sequence.pop(sequence.index(character))
        else:
            break
            
    return sequence

"""
=====================
SUPPORT FUNCTION INFO
=====================
'name_check' removes any looping, ordering or execution instructions within a section of the flow sequence

INPUT
rule_item: Text string with rule name + looping, ordering, execution instructions inside

OUTPUT
Rule name as present in rule register

Data Flow: Text string with rule name + flow instructions -> rule name as present in rule register
"""

def name_check(rule_item):  
    rule_check = list (rule_item) 
    characters = ['[', ']', '(', ')', '*', '+', '?']
    ruleName = ''
    
    # Clears loop sequences from string
    while True:
        if '{' in rule_check:
            check_char = rule_check.index('{')
            rule_check = rule_check[0:check_char]
        else:
            break
    
    # Clears other ordering instructions from string
    for char in characters:
        if char in rule_check:
            rule_check = iterative_pop(rule_check, char)

    for letter in rule_check:
        ruleName +=  letter

    return ruleName