
#-*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

# RHINO SPECIFIC IMPORTS | FROM PYTHON 3.5 to 2.7
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

# Rhino Scripting
import rhinoscriptsyntax as rs
import System
import Rhino as r
import scriptcontext as sc

# Python built-ins
from builtins import range
import os
import warnings

# Sortal Library Imports
from sortal import (rule, form, primitiveSort, vector2D, vector3D, description, sort, flow)

# Import for transformations
from mpmath import pi, sin, cos
from copy import deepcopy

import Rhino.Geometry as rg

"""
=============
FUNCTION INFO
=============
'create_rule' creates a new rule instance; it takes as input the rule name, rule description, LHS and RHS shape names, 
predicates (for LHS) and directives (RHS). Since create_rule() is linked to the sortal library, the act of creating a rule object checks 
if a rule with the same name as the new rule being created already exists. If yes, then the rule is not created and the function is exited. 

SUPPORT FUNCTIONS
add_predicates: Adds predicates to rule object
add_directives: Adds directives to rule object

INPUT
ruleName: Name of rule to be created (text string) 
ruleDesc: Description of rule to be created (text string)
lhs: Name of shape to become LHS of rule (text string)
rhs: Name of shape to become RHS of rule (text string)
predLHS: Agnostic dictionary of predicates (optional); usually, this is obtained from sgi.create_shape
dirLHS: Agnostic dictionary of directives (optional); usually, this is obtained from sgi.create_shape
prnt: Boolean value, corresponding to whether to print if rule was created successfuly, and as to which predicates/directives were added to the rule successfully 

OUTPUT
Sortal rule data structure, if successful
None, if unsuccessful

WARNINGS / ERRORS
MessageBox: If a rule with the same name already exists in the sortal register, then the user is prompted for whether they would like to overwrite 
            the pre-existing rule or to give the rule to be created a different name or to exit the create_rule function; 
            alternatively, if the geometry within the rule object's LHS is insufficient
TypeError: If the input for rule name or rule description or LHS shape name or RHS shape name is not a text string
KeyError: If the shapes corresponding to the inputs for the LHS and RHS shape names are not present in the form register
          or if 'rhino_shapes_c' is not present in the sort register, i.e. sort types in sortal library has not yet been set up

Data Flow: rule name, rule description (optional), LHS shape name, RHS shape name, predicates dictionary, agnostic dictionary
           -> checks if rule with the same name already exists (overwrite/rewrite)
           -> creates rule
           -> adds predicates/directives into rule object
           -> checks validity of rule geometry
           -> returns rule object as sortal rule data structure, if successful
"""

def create_rule(ruleName, ruleDesc, lhs, rhs, predLHS = None, dirRHS = None, prnt = False):
    if 'rhino_shapes_c' in sort.register:
        if ruleName in rule.register:
            go_ahead = rs.MessageBox("A rule with the same name as " + ruleName + " already exists in the rule register. Overwrite pre-existing rule?",  3, "Rule Overwrite")
        
            if go_ahead == 6:
                del rule.register[ruleName]
                
            elif go_ahead == 7:
                newName = raw_input('Please input new rule name for the rule to be created')
                ruleName = newName
                
            elif go_ahead == 2:
                warnings.warn('Exiting create_rule function')
                return None
        
        if type(ruleName) != type('a'):
            raise TypeError('Rule name input is not a string')
            
        if type(lhs) == type('a') and type(rhs) == type('a'):
            if lhs in form.formRegister:
                lhsMetaform = form.formRegister[lhs]
            else:
                raise KeyError('LHS shape', lhs, "does not exist in the sortal library's form register.")
                
            if rhs in form.formRegister:
                rhsMetaform = form.formRegister[rhs]
            else:
                raise KeyError('RHS shape', rhs, "does not exist in the sortal library's form register.")
        else:
            raise TypeError('Invalid input for left-hand side/right-hand side')

        # Creates rule instance using sortal library
        ruleNew = rule(ruleName, ruleDesc, lhsMetaform, rhsMetaform, keyCheck = True)  
        
        if predLHS != None and predLHS != False:
            if type(predLHS) == type([]):
                for pred in predLHS:
                    add_predicates(ruleNew, pred, prnt)
            else:
                add_predicates(ruleNew, predLHS, prnt)
        
        if dirRHS != None and dirRHS != False:
            if type(dirRHS) == type([]):
                for dir in dirRHS:
                    add_directives(ruleNew, dirRHS, prnt)
            else:
                add_directives(ruleNew, dirRHS, prnt)
        
        if ruleNew.isValid():
            if prnt:
                print (ruleName + ' has been created.')
            return ruleNew
        else:
            message = ('Rule was not created since there was insufficient information for LHS Shape.\nThe minimum requirements for defining the LHS are:\n'+
                                    '- 3 points (not coincident)\n'+
                                    '- 1 point + 1 line (not co-linear)\n'+
                                    '- 3 lines (no two parallel: all lines intersect one another)\n'+
                                    '- 3 lines (2 lines parallel with each other, the other line intersecting both)\n'+
                                    '- 2 lines (skew): However, this may have unexpected results; try not to have rules with only skew lines if possible')
                                    
            warnings.warn(message)
            del rule.register[ruleName]
            return None
    else:
        raise KeyError('Compound sort type "rhino_shapes_c" not present in the sort register. Please set the highest tier sort in your SDL file to a compound sort named "rhino_shapes_c" to ensure compatibility with the API.', 0, '"rhino_shapes_c" not present"')
                    
"""
=====================
SUPPORT FUNCTION INFO
=====================
'add_predicates' adds the predicate (max line, shortest line, void, bound line, longest line, no label) to the LHS of the rule object.

INPUT
ruleObject: Sortal rule object
predicates: Predicate coordinates and types (agnostic)
prnt: Boolean value; True: prints predicates added, False (default): does not print anything)

OUTPUT
Successfully added predicates to rule object, if successful

Data Flow: sortal rule object, predicates dictionary, classification of agnostic object -> rule object with predicates inside
"""

def add_predicates(ruleObject, predicates, prnt = False):
    ruleName = ruleObject.name
    lhs = ruleObject.lhs
    classification = predicates['classification']
    predKeys = {'max': [], 'shortest': [], 'longest': [], 'bound': [], 'nolabel':[], 'embeds':[]}
    addList = []
    
    for predType in predicates:
        geometry = predicates[predType]
        
        if len(geometry) > 0:
            if predType in predKeys and predType != 'bound' and predType != 'embeds':
                if predType == 'max':
                    predName = 'maxline'
                    keyName = 'max'
                    addName = 'Max Line'
                    
                elif predType == 'shortest':
                    predName = 'shortest_line'
                    keyName = 'shortest'
                    addName = 'Shortest Line'
                    
                elif predType == 'longest':
                    predName = 'longest_line'
                    keyName = 'longest'
                    addName = 'Longest Line' 
                    
                elif predType == 'nolabel':
                    predName = 'nolabel'
                    keyName = 'nolabel'
                    addName = 'No Label'
                    
                geometry =  predicates[keyName]
                
                for geomName in geometry:
                    if keyName in geomName:
                        for geomTag in geometry[geomName]:
                            if geomTag[0] == '@':
                                if ruleObject.retrieveKey(geomTag[1:]) not in predKeys[keyName]:
                                    predKeys[keyName].append(ruleObject.retrieveKey(geomTag[1:]))
                                elif ruleObject.retrieveKey(geomTag) not in predKeys[keyName]:
                                    predKeys[keyName].append(ruleObject.retrieveKey(geomTag))
                    else:
                        if geomName[0] == '@':
                            if ruleObject.retrieveKey(geomName[1:]) not in predKeys[keyName]:
                                predKeys[keyName].append(ruleObject.retrieveKey(geomName[1:]))
                        elif ruleObject.retrieveKey(geomName) not in predKeys[keyName]:
                            predKeys[keyName].append(ruleObject.retrieveKey(geomName))
            
            elif predType == 'bound':
                predName = 'bound'
                keyName = 'bound'
                addName = 'Bound Line'
                
                for tagName in geometry:
                    first = tagName[0]
                    last = tagName[-1]
                    for geom in geometry[tagName]:
                        object = ruleObject.retrieveKey(geom)
                        if first == '>':
                            object.addProperty(0)
                        if last == '<':
                            object.addProperty(1)
                        predKeys[keyName].append(object)
                
                # Cleans predKeys['bound'] to remove duplicates
                tempList = deepcopy(predKeys[keyName])
                cleanList = []
                
                for item in tempList:
                    if item not in cleanList:
                        cleanList.append(item)
                        
                predKeys[keyName] = cleanList
                
            elif predType == 'void':
                addName = 'Void'
                predName = predType
                voidSets = []
                
                for voidType in geometry:
                    vectors = []
                    void = geometry[voidType]
                    
                    if type(void[0][0]) == type('a'):
                        voidCoords = void[1:]
                        voidSorts = list(void[0])
                    else:
                        voidCoords = void
                        voidSorts = None
                        
                    for vector in voidCoords:
                        if len(vector) == 2:
                            if '3' in classification:
                                vectors.append(vector3D(vector[0], vector[1], 0))
                                
                            elif '2' in classification:
                                vectors.append(vector2D(vector[0], vector[1]))
                            
                        elif len(vector) == 3:
                            if '3' in classification:
                                vectors.append(vector3D(vector[0], vector[1], vector[2]))
                                
                            elif '2' in classification:
                                vectors.append(vector2D(vector[0], vector[1]))
                    
                    if voidSorts != None:
                        voidSets.append([voidSorts, vectors])
                    else:
                        voidSets.append(vectors)
            
            elif predType == 'embeds':
                predName = predType
                keyName = predType
                addName = 'Embeds'
                
                for groupCount in geometry:
                    tags = geometry[groupCount]
                    pair = []
                    
                    for item in tags:
                        pair.append(ruleObject.retrieveKey(item))
                        
                    predKeys[keyName].append(pair)
            
            # Embed not getting added successfully
            if predType == 'void':
                ruleObject.addPredicate(predType, voidSets)
            elif predType != 'classification':
                print ('\n', predName, predKeys[keyName])
                ruleObject.addPredicate(predName, predKeys[keyName])

            if predType != 'classification':
                if addName not in addList:
                    addList.append(addName)

    if prnt:
        print('predicates:')
        print(ruleObject.getPredicates())
        print ('The following predicates were added', addList)
        
    return

"""
=====================
SUPPORT FUNCTION INFO
=====================
'add_directives' adds the directives (point on line, distance, normal) to the RHS of the rule object.

INPUT
ruleObject: Sortal rule object
directives: Directives coordinates and types (agnostic)
prnt: Boolean value; True: prints predicates added, False (default): does not print anything)

OUTPUT
Successfully added directives to rule object, if successful

Data Flow: sortal rule object, directives dictionary, classification of agnostic object -> rule object with directives inside
"""   

def add_directives(ruleObject, directives, prnt = False):
    ruleName = ruleObject.name
    lhs = ruleObject.lhs
    classification = directives['classification']
    dirKeys = {}
    tags = ['pointOL', 'distance', 'normal']
    addList = []
    
    for dirType in directives:
        geometry = directives[dirType]
        
        if dirType in tags and len(geometry) > 0:
            if dirType == 'pointOL':
                dirName = 'point_on_line'
                keyName = 'point_on_line'
                addName = 'Point on Line'
                
            elif dirType == 'distance':
                dirName = 'distance'
                keyName = 'distance'
                addName = 'Distance'
            
            elif dirType == 'normal':
                dirName = 'normal'
                keyName = 'normal'
                addName = 'Normal'
                
            for geom in geometry:
                tempGeom = []
                expressions = []

                if type(geom) == type([]):
                    for a in geom:
                        if type(a) == type(''):
                            if '@' in a:
                                indivKey = ruleObject.retrieveKey(a[1:])
                                tempGeom.append(indivKey)
                            else:
                                expressions.append(description(sort.getSort('__dummyDesc__'), '`' + a + '`'))
                        elif type(a) == type(0) or type(a) == type(0.0):
                            expressions.append(a)
                else:
                    tagData = geometry[geom]
                    for a in tagData:
                        if type(a) == type(''):
                            if a[0] == '@':
                                indivKey = ruleObject.retrieveKey(a[1:])
                                tempGeom.append(indivKey)
                            elif a.isdigit():
                                tempGeom.append(float(a))
                            elif '`' in a or type(a) == type('a'):
                                expressions.append(description(sort.getSort('__dummyDesc__'), '`' + a + '`'))
                            else:
                                tempGeom.append(float(a))
                                
                        elif type(a) == type(0) or type(a) == type(0.0):
                            expressions.append(a)
                    
                    # Removes identifying planeSegment or lineSegment tag
                    if dirType == 'normal':
                        tempGeom.pop(0)  
                        
                for i in expressions:
                    tempGeom.append(i)

                if keyName not in dirKeys:
                    dirKeys[keyName] = []
                    
                if None not in tempGeom:    
                    dirKeys[keyName].append(tempGeom)

            ruleObject.addDirective(dirName, dirKeys[keyName])
            addList.append(addName)
    
    if prnt:
        print('directives:')
        print(ruleObject.getDirectives())
        print ('The following directives were added', addList)