#-*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

# RHINO SPECIFIC IMPORTS | FROM PYTHON 3.5 to 2.7
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

# Rhino Scripting
import rhinoscriptsyntax as rs
import System
import Rhino as r
import scriptcontext as sc

# Python built-ins
from builtins import range
import os
import warnings

# Sortal Library Imports
from sortal import (coordinate, transformSet, rule, primitiveSort, point3D, point2D, lineSegment3D, planeSegment2D, 
                    lineSegment2D, line3D, label, rWeight, disjunctiveSort, attributeSort, affineTransform, rule, sort, 
                    vector3D, vector2D, planeSegment3D, sort, affineTransform3D, description, graphMap3D, sdlParser, 
                    compoundSort, disjunctiveSort, graphMap, form, key)
            
# Import for transformations
from mpmath import pi, sin, cos
from copy import deepcopy

# API imports
from ..sortal_lib_api.convert.convert_agnostic_to_sortal import convertA
from ..sortal_lib_api.convert.convert_rhino_to_agnostic import convertR

import Rhino.Geometry as rg

convertA = convertA()
convertR = convertR()


"""
=============
FUNCTION INFO
=============
'create_shape' creates a sortal shape from a collection of Rhino GUIDs. It names the shape according to the input for the shapeName variable in the sortal library.
It returns the predicate/directive dictionaries. 

INPUT
shapeName: Name of shape (text string)
shapeData: List of Rhino GUIDs that will compose the sortal shape
descriptions: Text string of descriptions to include in the sortal shape, e.g.
              'label1@("A2", 1, 1234);("A3", 4, 1234)|label2@("A1", 1, 1234)|label2@("A9", 8, 1234)'
              where there are two description types, 'label1' (followed by an ampersand @; 2 individuals separated by a semicolon ;),         
              and 'label2' (2 individuals); declaration of different description types and their individuals is separated by a vertical dash |
classification: Name of target disjunctive sort type (text string); this sort type must already be present in the sort register 
                under the compound sort 'rhino_shapes_c'
refPoint: Reference point to serve as 'origin' point for shape; the default origin point is (0,0,0)
prnt: Boolean value (True/False); True - prints the resulting sortal shape according to the output of the 'printE' command in the sortal library,
      False (default) - does not print anything

OUTPUT
predicates: Predicates dictionary
directives: Directives dictionary

WARNINGS / ERRORS
TypeError: If shapeName input is not a text string, or if shapeData input is not a list of Rhino GUIDs, or if shapeData has any elements that are not Rhino GUIDs inside
Warning: If descriptions input is not a text string
KeyError: If the result Rhino GUIDs to Agnostic Dictionary conversion is empty 
          OR if 'rhino_shapes_c' is not present in the sort register, i.e. sort types in sortal library has not yet been set up

Data Flow: Rhino GUIDs + description text string 
           -> sortal shape 
           -> register shape in form register 
           -> print output (optional)
           -> return predicates, directives dictionaries
"""

def create_shape(shapeName = None, shapeData = None, descriptions = '',  classification = None, refPoint = rg.Point3d(0,0,0), prnt = False):
    if 'rhino_shapes_c' in sort.register:
        if type(shapeName) != type('a'):
            raise TypeError('Shape name is not a text string.') 
            
        if type(shapeData) == type([]):
            try:
                rs.ObjectType(shapeData[0])
            except:
                raise TypeError('Shape data is not a list of Rhino GUIDs.')
                
            if type(descriptions) != type('a'):
                warnings.warn('Descriptions input is not a text string. Descriptions will not be included in the sortal shape.')
                descriptions = ''
               
            if refPoint is None:
                refPoint = rg.Point3d(0,0,0)
    
            shapeAg, shapeIDs, predicates, directives = convertR.toAgnostic(shapeData, classification, descriptions, refPoint)

            #if len(shapeAg) == 0:
            #    raise KeyError('No shape data received')

            # Converts agnostic dictionary to sortal form
            shapeMetaform = convertA.toSortal(shapeAg)

            # Registers shape object name
            shapeMetaform.register(shapeName)
            
            if prnt:
                # Outputs complete sortal metaform (geometry + external descriptions) using sortal form printE() function
                print('The following sortal shape (named ' + shapeMetaform.name + ') was created:\n')
                shapeMetaform.printE()
    
            return predicates, directives 
        
        else:
            raise TypeError('Shape data is not a list of Rhino GUIDs')

    else:
        raise KeyError('Compound sort type "rhino_shapes_c" not present in the sort register. Please set the highest tier sort in your SDL file to a compound sort named "rhino_shapes_c" to ensure compatibility with the API.', 0, '"rhino_shapes_c" not present"')
                
            