#-*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

# RHINO SPECIFIC IMPORTS | FROM PYTHON 3.5 to 2.7
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

# Rhino Scripting
import rhinoscriptsyntax as rs
import System
import Rhino as r

# Python built-ins
from builtins import range
import os
import warnings

# Sortal library imports
from sortal import sort, rule, sdlParser, form

# API support code imports
from .convert.convert_sortal_to_agnostic import convertS
from ..rhino_specific.visualAgnostic import visualA
from .move import translation_vector

# Conversion class initialization
convertS = convertS()

"""
=============
FUNCTION INFO
=============
'draw_shape' draws the shape in the Rhino workspace. 
This function assumes that the shape is already present in the form register.
When a reference point is given (tuple/list/vector/point), the shape GUIDs 
are moved to the reference point. The default reference point is the origin. 

SUPPORT FUNCTIONS
convertS: Converts a sortal shape to an agnostic dictionary to serve as input for the drawing function, visualA
visualA: Draws an agnostic dictionary in the Rhino workspace
translation_vector: Moves shape to frame of reference with refPt input as origin

INPUT
shapeName: Name of shape object to draw from inside form register (text string)
refPt: Reference point (optional - tuple / list of integers or floats / vectors / points (3Ds)); relocates
       drawing of shape with reference point serving as the origin; default origin is (0,0,0)

OUTPUT
shapeNewRhino: List of Rhino Geometry [shape GUIDs]; this is returned as an empty list the shape does not exist in the shape register

WARNINGS / ERRORS
Warning: Shape with the inputted shape name does not exist in the form register
TypeError: Shape name input is not a text string
KeyError: If 'rhino_shapes_c' is not present in the sort register, i.e. sort types in sortal library has not yet been set up

Data Flow: shape name in form register  
           -> sortal object retrieved from SDL 
           -> sortal object 
           -> agnostic dictionary of shape 
           -> list of Rhino GUIDs (visible in Rhino viewport)
"""
            
def draw_shape(shapeName, refPt = r.Geometry.Point3d(0,0,0)):
    if 'rhino_shapes_c' in sort.register:
        if type(shapeName) == type('a'):
            if shapeName in form.formRegister:
                shape = form.formRegister[shapeName]                    # Retrieves shape
                shapeAgnostic = convertS.toAgnostic(shape)              # Converts sortal form to agnostic tuple
                visObjects = visualA(shapeAgnostic)                     # Initializes agnostic tuple input in visualA
    
                shapeNewRhino = visObjects.plotAgnostic()               # Draws geometry
                bbox = rs.BoundingBox(shapeNewRhino)
                translationVec = translation_vector(refPt, bbox, 0)
                        
                if translationVec != None:
                    # Moves geometry to location with respect to reference point
                    shapeNewRhino = rs.MoveObjects(shapeNewRhino, translationVec)
                
                return shapeNewRhino
            else:
                warnings.warn('Shape does not exist in the form register.') 
                return []
        else:
            raise TypeError('Input for shape name is not a text string.')
    else:
        raise KeyError('Compound sort type "rhino_shapes_c" not present in the sort register. Please set the highest tier sort in your SDL file to a compound sort named "rhino_shapes_c" to ensure compatibility with the API.', 0, '"rhino_shapes_c" not present"')
               