#-*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

# RHINO SPECIFIC IMPORTS | FROM PYTHON 3.5 to 2.7
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

# Rhino Scripting
import rhinoscriptsyntax as rs
import System
import Rhino as r
import scriptcontext as sc

# Python built-ins
from builtins import range
import os
import warnings
from copy import deepcopy
import math

# Sortal library import
from sortal import sort

"""
=============
FUNCTION INFO
=============
'move' spaces out the shapes in a list based on either the translationVec variable (if a valid input is giving)
or based on the bounding box dimensions multiplied by the value of sc.sticky in vertical order. The altered Rhino GUIDs are returned.

SUPPORT FUNCTION
translation_vector: Creates the translation vector for a shape based on its bounding box, and the reference point input (which serves as the origin)

INPUT
shapes: List of Rhino GUID lists or list of Rhino GUIDs, translation vector (optional)
alignment: Vector that will be reduced to unit vector to determine axis of moved shapes (e.g. vertical ascending or horizontal going to the right, etc.)
translationVec: Single vector or list of vectors, that serve.s as the spacing reference vector/s between shapes. The space between the first
                and second shapes will follow the first vector element of the list, the space between the second and third shapes will follow the
                second vector element of the list, and so on.
    
OUTPUT
movedShapes: List of Rhino GUID lists after moving the geometries
None, if unsuccessful

WARNINGS / ERRORs
TypeError: if data type of alignment and/or translation vector is not a tuple or list of three numbers or a Vector3d object
KeyError: If 'rhino_shapes_c' is not present in the sort register, i.e. sort types in sortal library has not yet been set up
Warning: If the input for the reference point is invalid (not a tuple or a list of three numbers or a Rhino point geometry)

Data Flow: List of Rhino GUID lists -> move the GUIDs -> altered list of Rhino GUID
"""
   
def move(shapes,  alignment = False, translationVec = False):
    if 'rhino_shapes_c' in sort.register:
        movedShapes = [] 
        start_distance_x = 0.0
        start_distance_y = 0.0
        
        if (type(alignment) != type([]) and type(alignment) != type(()) and type(alignment) != r.Geometry.Vector3d) and alignment != False:
            raise TypeError('Alignment vector is not a tuple/list of three numbers or a vector')
            
        else:
            if type(alignment) == type([]) or type(alignment) == type(()):
                alignment = r.Geometry.Vector3d(alignment[0], alignment[1], alignment[2])
                alignment.Unitize()
    
        if (type(translationVec) != type([]) and type(translationVec) != type(()) and type(translationVec) != r.Geometry.Vector3d) and translationVec != False:
            raise TypeError('Translation spacing vector is not a tuple/list of three numbers or a vector')
            
        else:
            if type(translationVec) == type([]) or type(translationVec) == type(()):
                if (type(translationVec[0]) == type(0) or type(translationVec[0]) == type(0.0)) and len(translationVec) == 3:
                    translationVec = r.Geometry.Vector3d(translationVec[0], translationVec[1], translationVec[2])
                elif type(translationVec[0]) == type([]):
                    vectors = []
                    for vector in translationVec:
                        vectors.append(r.Geometry.Vector3d(vector[0], vector[1], vector[2]))
                    translationVec = vectors
                else:
                    raise TypeError('Translation spacing vector is not a tuple/list of three numbers or a vector')
    
        if type(shapes) == type([]):
            if type(shapes[0]) == type([]):
               
                for row_count, shape in enumerate(shapes):
                    bbox = rs.BoundingBox(shape)
                    translation = None 
        
                    if start_distance_y == 0.0:
                        if (bbox[1].Y - bbox[0].Y) > 0:
                            start_distance_y = (bbox[1].Y - bbox[0].Y)*sc.sticky['fixed_factor'] * row_count
                        else:
                            start_distance_y = (bbox[2].Y - bbox[0].Y)*sc.sticky['fixed_factor'] * row_count
                    else: 
                        if (bbox[1].Y - bbox[0].Y) > 0: 
                            start_distance_y += (bbox[1].Y- bbox[0].Y)*sc.sticky['fixed_factor']
                        else:
                            start_distance_y += (bbox[2].Y- bbox[0].Y)*sc.sticky['fixed_factor']
                    
                    if type(translationVec) == r.Geometry.Vector3d or translationVec == False:
                        translation = translation_vector(translationVec, bbox, start_distance_y)
                    elif type(translationVector) == type([]):
                        translation = translation_vector(translationVec[row_count], bbox, start_distance_y)
    
                    # Aligns translation vector to alignment axis
                    if alignment != False:
                        distance = math.sqrt(translation.X**2 + translation.Y**2 + translation.Z**2)
                        axis = r.Geometry.Vector3d.Multiply(alignment, row_count+1)
                        translation = r.Geometry.Vector3d.Multiply(axis, distance)
    
                    if translation != None:
                        # Moves geometry to location with respect to reference point
                        rs.ShowObjects(shape)
                        movedShapes.append(rs.MoveObjects(shape, translation)) 
                    else:
                        warnings.warn('Invalid input for translation vector')
                        return None
            else:
                bbox = rs.BoundingBox(shapes)
                translation = None 
                
                translation = translation_vector(translationVec, bbox, 0.0)
    
                if translation != None:
                    # Moves geometry to location with respect to reference point
                    movedShapes.append(rs.MoveObjects(shapes, translation))
                    rs.ShowObjects(shapes)
                else:
                    warnings.warn('Invalid input for translation vector')
                    return None
        
        return movedShapes
    else:
        raise KeyError('Compound sort type "rhino_shapes_c" not present in the sort register. Please set the highest tier sort in your SDL file to a compound sort named "rhino_shapes_c" to ensure compatibility with the API.', 0, '"rhino_shapes_c" not present"')
                
"""
=============
FUNCTION INFO
=============
'translation_vector' computes the translation vector based on the dimensions of the bbox, the given translation vector by the user, and the value
of variableDist (i.e the starting distance, which could be based on the fixed factor inside sc.sticky['fixedFactor'] or the dimensions of the bounding box
and the row count

INPUT
Translation vector (Rhino Geometry Vector3d), bounding box of relevant geometry, space between boxes

OUTPUT
Rhino Geometry Vector3d

Data Flow: Vector, bbox, spac of units between boxes
"""

def translation_vector(translationVec, bbox, variableDist):

    if translationVec == False:
        translation = r.Geometry.Vector3d((bbox[1].X - bbox[0].X) * sc.sticky['fixed_factor'], variableDist, bbox[1].Z - bbox[0].Z)
    elif type(translationVec) == r.Geometry.Vector3d:
        translation = r.Geometry.Vector3d(translationVec.X -  bbox[0].X, variableDist - (bbox[0].Y - translationVec.Y), translationVec.Z -  bbox[0].Z)
    else:
        try:
            if rs.ObjectType(translationVec) == 1:
                coords = rs.PointCoordinates(translationVec)
                translation = r.Geometry.Vector3d(coords.X - bbox[0].X, variableDist - (bbox[0].Y - coords.Y), coords.Z -  bbox[0].Z)
        except:
            warnings.warn('Reference point input is invalid')  
            return None

    return translation
