#-*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

# RHINO SPECIFIC IMPORTS | FROM PYTHON 3.5 to 2.7
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

# Rhino Scripting
import rhinoscriptsyntax as rs
import System
import Rhino as r
import scriptcontext as sc

# Python built-ins
from builtins import range
import os
import warnings

# Sortal Library Imports
from sortal import (rule, form, sort, sdlParser, disjunctiveSort, compoundSort, primitiveSort, description, flow)
            
# Python imports
from copy import deepcopy

attribute_bases = ['label', 'enumerative', 'weight', 'color']
base_geometry = ['lineSeg', 'point', 'planeSeg', 'ellipse', 'circle', 'Arc', 'bezier', 'description']

"""
=============
FUNCTION INFO
=============
'read_sdl' opens up an SDL file and merges any compound sorts inside with the prevailing iteration of the compound sort 'rhino_shapes_c' 
in the sort register. It also updates any rules and shapes with outdated sort types not set to the prevailing compound sort 'rhino_shapes_c'
and rewrites the SDL file to ensure compatibility with the API.

SUPPORT FUNCTIONS
clean_sorts: Removes dummy disjunctive sorts (sorts with two description sort types inside) from the compound sort 'rhino_shapes_c'
update_sort: Updates the sort type of a shape/rule side to the compound sort 'rhino_shapes_c'

INPUT
SDL file name

OUTPUT
True, if successful
False, if unsuccessful

WARNINGS \ ERRORS
KeyError: if the file name does not exist

Data Flow: text string fileName 
           -> checking of presence of 'rhino_shapes_c' in sort register and in SDL file 
           -> update of contents of 'rhino_shapes_c' if necessary 
           -> update of contents (rules and shapes) inside SDL file if necessary 
           -> return Boolean True or False depending on success of operation
"""

def read_sdl(fileName):
    temp_sort_dict = {}
    temp_shape_dict = {}
    temp_rule_dict = {}
    temp_flow_dict = {}
    sdl_shapes = {}
    sdl_rules = {}
    sdl_flows = []
    
    if '__dummyDescription__' not in sort.register:
        if '__dummyDesc1__' not in sort.register:
            __dummy1__ = primitiveSort('__dummyDesc1__', description)
        else:
            __dummy1__ = sort.register['__dummyDesc1__']
            
        if '__dummyDesc2__' not in sort.register:
            __dummy2__ = primitiveSort('__dummyDesc2__', description)
        else:
            __dummy2__ = sort.register['__dummyDesc2__']
            
        __dummyDesc__ = disjunctiveSort('__dummyDescDisjunctive__', [__dummy1__, __dummy2__])
    else:
        __dummyDesc__ = sort.register['__dummyDescDisjunctive__']
        sc.sticky['__dummySort__'] = '__dummyDescDisjunctive__'

    for sortType in sort.register:
        temp_sort_dict[sortType] = deepcopy(sort.register[sortType])
        
    for shape in form.formRegister:
        temp_shape_dict[shape] = form.formRegister[shape].duplicate()
        
    for flow_item in flow.register:
        temp_flow_dict[flow_item] = deepcopy(flow_item)
        
    sort.cleanup()
    form.formRegister = {}
    flow.register = {}
    sdlParser.clear()
    
    # Checks if file exists
    if os.path.exists(fileName):
        
        if '.sdl' in fileName:
            sdlParser.parse(fileName)
            # Sets up shapes from SDL file in the sortal form register
            for shapeName in sdlParser.forms:
                form.formRegister[shapeName] = sdlParser.forms[shapeName].duplicate()

            for shape_name in sdlParser.forms:
                sdl_shapes[shape_name] = sdlParser.forms[shape_name].duplicate()

            for rule_name in sdlParser.rules:
                sdl_rules[rule_name] = deepcopy(sdlParser.rules[rule_name])
            
            for flow_name in sdlParser.flows:
                sdl_flows.append(deepcopy(sdlParser.flows[flow_name]))
                
            # Case 1: rhino_shapes_c present in opened SDL file
            # 1a: rhino_shapes_c present in previous sort register (temp_sort_dict)
            # 1b: rhino_shapes_c not present in previous sort register (temp_sort_dict)
            # Merge rhino_shapes_c (SDL) and rhino_shapes_c (temp_sort_dict)
            # Bring back all sorts inside temp_sort_dict
            # Update contents of file

            if 'rhino_shapes_c' in sort.register:                       # This rhino_shapes_c will inevitably come from the SDL file due to previous sort.register.clear()
                    
                comp_list_sdl = sort.register['rhino_shapes_c'].component
                
                # Case 1a
                # rhino_shapes_c present in previous sort register (temp_sort_dict)
                # Merge two rhino_shape_c sorts
                if 'rhino_shapes_c' in temp_sort_dict:
                    comp_list_temp = temp_sort_dict['rhino_shapes_c'].component
                    
                    for sort_type in comp_list_sdl:
                        add = True
                        for component_sort in comp_list_temp:
                            if sort_type.name == component_sort.name:
                                add = False
                                break
                        
                        if add: 
                            comp_list_temp.append(sort_type)
                  
                    del temp_sort_dict['rhino_shapes_c']

                # Case 1b
                # rhino_shapes_c not present in previous sort register (temp_sort_dict)
                # Merge contents of rhino_shapes_c (SDL) with highest tier geometric disjunctive and description disjunctive sorts in temp_sort_dict
                else:
                    comp_list_temp = []
                    for sort_type in temp_sort_dict:
                        if temp_sort_dict[sort_type].type == 'sort-disjunctive':
                            for base_geom in base_geometry:
                                if base_geom in temp_sort_dict[sort_type].component[0].baseSort().characteristic.type:
                                    comp_list_temp.append(temp_sort_dict[sort_type])
                                    break
                        elif sort.register[sort_type].type == 'sort-compound':
                            for component in sort.register[sort_type].type:
                                comp_list_temp.append(sort.register[sort_type])            
                    
                    if comp_list_temp > 0:
                        for sort_type in comp_list_sdl:
                            if sort_type not in comp_list_temp:
                                comp_list_temp.append(sort_type)
                
                # Removes dummy description sorts
                for i in range(2):
                    comp_list_temp = clean_sorts(comp_list_temp)
                    
                # Deletes rhino_shapes_c from sort register
                del sort.register['rhino_shapes_c']
                
                # Adds disjunctive dummy sort to prevent ejection from rhino_shapes_c sort later on
                if len(comp_list_temp) <2:
                    comp_list_temp.append(__dummyDesc__)
                    
                # Creates new rhino_shapes_c with combined contents of both rhino_shapes_c and disjunctive sorts
                compound_sort = compoundSort('rhino_shapes_c', comp_list_temp)
                  
                for sort_type in temp_sort_dict:
                    if sort_type not in sort.register:
                        sort.register[sort_type] = temp_sort_dict[sort_type]                   
                
            # Case 2: rhino_shapes_c not present in opened SDL file
            # 2a: Search for compound sorts with other names from rhino_shapes_c
            # 2b: Search for highest tier disjunctive sort composed of either min. one geometry and one description or min. two descriptions
            # 2c: Search for compound sorts not named 'rhino_shapes_c' and combine their contents to form one compound sort named 'rhino_shapes_c'
            # Merge with rhino_shapes_c in temp_sort_dict
            # Update contents of file
            elif 'rhino_shapes_c' in temp_sort_dict and 'rhino_shapes_c' not in sort.register:
                comp_list_temp = temp_sort_dict['rhino_shapes_c'].component
                comp_list_sdl = []
                
                # Collect all potentially other-named compound sorts inside sort register (from SDL) AND all disjunctive sorts 
                for sort_type in sort.register:
                    if sort.register[sort_type].type == 'sort-disjunctive':
                        for base_geom in base_geometry:
                            if base_geom in sort.register[sort_type].component[0].baseSort().characteristic.type:
                                comp_list_sdl.append(sort.register[sort_type])
                                break
                    elif sort.register[sort_type].type == 'sort-compound':
                        for component in sort.register[sort_type].component:
                            comp_list_sdl.append(sort.register[sort_type])
                
                for sort_type_sdl in comp_list_sdl:
                    add = True
                    for sort_type_temp in comp_list_temp:
                        if sort_type_sdl.name == sort_type_temp.name or sort_type_sdl.type == 'sort-compound':
                            add = False
                        
                    if add:
                        comp_list_temp.append(sort_type_sdl)

                # Deletes rhino_shapes_c from sort register
                if 'rhino_shapes_c' in sort.register:
                    del sort.register['rhino_shapes_c']
                
                # Removes dummy description sorts
                for i in range(2):
                    comp_list_temp = clean_sorts(comp_list_temp)
                
                # Adds disjunctive dummy sort to prevent ejection from rhino_shapes_c sort later on
                if len(comp_list_temp) <2:
                    comp_list_temp.append(__dummyDesc__)
                
                # Creates new rhino_shapes_c with combined contents of both rhino_shapes_c and disjunctive sorts
                compound_sort = compoundSort('rhino_shapes_c', comp_list_temp)
                    
                for sort_type in temp_sort_dict:
                    if sort_type not in sort.register:
                        sort.register[sort_type] = temp_sort_dict[sort_type]    

            # Neither are available -> but other compound sorts with names other than 'rhino_shapes_c' may exist
            # 3a: Check if compound sort(s) named other than 'rhino_shapes_c' exists in SDL file
            # Obtain, merge together to form rhino_shapes_c and delete existence of other-named compound sorts from sort register
            # If this is also not available, raise error
            else:
                cmpd_sorts = []
                for sort_type in sort.register:
                    if sort.register[sort_type].type == 'sort-compound':
                        cmpd_sorts.append(sort.register[sort_type])
                
                if len(cmpd_sorts) > 0:
                    comp_list_sdl = []
                    
                    for compound_sort in cmpd_sorts:
                        for component in compound_sort.component:
                            if component not in comp_list_sdl:
                                comp_list_sdl.append(component)
                    
                    # Adds disjunctive dummy sort to prevent ejection from rhino_shapes_c sort later on
                    if len(comp_list_sdl) <2:
                        comp_list_sdl.append(__dummyDesc__)
                    
                    compound_sort = compoundSort('rhino_shapes_c', comp_list_sdl)
                    
                    for sort_type in temp_sort_dict:
                        if sort_type not in sort.register:
                            sort.register[sort_type] = temp_sort_dict[sort_type]  
                        
                else:
                    comp_list_sdl = []
                
                    # Collect all disjunctive sorts inside sort register
                    for sort_type in sort.register:
                        if sort.register[sort_type].type == 'sort-disjunctive':
                            for base_geom in base_geometry:
                                if base_geom in sort.register[sort_type].component[0].baseSort().characteristic.type and sort.register[sort_type] not in comp_list_sdl:
                                    comp_list_sdl.append(sort.register[sort_type])
                                    break
                    
                    # Removes dummy description sorts
                    for i in range(2):
                        comp_list_temp = clean_sorts(comp_list_sdl)
                    
                    # Adds disjunctive dummy sort to prevent ejection from rhino_shapes_c sort later on
                    if len(comp_list_temp) <2:
                        comp_list_temp.append(__dummyDesc__)

                    # Creates new rhino_shapes_c with combined contents of both rhino_shapes_c and disjunctive sorts
                    compound_sort = compoundSort('rhino_shapes_c', comp_list_temp)
                        
                    for sort_type in temp_sort_dict:
                        if sort_type not in sort.register:
                            sort.register[sort_type] = temp_sort_dict[sort_type]    
                        
            new_sdl_rules = []
            new_sdl_shapes = []
            
            # File update
            # Update rule sides to prevailing 'rhino_shapes_c' sort
            for rule_name in sdl_rules:
                if sdl_rules[rule_name].lhs.sort is not sort.register['rhino_shapes_c']:
                    # Obtains all prior data about rule before shift to new 'rhino_shapes_c'
                    rule_object = deepcopy(rule.register[rule_name])
                    rule_desc = rule_object.description
                    predicates = rule_object.getPredicates()
                    directives = rule_object.getDirectives()
                    
                    # Deletes old rule
                    del rule.register[rule_name]
                    
                    # Creates new rule
                    new_lhs  = update_sort(rule_object.lhs.name, rule_object.lhs)
                    new_rhs = update_sort(rule_object.rhs.name, rule_object.rhs)
                    new_rule = rule(rule_name, rule_desc, new_lhs, new_rhs)
                    
                    # Adds in predicate/directive information from old rule before sort shift
                    if type(predicates) == type([]):
                        for predicate in predicates:
                            new_rule.addPredicate(predicate[0], predicate[1])
                            
                    if type(directives) == type([]):
                        for directive in directives:
                            new_rule.addDirective(directive[0], directive[1])
                            
                    new_sdl_rules.append(new_rule)
                else:
                    new_sdl_rules.append(sdl_rules[rule_name])
                    
            # Update shape to prevailing 'rhino_shapes_c' sort
            for shape_name in sdl_shapes:
                if sdl_shapes[shape_name].sort is not sort.register['rhino_shapes_c']:
                    shape_object = sdl_shapes[shape_name].duplicate()
                    del form.formRegister[shape_name]
                    new_sdl_shapes.append(update_sort(shape_name, shape_object))
                else:
                    new_sdl_shapes.append(sdl_shapes[shape_name])
            
            sc.sticky['mainSort'] = 'rhino_shapes_c'
            sdlParser.writeGrammar(fileName, new_sdl_rules, new_sdl_shapes, sdl_flows, [sort.register['rhino_shapes_c']])
            
            for new_rule in new_sdl_rules:
                sdlParser.rules[new_rule.name] = deepcopy(new_rule)
                
            for new_shape in new_sdl_shapes:
                sdlParser.forms[new_shape.name] = deepcopy(new_shape)
            
            for shape_name in temp_shape_dict:
                form.formRegister[shape_name] = temp_shape_dict[shape_name]
            
            return True
                    
    else:
        raise KeyError(fileName + " does not exist")
        return False
            
"""
=====================
SUPPORT FUNCTION INFO
=====================
'clean_sorts' removes any dummy disjunctive sorts or disjunctive sorts composed only of attribute sort types from a list

INPUT
List of sorts

OUTPUT
List of sorts containing only disjunctive sorts with geometric bases/description bases

Data Flow: List of disjunctive sorts
        -> List of disjunctive sorts (only those with geometric/description bases)
"""

def clean_sorts(sort_list):
    for sort_type in sort_list:
        if sort_type.type == 'sort-disjunctive' and sort_type.name == '__dummyDescDisjunctive__' and sort_type.name == '__dummyDescDisj__':
            sort_list.pop(sort_list.index(sort_type))
        else:
            move_on = False
            for component in sort_type.component:
                for attribute_type in attribute_bases:
                    if attribute_type in component.baseSort().characteristic.type:
                        sort_list.pop(sort_list.index(sort_type))
                        move_on = True
                        break
                        
                if move_on:
                    break
                    
    return sort_list

"""
=====================
SUPPORT FUNCTION INFO
=====================
'update_sort' updates the highest tier sort type of a rule object or shape to the compound sort 'rhino_shapes_c'

INPUT
Object (sortal rule or shape object)

OUTPUT
Sortal form with sort type set to prevailing iteration of rhino_shapes_c

Data Flow: Sortal form with sort type other than prevailing iteration of compound sort rhino_shapes_c
        -> sortal form with sort type being prevailing iteration of compound sort rhino_shapes_c
"""

def update_sort(shape_name, object):
    sortal_object = sort.register['rhino_shapes_c'].newForm()
    
    if object.sort.type == 'sort-compound':
        elements = object.elements
        for element in elements:
            sortal_object.insert(element)
    elif object.sort.type == 'sort-primitive' or object.sort.type == 'sort-attribute': # think about this case
        sort_name = object.sort + '_disjunctive'
        disjunctiveSort(sort_name, [object.sort, __dummyDesc__])
        newForm = sort.register[sort_name].newForm()
        newForm.insert(sortal_object)
        sortal_object.insert(newForm)
    else:
        sortal_object.insert(object)
        
    sortal_object.register(shape_name)

    return sortal_object
