
# -*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

# RHINO SPECIFIC IMPORTS | FROM PYTHON 3.5 to 2.7
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

# Rhino Scripting
import rhinoscriptsyntax as rs
import Rhino as r
import scriptcontext as sc

# Python built-ins
from builtins import range

# Sortal Library Imports
from sortal import (transformSet, rule, primitiveSort, point3D, point2D, coordinate, lineSegment3D, planeSegment2D, 
                    lineSegment2D, line3D, label, rWeight, disjunctiveSort, attributeSort, affineTransform,  
                    vector3D, vector2D, planeSegment3D, sort, coordinate, sort, flow)
                   
# Import for transformations
from mpmath import mp, pi, sin, cos, tan
from math import sqrt
import random

# For deepcopying variable contents
from copy import deepcopy

"""
Convert CLASS
Sorts objects into points, line segments, with their accompanying attributes
Performs conversion of Rhino GUIDs to agnostic form

INPUT
Two options:
1 Free selection of Rhino objects (sample code below after the class definition)
2 From Layers + Frames (using extract Class from 'extract.py', inside convert file)

OUTPUT
Agnostic form dictionary and corresponding list of Rhino GUIDs, sorted into geometric object types 
(i.e. ({'points': [<points>], 'line segments': [<lines>], 'plane segments': [<lines>]))

CURRENTLY SUPPORTS
Individuals: Points, lines, polylines, rectangles, BREPs/surfaces, meshes
Attributes: Labels, label descriptions, enumerative, weights (0 to 1, line widths), weights (grayscale from 0 to 255), color (RGB)
"""

# Searches for original Rhino geometry of GUID
rhdoc = r.RhinoDoc.ActiveDoc
testPt = r.Geometry.Point3d(0,0,0)

class convertR(object):
    
    # Input: object GUIDs of form, insertion point of frame selected, disjunctive sort tuple
    # Object IDs - Rhino GUIDs; disjunctive sort tuple - what individuals and attributes (if any) are to compose the form
    # Returns agnostic form (dictionary) with format: {'points': [<p1>, <p2>], 'line segments': [<ls1>, <ls2>]}
    
    def __init__(self):
        pass
    
    # Default refPoint is origin
    def toAgnostic(self, objectIds, classification = None, descriptions = '', refPoint = r.Geometry.Point3d(0,0,0)):
        self.objectIds = objectIds 
        self.lines = []                 # Contains GUIDs of all curve objects
        self.points = []                # Contains GUIDs of all point and text dot objects
        self.planes = []
     
        # Contains GUIDs of sorted curve and point/text dot objects in dictionary; subdictionaries are arranged by sort name 
        #(determined either by layer name or by random selection
        self.objects = {'point':{}, 'line_segment':{}, 'plane_segment':{}, 'circle': {}, 'ellipse': {}, 'cArc':{}, 'eArc': {}, 'bezier_curve': {}}  
        self.form = None                                # Contains final agnostic form
        self.descriptions = descriptions                # Sets description string

        # Sets reference point used to obtain relative vector coordinates
        if type(refPoint) == r.Geometry.Point3d:
            self.framePosition = r.Geometry.Vector3d(refPoint.X,refPoint.Y,refPoint.Z)  
        elif type(refPoint) == type([]) or type(refPoint) == type(()):
            self.framePoistion = r.Geometry.Vector3d(refPoint[0], refPoint[1], refPoint[2])
        else:
            try:
                if rs.ObjectType(refPoint) == 1:
                    coords = rs.PointCoordinates(refPoint)
                    coords = r.Geometry.Vector3d(coords.X,coords.Y,coords.Z)  
                elif rs.ObjectType(refPoint) == 8192:
                    coords = rs.TextDotPoint(refPoint)
                    coords = r.Geometry.Vector3d(coords.X,coords.Y,coords.Z)  
                self.framePosition = coords
            except:
                raise ValueError('Invalid reference point input')
     
        self.setDesc = False                            # Tracks if description sorts are present in overarching sort type
        self.classification = ''
        self.compSort = False
        self.attributes = {}                            # Tracks enabled attributes per major geometric sort type
        self.lineCount = 0
        layerNames = rs.LayerNames(True)
        
        # Compound sort is defined by API OR
        # sgi.openSDL was used to load up an SDL file -> sc.sticky['mainSort'] is present
        if 'rhino_shapes_c' in sort.register:
            if len(sort.register['rhino_shapes_c'].component) == 1:
                self.discSort = sort.register['rhino_shapes_c'].component[0]
                
            elif (sort.register['rhino_shapes_c'].component) > 1:
                sortType = None
                if classification != None:
                    if classification in sort.register:
                        for sortType in sort.register['rhino_shapes_c'].component:
                            if sortType.name == classification:
                                self.discSort = sortType
                                break
                    #elif classification not in sort.register:
                    #    if len(sort.register['rhino_shapes_c'] == 1:
                    #         self.discSort = sort.register['rhino_shapes_c'].component[0]
                    #    for sortType in sort.register['rhino_shapes_c'].component:
                            
                elif classification == None:
                    breakVal = False
                    
                    for sortType in sort.register['rhino_shapes_c'].component:
                        if sortType.type == 'sort-disjunctive':
                            for subSortType in sortType.component:
                                if subSortType.type == 'sort-attribute' or (subSortType.type == 'sort-primitive' and 'description' not in subSortType.baseSort().characteristic.type):
                                    self.discSort = sortType
                                    breakVal = True
                                    break
                            
                            if breakVal == True:
                                break
                                   
                if sortType == None:
                    go_ahead = rs.MessageBox(str(classification) + " is not present in the compound sort 'rhino_shapes_c'",  0, "Sort type not present in compound sort")         
                    return None, None
            
            else: 
                go_ahead = rs.MessageBox("The compound sort 'rhino_shapes_c' is empty.",  0, "Empty compound sort")         
                return None, None
                
        else:
            go_ahead = rs.MessageBox("The compound sort 'rhino_shapes_c' has not been set up.",  0, "No compound sort")         
            return None, None
      
        # List of sort types of main sorts
        # Agnostic object is mapped to structure of self.discSort
        self.mainSorts =  self.discSort.component           
        
        vectorType = '3D'
        for mainSort in self.mainSorts:
           if '2D' in mainSort.baseSort().characteristic.type:
                vectorType = '2D'
                break 
                
        self.predicates = {'max':[], 'shortest':[], 'longest':[], 'void':{}, 'bound': {}, 'nolabel': [], 'embeds': {}}
        self.directives = {'pointOL': [], 'distance': {}, 'normal': {}}
        self.voidSorts = {}
        self.retrieve = {'pointOL': 'seg', 'length' : 'len', 'distance': 'dist', 'normal': 'norm'}
        organize = ['pointOL']
        
        # Checks if convert class received any objects
        if self.objectIds == None:
            ('No objects selected')
            return None, None        

        # Separates objects into individual types (points, lines, etc.) and handles selectively
        # based on metaform sort type
        self.separateObjects(self.objectIds, self.mainSorts)

        # Extracts data from object ids and sorts them according to individuals or as a list of individuals (e.g. polylines)
        self.agnosticObject, self.formIds = self.transformObjects(self.objects)
        # Adds descriptions to agnostic object
        if descriptions != '' and self.setDesc:
            self.addDescriptions()
        
        count = 0
        descCount = 0
        
        # Checks if agnostic object is empty from descriptions and geometries
        for geomType in self.agnosticObject:
            if len(self.agnosticObject[geomType]) == 0:
                count += 1
                
        #if count == len(self.agnosticObject):
        #    raise ValueError('The geometries/descriptions you have inputted are not supported by the currently active compound sort', 0, 'Geometries not supported by compound sort')
        
        self.finalAgnostic = {}
        
        # PROBLEM HERE
        # Sets up sort type classification (e.g. 2D, 3D, P2D, P3D as the key name in a compound agnostic object) of current agnostic object based on user input
        # Returns agnostic object as is (not classed under disjunctive sort name) if there is only one disjunctive sort aside from the dummy descriptions disjunctive sort present in rhino_shapes_c
        # Otherwise, returns agnostic object classed under its disjunctive sort
        if len(sort.register['rhino_shapes_c'].component) == 2:
            if '__dummySort__' in sc.sticky:
                if sc.sticky['__dummySort__'] in sort.register:
                    if sort.register['__dummyDescDisjunctive__'] in sort.register['rhino_shapes_c'].component:
                        self.finalAgnostic = self.agnosticObject
            elif '__dummyDescDisjunctive__' in sort.register:
                if sort.register['__dummyDescDisjunctive__'] in sort.register['rhino_shapes_c'].component:
                    self.finalAgnostic = self.agnosticObject
                        
        elif len(sort.register['rhino_shapes_c'].component) >= 2: 
            if classification == None :
                self.finalAgnostic[self.discSort.name] = [self.agnosticObject]
            else:
                self.finalAgnostic[classification] = [self.agnosticObject]
        else:
            self.finalAgnostic = self.agnosticObject

        # Used as basis for determining type of vectors to use in creating predicates
        self.predicates['classification'] = vectorType     
        self.directives['classification'] = vectorType

        # Cleans through predicates and directives to remove any redundancies
        # Currently using the list of strings 'organize' to allow for cleaning as this behavior may change
        for i in self.predicates:
            if i in organize:
                self.cleanTags(self.predicates, i)

        for i in self.directives:
            if i in organize:
                self.cleanTags(self.directives, i)
                
        return self.finalAgnostic, self.formIds, self.predicates, self.directives

    def determineAttributes(self, sortType, attributes):
        if sortType.name not in self.attributes:
            self.attributes[sortType.name] = []
        
        for att in attributes:
            if 'description' in att.baseSort().characteristic.type:
                self.attributes[sortType.name].append('desc')
            
            if 'enum' in att.baseSort().characteristic.type:
                self.attributes[sortType.name].append('enum')
            
            if 'label' in att.baseSort().characteristic.type:
                self.attributes[sortType.name].append('label')
                
            if 'rWeight' in att.baseSort().characteristic.type:
                if 'weight255' in att.name:
                    self.attributes[sortType.name].append('weight255')
                elif 'weight' in att.name:
                    self.attributes[sortType.name].append('weight')
            
            if 'color' in att.baseSort().characteristic.type:
                self.attributes[sortType.name].append('color')
                
        return
        
    """
    =================================
    CONVERTING GUIDs to AGNOSTIC FORM 
    =================================
    """
    
    # Separates objects into geometric ids and chooses which ones to handle, based on user sort type input
    # Example: if a user only wants to create a form of lines, this function will select 
    # only the lines/curve objects to put inside the form
    def separateObjects(self, objectIds, mainSorts):
        # Sorts objects into curves (line segments) and points/text dots
        # If block instance or polyline, explodes block instance and adds the GUIDs to the objectIds list
        pointHold = None
        lineSegHold = None
        planeSegHold = None
        cHold = None
        eHold = None
        cArcHold = None
        eArcHold = None
        bezHold = None
        cArcHold = None
        eArcHold
        
        for a in self.mainSorts:
            if 'point' in a.baseSort().characteristic.type:
                self.objects['point'][a.name] = []
                if pointHold == None:
                    pointHold = a.name
                    
            elif 'lineSegment' in a.baseSort().characteristic.type:
                self.objects['line_segment'][a.name] = []
                if lineSegHold == None:
                    lineSegHold = a.name
                    
            elif 'planeSegment' in a.baseSort().characteristic.type:
                self.objects['plane_segment'][a.name] = []
                if planeSegHold == None:
                    planeSegHold = a.name
            
            elif 'circle' in a.baseSort().characteristic.type:
                self.objects['circle'][a.name] = []
                if cHold == None:
                    cHold = a.name
            
            elif 'ellipse' in a.baseSort().characteristic.type:
                self.objects['ellipse'][a.name] = []
                if eHold == None:
                    eHold = a.name
            
            elif 'Arc' in a.baseSort().characteristic.type:
                if 'circular' in a.baseSort().characteristic.type:
                    self.objects['cArc'][a.name] = []
                    if cArcHold == None:
                        cArcHold = a.name
                #"""        
                elif 'elliptical' in a.baseSort().characteristic.type:
                    self.objects['eArc'][a.name] = []
                    if eArcHold == None:
                        eArcHold = a.name
                #"""        
            elif 'bezier' in a.baseSort().characteristic.type:
                self.objects['bezier_curve'][a.name] = []
                if bezHold == None:
                    bezHold = a.name
                    
            if 'description' in a.baseSort().characteristic.type:
                self.setDesc = True
                
            if self.classification == '':
                if '2D' in a.baseSort().characteristic.type or '3D' in a.baseSort().characteristic.type:
                    self.classification = a.baseSort().characteristic.type[-2:]
                    
                if 'P' in a.baseSort().characteristic.type:
                    self.classification =  'P' + self.classification
            
            if a.type == 'sort-attribute':
                if a.weight.type == 'sort-disjunctive':
                    attributes = a.weight.component
                elif a.weight.type == 'sort-primitive':
                    attributes = [a.weight]
                    
                self.determineAttributes(a, attributes)
            
            else:
                self.attributes[a.name] = []
                
        for object in objectIds:
            if rhdoc.Objects.Find(object) != None:
                layerName = rs.ObjectLayer(object)
                # Handles all line segment/curve objects: line segments, circles, ellipses, arcs, Bezier curves
                try:
                    if (rs.IsCurve(object) or rs.IsPolyline(object)) and rs.IsCircle(object) == False and rs.IsArc(object) == False and rs.IsEllipse(object) == False and self.isEllipse(object)[0] == False:
                        crv = rs.coercecurve(object)
                        
                        if rs.IsCurve(object) and len(rs.CurvePoints(object)) == 3 and crv.HasBezierSpans: 
                            if layerName in sort.register:
                                self.objects['bezier_curve'][layerName].append(object)
                            elif bezHold != None:
                                self.objects['bezier_curve'][bezHold].append(object)
                            
                        elif (type(crv) == r.Geometry.LineCurve or type(crv) == r.Geometry.PolyCurve or type(crv) == r.Geometry.PolylineCurve or type(crv) == r.Geometry.Rectangle3d): 
                            if layerName in sort.register:
                                self.objects['line_segment'][layerName].append(object)
                            elif lineSegHold != None:
                                self.objects['line_segment'][lineSegHold].append(object)
                            
                    elif rs.IsCircle(object) and rs.IsCurveClosed(object):
                        if layerName in self.objects['circle'] and layerName in sort.register:
                            self.objects['circle'][layerName].append(object)
                        elif cHold != None:
                            self.objects['circle'][cHold].append(object)
                    
                    elif ((rs.IsEllipse(object) and rs.IsCurveClosed(object)) or (self.isEllipse(object)[0] and rs.IsCurveClosed(object))):
                        if layerName in self.objects['ellipse'] and layerName in sort.register:
                            self.objects['ellipse'][layerName].append(object)
                        elif eHold != None:
                            self.objects['ellipse'][eHold].append(object)
                        
                    elif rs.IsArc(object):
                        if layerName in self.objects['cArc'] and layerName in sort.register:
                            self.objects['cArc'][layerName].append(object)
                        elif cArcHold != None:
                            self.objects['cArc'][cArcHold].append(object)
                        
                    elif self.checkEllipse(object) == True:
                        crv = rs.coercecurve(object)
                        if type(crv) == r.Geometry.PolyCurve or type(crv) == r.Geometry.PolylineCurve or type(crv) == r.Geometry.NurbsCurve:
                            if crv.HasBezierSpans and len(rs.CurvePoints(object)) == 3:
                                if layerName in self.objects['bezier_curve'] and layerName in sort.register:
                                    self.objects['bezier_curve'][layerName].append(object)
                                elif bezHold != None:
                                    self.objects['bezier_curve'][bezHold].append(object)
                                    
                            # Originally for eArcs
                            #"""
                            elif type(rhdoc.Objects.Find(object).Geometry) == r.Geometry.PolyCurve:
                                nurbs_obj = rhdoc.Objects.Find(object).Geometry.Duplicate().ToNurbsCurve()
                                if layerName in sort.register:
                                    self.objects['eArc'][layerName].append([object, nurbs_obj])
                                elif eArcHold != None:
                                    self.objects['eArc'][eArcHold].append([object, nurbs_obj])
                            elif eArcHold != None:
                                self.objects['eArc'][eArcHold].append(object) 
                        #"""
                #"""     
                except:
                    if (rs.ObjectType(object) == 1 or rs.ObjectType(object) == 8192 or rs.IsPoint(object) or rs.IsTextDot(object)):
                        if layerName in self.objects['point'] and layerName in sort.register:
                            self.objects['point'][rs.ObjectLayer(object)].append(object)
                        elif 'visual' not in rs.GetUserText(object) and pointHold != None:
                            self.objects['point'][pointHold].append(object)
                            
                    elif (rs.ObjectType(object) == 8 or rs.ObjectType(object) == 16 or rs.ObjectType(object) == 32):
                        if layerName in self.objects['plane_segment'] and layerName in sort.register:
                            self.objects['plane_segment'][layerName].append(object)
                        elif planeSegHold != None:
                            self.objects['plane_segment'][planeSegHold].append(object)
    
                    elif rs.IsBlockInstance(object):  
                        blockObjects = rs.ExplodeBlockInstance(object)
                        index = objectIds.index(object)
                        objectIds.pop(index)
                        
                        for obj in blockObjects:
                            objectIds.append(object)
                #"""
        return
                
    # Transforms objects into agnostic form
    def transformObjects (self, objects):
        form = {}       # Stores agnostic individuals according to hash keys based on geometry type
        formIds = {}    # Stores corresponding GUIDs of agnostic individuals into corresponding lists based on geometry type
        geomSorts = ['point', 'line_segment', 'plane_segment', 'circle', 'ellipse', 'cArc', 'bezier_curve']
        
        for geomType in objects:
            for sortName in objects[geomType]:
                attribCheck = self.attributes[sortName]
                geomObject = objects[geomType]
                
                if sortName not in form and sortName not in formIds:
                    form[sortName] = []
                    formIds[sortName] = []
                
                for geometry in geomObject[sortName]:
                    newObject = None
                    indivType = ''
                    
                    if type(geometry) == type([]):
                        indiv = geometry[0]
                    else:
                        indiv = geometry
                        
                    if 'visual' in rs.GetUserText(indiv):
                        continue

                    if geomType == 'bezier_curve' and len(rs.CurvePoints(indiv)) == 3:
                        newObject = self.bezCurve(indiv, sortName)
                 
                    elif geomType == 'line_segment':
                        # Handles individual line
                        # If line start == line end, then this creates an error and the code proceeds to handle the curve
                        # as a polyline    
                        try:
                            # Obtains line information (coords, attribute information) from line GUID
                            # Returns agnostic tuple form of line 
                            newObject = self.Line(indiv, sortName)  
                        # Handles polyline
                        except:
                            # Obtains line information (coords, attribute information) from polyline GUID
                            # Returns agnostic tuple form of polyline 
                            newObject = self.Polyline(indiv, sortName)
                            
                        self.lineCount += 1

                    elif geomType == 'circle':
                        newObject = self.Circle(indiv, sortName)

                    elif geomType == 'ellipse':
                        newObject = self.Ellipse(indiv, sortName)
                        
                    elif geomType == 'cArc':
                        newObject = self.cArc(indiv, sortName)

                    elif geomType == 'eArc':
                        if self.checkEllipse(indiv):
                            newObject = self.eArc(indiv, sortName)
                        elif type(geometry) == type([]):
                            if type(geometry[1]) == r.Geometry.NurbsCurve:
                                newObject = self.eArc(geometry, sortName)
                        else:
                            warnings.warn('Arc object cannot be converted to elliptical arc and will not be included in the agnostic shape')

                    # Checks if GUID is point or text dot
                    elif geomType == 'point':
                        # Obtains point information (coords, attributes) from point GUID
                        # Returns agnostic tuple form of point
                        newObject = self.Point(indiv, sortName)
                    elif geomType == 'plane_segment':
                        # Obtains point information (coords, attributes) from point GUID
                        # Returns agnostic tuple form of point
                        newObject = self.Plane(indiv, sortName)
            
                    # If newObject has been created, it is appended to form and its GUID is appended to formIds
                    if newObject != None:
                        if geomType == 'line_segment' and len(newObject) > 0:
                            formIds[sortName].append(indiv)  

                            if type(newObject[0][0][0]) == type(1.0) or type(newObject[0][0][0]) == type(0):
                                form[sortName].append(newObject)
                            else:
                                for obj in newObject:
                                    form[sortName].append(obj)

                        elif geomType in geomSorts:
                            formIds[sortName].append(indiv)
                            form[sortName].append(newObject)

        # To put together similarly named void groups or coordinates that belong in the same void group
        for groupNameA in self.predicates['void']:
            for groupNameB in self.voidSorts:
                if groupNameA == groupNameB:
                    self.predicates['void'][groupNameA] = list(self.predicates['void'][groupNameA])
                    self.predicates['void'][groupNameA].insert(0, self.voidSorts[groupNameB])
                    self.predicates['void'][groupNameA] = tuple(self.predicates['void'][groupNameA])
        
        # To remove double coordinates within the void groups 
        # (i.e. triangular voids should have three elements, square or rectangular voids should have four elements)
        for groupName in self.predicates['void']:
            coordinates = deepcopy(self.predicates['void'][groupName])
            coordinates = list(coordinates)
            for coordA in self.predicates['void'][groupName]:
                count = 0
                for coordB in coordinates:
                    if coordA == coordB:
                        if count == 0:
                            count+=1
                        elif count >= 1:
                            coordinates.pop(coordinates.index(coordB))
                            count+=1
            self.predicates['void'][groupName] = deepcopy(coordinates)
            
        return form, formIds
    
    """
    ==========================
    EXTRACTING DATA from GUIDs
    ==========================
    """

    # Extracts data from curve (line segment) GUID
    def Line(self, id, sortName):
        tail, head = 0, 0
        # Adjusts coordinates based on frame's Block Instance Insertion Point 
        p1_abs = rs.CurveStartPoint(id)
        p2_abs = rs.CurveEndPoint(id)
        p1_rel = rs.PointSubtract(p1_abs, self.framePosition)
        p2_rel = rs.PointSubtract(p2_abs, self.framePosition)
        
        # Determines order of plotting
        if p1_rel < p2_rel:
            tail = p1_rel
            head = p2_rel
        elif p1_rel > p2_rel:
            tail = p2_rel
            head = p1_rel
        else:
            raise ValueError
        
        # Creates vectors 
        if '3' in self.classification:
            v1 = (head[0], head[1], head[2])
            v2 = (tail[0], tail[1], tail[2])
        else: 
            v1 = (head[0], head[1])
            v2 = (tail[0], tail[1])
 
        coords = (v1,v2)
        
        # Creates attributes dictionary for point
        att = self.createAtt('lineSegment', id, coords, sortName)
        # Extracts predicates/directives from geometry
        self.createTags(att, 'lineSegment', id, coords, sortName)
        
        if 'tags' not in att:
            att['tags'] = []
        
        tagName = rs.GetUserText(id, 'tag')
        
        return (coords, att)
    
    # Extracts data from curve (polyline) GUID
    def Polyline(self, id, sortName):
        vertices = rs.PolylineVertices(id)
        vectors_temp = []
        vectors = []

        # Adjusts coordinates based on frame's Block Instance Insertion Point 
        # Creates vectors
        for vector in vertices:
            vector_rel  = rs.PointSubtract(vector, self.framePosition)
            
            # Creates vectors 
            if '3' in self.classification:
                vectors_temp.append((vector_rel.X, vector_rel.Y, vector_rel.Z))
            else: 
                vectors_temp.append((vector_rel.X, vector_rel.Y))

        seen = set()
        vectors[:] = [vector for vector in vectors_temp if vector not in seen and not seen.add(vector)]
        vectorList = []
        # Creates attributes dictionary for point
        att = self.createAtt('lineSegment', id, vectors, sortName)
        # Extracts predicates/directives from geometry
        self.createTags(att, 'polyline', id, vectors, sortName)
        
        for count, vector in enumerate(vectors):

            if count == (len(vectors)-1): 
                if len(vectors) > 3 and vector != vectors[0]:
                    newVector = ((vector, vectors[0]), att)
                else:
                    break
            else:
                newVector = ((vector, vectors[count+1]), deepcopy(att))
               
            vectorList.append((newVector))

        return vectorList
   
    # Extracts data from point (point/text dot) GUID
    def Point(self, id, sortName):
        # Obtains color information from point
        color = rs.coercecolor(rs.ObjectColor(id))
        rgb = [color.R, color.G, color.B]
        
        if rs.IsTextDot(id):
            coord_abs = rs.TextDotPoint(id) 
        else:
            coord_abs = rs.PointCoordinates(id)

        coord_rel = rs.PointSubtract(coord_abs, self.framePosition)

        if '3' in self.classification:
            coord = (coord_rel.X, coord_rel.Y, coord_rel.Z)
        else:
            coord = (coord_rel.X, coord_rel.Y)
        
        # Creates attributes dictionary for point
        att = self.createAtt('point', id, coord, sortName)
        # Extracts predicates/directives from geometry
        self.createTags(att, 'point', id, coord, sortName)
        
        return (coord, att)

    # Extracts data from plane segment (mesh/surface/brep) GUID
    def Plane(self, id, sortName):
        # Obtains color information from point
        color = rs.coercecolor(rs.ObjectColor(id))
        rgb = [color.R, color.G, color.B]
        obj = rhdoc.Objects.Find(id)
        plane = []

        if rs.IsBrep(id) or rs.IsSurface(id):
            plane = {}
            plane['inner'] = []
            outer = ()
            # BrepObjects
            if type(obj) == r.DocObjects.BrepObject:
                for curve in obj.BrepGeometry.Loops:
                    thing = curve.To3dCurve()
                    curves = thing.DuplicateSegments()
                    outerBound = []
    
                    # Holes
                    # May be more than 1
                    if curve.LoopType == r.Geometry.BrepLoopType.Inner:
                        inner = ()
                        for thing in curves:
                            coord_abs = thing.PointAtStart
                            coord_rel = rs.PointSubtract(coord_abs, self.framePosition)
                            
                            if '3' in self.classification:
                                coord = (coord_rel.X, coord_rel.Y, coord_rel.Z)
                            else:
                                coord = (coord_rel.X, coord_rel.Y,0)
                            inner = inner + (coord,)
                        
                        plane['inner'].append(inner)
                    
                    # Outer bound
                    # Only 1 per brep
                    elif curve.LoopType == r.Geometry.BrepLoopType.Outer:
                        for thing in curves:
                            coord_abs = thing.PointAtStart
                            coord_rel = rs.PointSubtract(coord_abs, self.framePosition)
                            
                            if '3' in self.classification:
                                coord = (coord_rel.X, coord_rel.Y, coord_rel.Z)
                            else:
                                coord = (coord_rel.X, coord_rel.Y)
                                
                            outer = outer + (coord,)
    
                        plane['outer']= outer
            # Breps
            elif type(obj) == r.Geometry.Brep:
                for curve in obj.Loops:
                    thing = curve.To3dCurve()
                    curves = thing.DuplicateSegments()
                    outerBound = []
    
                    # Holes
                    # May be more than 1
                    if curve.LoopType == r.Geometry.BrepLoopType.Inner:
                        inner = ()
                        for thing in curves:
                            coord_abs = thing.PointAtStart
                            coord_rel = rs.PointSubtract(coord_abs, self.framePosition)
                            
                            if '3' in self.classification:
                                coord = (coord_rel.X, coord_rel.Y, coord_rel.Z)
                            else:
                                coord = (coord_rel.X, coord_rel.Y)
                                
                            inner = inner + (coord,)
                        
                        plane['inner'].append(inner)
                    
                    # Outer bound
                    # Only 1 per brep
                    elif curve.LoopType == r.Geometry.BrepLoopType.Outer:
                        for thing in curves:
                            coord_abs = thing.PointAtStart
                            coord_rel = rs.PointSubtract(coord_abs, self.framePosition)
                            
                            if '3' in self.classification:
                                coord = (coord_rel.X, coord_rel.Y, coord_rel.Z)
                            else:
                                coord = (coord_rel.X, coord_rel.Y)
                                
                            outer = outer + (coord,)
    
                        plane['outer']= outer
            
        # Meshes
        elif rs.IsMesh(id):
            if type(obj) == r.DocObjects.MeshObject:
                count =  obj.Geometry.Faces.Count
                plane = [] 
                
                for ind in range(count):
                    vertices = obj.Geometry.Faces.GetFaceVertices(ind)[1:]
                    
                    for vertex in vertices:
                        coord_abs = vertex
                        coord_rel = rs.PointSubtract(coord_abs, self.framePosition)
                        
                        if '3' in self.classification:
                            coord = (coord_rel.X, coord_rel.Y, coord_rel.Z)
                        else:
                            coord = (coord_rel.X, coord_rel.Y)
                            
                        plane.append(coord)
                        
            elif type(obj) == r.Geometry.Mesh:
                count =  obj.Faces.Count
                plane = [] 
                
                for ind in range(count):
                    vertices = obj.Faces.GetFaceVertices(ind)[1:]
                    
                    for vertex in vertices:
                        coord_abs = vertex
                        coord_rel = rs.PointSubtract(coord_abs, self.framePosition)
                        
                        if '3' in self.classification:
                            coord = (coord_rel.X, coord_rel.Y, coord_rel.Z)
                        else:
                            coord = (coord_rel.X, coord_rel.Y)
                            
                        plane.append(coord)
        
        # Creates attributes dictionary for plane
        att = self.createAtt('planeSegment', id, plane, sortName)
        # Extracts predicates/directives from geometry
        self.createTags(att, 'planeSegment', id, plane, sortName)
        
        return (plane, att)
    
    
    # Extracts data from circle GUID (full 360 deg circle)
    def Circle(self, id, sortName):
        ctr = rs.CircleCenterPoint(id)
        rad = rs.CircleRadius(id)
        c_pt1 = rs.EvaluateCurve(id,0)
        c_pt2 = rs.EvaluateCurve(id,2)
        c_pt3 = rs.EvaluateCurve(id,10)         # Generate third point for rhino plotting purposes

        # Computes points relative to frame position/reference point
        ctr = rs.PointSubtract(ctr, self.framePosition)
        c_pt1 = rs.PointSubtract(c_pt1, self.framePosition)
        c_pt2 = rs.PointSubtract(c_pt2, self.framePosition)
        c_pt3 = rs.PointSubtract(c_pt3, self.framePosition)

        if '3' in self.classification:
            ctr_coords = (ctr.X, ctr.Y, ctr.Z)
            c_pt1_coords = (c_pt1.X, c_pt1.Y, c_pt1.Z)
            c_pt2_coords = (c_pt2.X, c_pt2.Y, c_pt2.Z)
            c_pt3_coords = (c_pt3.X, c_pt3.Y, c_pt3.Z)
            circle = (ctr_coords, rad, c_pt1_coords,  c_pt2_coords, c_pt3_coords)
        else:
            ctr_coords = (ctr.X, ctr.Y)
            c_pt1_coords = (c_pt1.X, c_pt1.Y)
            c_pt2_coords = (c_pt2.X, c_pt2.Y)
            circle = (ctr_coords, rad, c_pt1_coords,  c_pt2_coords)
        
        """
        rs.AddTextDot('center', ctr_coords)
        rs.AddTextDot('p1', c_pt1_coords)
        rs.AddTextDot('p2', c_pt2_coords)
        rs.AddTextDot('p3', c_pt3_coords)
        """
        
        # Creates attributes dictionary for circle
        att = self.createAtt('circle', id, circle, sortName)
        # Extracts predicates/directives from geometry
        self.createTags(att, 'circle', id, circle, sortName)
        return (circle, att)
    
    # Extracts data from ellipse GUID
    def Ellipse(self, id, sortName):
        if rs.IsEllipse(id):
            quad_points = rs.EllipseQuadPoints(id)
            center = rs.EllipseCenterPoint(id) 
            long_pt = quad_points[0]     # Major axis
            short_pt = quad_points[2]    # Minor axis
            
            rs.AddTextDot('center', center)
            rs.AddTextDot('p1', long_pt)
            rs.AddTextDot('p2', short_pt)
            
            # Computes points relative to frame position/reference point
            center = rs.PointSubtract(center, self.framePosition)
            long_pt = rs.PointSubtract(long_pt, self.framePosition)
            short_pt = rs.PointSubtract(short_pt, self.framePosition)
            rs.AddTextDot('center1', center)
            rs.AddTextDot('p11', long_pt)
            rs.AddTextDot('p21', short_pt)

            if '3' in self.classification:
                long_pt_coords = (long_pt.X, long_pt.Y, long_pt.Z)
                short_pt_coords = (short_pt.X, short_pt.Y, short_pt.Z)
            else:
                long_pt_coords = (long_pt.X, long_pt.Y)
                short_pt_coords = (short_pt.X, short_pt.Y)
            
            # Compute two foci of ellipse
            aDist = ((long_pt.X - center.X)**2 + (long_pt.Y - center.Y)**2 + (long_pt.Z - center.Z)**2)
            bDist = ((short_pt.X - center.X)**2 + (short_pt.Y - center.Y)**2 + (short_pt.Z - center.Z)**2)
            cDist = sqrt(aDist - bDist)
        
        # Ellipse is irregular or cannot be identified by Rhino
        elif self.isEllipse(id)[0] or type(self.isEllipse(id)[1]) == r.Geometry.Ellipse:
            ell = self.isEllipse(id)[1]
            center = ell.Plane.Origin
            norm = ell.Plane.Normal
            normDist = sqrt(norm.X**2 + norm.Y**2 + norm.Z**2)
            normUnit = r.Geometry.Vector3d(norm.X/normDist, norm.Y/normDist, norm.Z/normDist)
            radx = ell.Radius1   # x-distance
            rady = ell.Radius2   # y-distance
            short_pt = rs.EvaluateCurve(id, rs.CurveClosestPoint(id, center))
            shortVec = short_pt - center
            shortDist = sqrt(shortVec.X**2 + shortVec.Y**2 + shortVec.Z**2)
            shortUnit = r.Geometry.Vector3d(shortVec.X/shortDist, shortVec.Y/shortDist, shortVec.Z/shortDist)
            # Generates major axis' unit vector
            longUnit = r.Geometry.Vector3d.CrossProduct(shortUnit, normUnit)
            
            if rady > radx:
                cDist = sqrt(rady**2 - radx**2)
                long_pt = center + r.Geometry.Vector3d(radx, 0, 0)
                bDist = radx
                aDist = rady
            else:
                cDist = sqrt(radx**2 - rady**2)
                long_pt = center + r.Geometry.Vector3d(0, rady, 0)
                bDist = rady
                aDist = radx
            
            long_pt = center + r.Geometry.Vector3d.Multiply(longUnit, aDist)
            
        fociVec = long_pt - center
        unitDist = sqrt(fociVec.X**2 + fociVec.Y**2 + fociVec.Z**2)
        unitVec = r.Geometry.Vector3d(fociVec.X/unitDist, fociVec.Y/unitDist, fociVec.Z/unitDist)
        focus1 = center + r.Geometry.Vector3d.Multiply(unitVec, cDist)
        focus2 = center - r.Geometry.Vector3d.Multiply(unitVec, cDist)
        short_pt = rs.EvaluateCurve(id, rs.CurveClosestPoint(id, center))
        short_pt = rs.PointSubtract(short_pt, self.framePosition)
        
        center_coords = (center.X, center.Y, center.Z)
        foc1_coords = (focus1.X, focus1.Y, focus1.Z)
        foc2_coords = (focus2.X, focus2.Y, focus2.Z)
        min_coords = (short_pt.X, short_pt.Y, short_pt.Z)
        maj_coords = (long_pt.X, long_pt.Y, long_pt.Z)
        
        ellipse = (center_coords, foc1_coords, foc2_coords, min_coords, maj_coords)
        
        """
        rs.AddTextDot('center', (center[0], center[1], center[2]))
        rs.AddTextDot('foc1', (foc1_coords[0], foc1_coords[1], foc1_coords[2]))
        rs.AddTextDot('foc2', (foc2_coords[0], foc2_coords[1], foc2_coords[2]))
        rs.AddTextDot('short', (min_coords[0], min_coords[1], min_coords[2]))
        rs.AddTextDot('long', (maj_coords[0], maj_coords[1], maj_coords[2]))
        """
        
        # Creates attributes dictionary for circle
        att = self.createAtt('ellipse', id, ellipse, sortName)
        # Extracts predicates/directives from geometry
        self.createTags(att, 'ellipse', id, ellipse, sortName)
        return (ellipse, att)
        
    # Extracts data from circular arc GUID
    def cArc(self, id, sortName):
        center = rs.ArcCenterPoint(id)
        start_point = rs.CurveStartPoint(id)
        end_point = rs.CurveEndPoint(id)
        pin = rs.ArcMidPoint(id)
        
        # Computes points relative to frame position/reference point
        center = rs.PointSubtract(center, self.framePosition)
        start_point = rs.PointSubtract(start_point, self.framePosition)
        end_point = rs.PointSubtract(end_point, self.framePosition)
        pin = rs.PointSubtract(pin, self.framePosition)
        
        if '3' in self.classification:
            center_coords =(center.X, center.Y, center.Z)
            start_point_coords =(start_point.X, start_point.Y, start_point.Z)
            end_point_coords =(end_point.X, end_point.Y, end_point.Z)
            pin_coords =(pin.X, pin.Y, pin.Z)
        
        else:
            center_coords =(center.X, center.Y)
            start_point_coords =(start_point.X, start_point.Y)
            end_point_coords =(end_point.X, end_point.Y)
            pin_coords =(pin.X, pin.Y)
        
        #rs.AddTextDot('center', (center_coords[0], center_coords[1], center_coords[2]))
        #rs.AddTextDot('start', (start_point_coords[0], start_point_coords[1], start_point_coords[2]))
        #rs.AddTextDot('end', (end_point_coords[0], end_point_coords[1], end_point_coords[2]))
        #rs.AddTextDot('pin', (pin_coords[0], pin_coords[1], pin_coords[2]))
        
        cArc = (center_coords, start_point_coords, end_point_coords, pin_coords)
        # Creates attributes dictionary for circle
        att = self.createAtt('cArc', id, cArc, sortName)
        # Extracts predicates/directives from geometry
        self.createTags(att, 'cArc', id, cArc, sortName)
        return (cArc, att)
        
    # Extracts data from elliptical arc GUID    
    # Data set: (<center>, <foc1>, <foc2>, <start>, <end>, <maj_axis_pt or randomly generated point serves as pin>, <minor axis quadrant point>, 
    # <major axis quadrant point>)
    def eArc(self, id, sortName):
        if type(id) == type([]):
            try:
                result1 = rs.IsEllipse(id[0])
            except:
                result1 = None
        
            result2 = self.isEllipse(id[1])
        else:
            try:
                result1 = rs.IsEllipse(id)
            except:
                result1 = None
        
            result2 = self.isEllipse(id)
            
        if result1 or result2:
            if result1 == True:
                quad_points = rs.EllipseQuadPoints(id)
                center = rs.EllipseCenterPoint(id) 
                ellipse_obj = self.isEllipse(id)[1]
                normal = ellipse_obj.Plane.Normal
                long_pt = quad_points[1]     # Major axis
                short_pt = quad_points[2]    # Minor axis
                start_pt = rs.CurveStartPoint(id)
                end_pt = rs.CurveEndPoint(id)
                pin_pt = long_pt
                radx = ellipse_obj.Radius1
                rady = ellipse_obj.Radius2
            elif result2:
                if type(id) == type([]):
                    nurbs_obj = id[1]
                    ellArc = id[0]
                    if nurbs_obj.IsEllipse():
                        ellipse_obj = nurbs_obj.TryGetEllipse()[1]
                        center =  ellipse_obj.Plane.Origin
                        short_pt = nurbs_obj.PointAt(nurbs_obj.ClosestPoint(center)[1])
                        start_pt = rs.CurveStartPoint(ellArc)
                        end_pt = rs.CurveEndPoint(ellArc)
                    
                    else:
                        warnings.warn('Arc-like object cannot be processed by API')
                        return

                else:
                    ellipse_obj = self.isEllipse(id)[1]
                    center =  ellipse_obj.Plane.Origin
                    short_pt = rs.EvaluateCurve(id, rs.CurveClosestPoint(id, center))
                    start_pt = rs.CurveStartPoint(id)
                    end_pt = rs.CurveEndPoint(id)
                    ellArc = id
                    
                normal = ellipse_obj.Plane.Normal
                long_pt = None
                radx = ellipse_obj.Radius1
                rady = ellipse_obj.Radius2
            
            # Computes points relative to frame position/reference point
            center = rs.PointSubtract(center, self.framePosition)
            short_pt = rs.PointSubtract(short_pt, self.framePosition)
            long_pt = rs.PointSubtract(long_pt, self.framePosition)
            pin = rs.PointSubtract(pin, self.framePosition)
            normal = normal.Subtract(self.framePosition)
        
            focus1, focus2, long_pt, aDist, bDist = self.fociCompute(center, normal, radx, rady, short_pt, long_pt)
            
            # Does not proceed with elliptical arc agnostic object creation if computation is incorrect
            # Checks if minor axis point location corresponds to minor distance wrt to center or vice versa with major axis point location
            # Checks against radiusX, radiusY obtained from Rhino Common ellipse approximation
            if round(rs.Distance(short_pt, center), coordinate.precision()) != round(bDist, coordinate.precision()) or round(rs.Distance (long_pt, center), coordinate.precision()) != round(aDist, coordinate.precision()):
                warnings.warn('Arc object cannot be converted to elliptical arc')
                return
            """
                nurbs_obj = deepcopy(ellipse_obj).ToNurbsCurve()
                # Handles elliptical arcs whose range does not include either major/minor axes points
                ellipse_obj = nurbs_obj.TryGetEllipse()[1]
                ellipse_temp = ellipse_obj.ToNurbsCurve()
                short_pt = nurbs_obj.PointAt(ellipse_temp.ClosestPoint(ellipse_obj.Plane.Origin)[1])
                rs.AddTextDot('shortTest', short_pt)
                pin_pt = nurbs_obj.PointAt(10)
                start = obj.PointAtStart
                end = obj.PointAtEnd
                radX = ellipse_obj.Radius1
                radY = ellipse_obj.Radius2
                focus1, focus2, long_pt, aDist, bDist = fociCompute(center, normal, radx, rady, short_pt, long_pt)
            """
            
            # Generates pin pt on elliptical arc
            if self.checkPoint(rs.EvaluateCurve(ellArc, rs.CurveClosestPoint(ellArc, short_pt)), short_pt) and self.checkPoint(rs.EvaluateCurve(ellArc, rs.CurveClosestPoint(ellArc, short_pt)), short_pt) and self.checkPoint(short_pt, start_pt) == False and self.checkPoint(short_pt, end_pt) == False:
                pin_pt = short_pt
            else:
                x_coord = random.uniform(rs.CurveClosestPoint(ellArc, start_pt), rs.CurveClosestPoint(ellArc, end_pt))
                pin_pt =  rs.EvaluateCurve(ellArc, x_coord)
                                    
            if '3' in self.classification:
                center_coords = (center.X, center.Y, center.Z)
                foc1_coords = (focus1.X, focus1.Y, focus1.Z)
                foc2_coords = (focus2.X, focus2.Y, focus2.Z)
                start_pt_coords = (start_pt.X, start_pt.Y, start_pt.Z)
                end_pt_coords = (end_pt.X, end_pt.Y, end_pt.Z)
                pin_pt_coords = (pin_pt.X, pin_pt.Y, pin_pt.Z)
                short_pt_coords = (short_pt.X, short_pt.Y, short_pt.Z)
                long_pt_coords = (long_pt.X, long_pt.Y, long_pt.Z)
                
            else:
                center_coords = (center.X, center.Y)
                foc1_coords = (focus1.X, focus1.Y)
                foc2_coords = (focus2.X, focus2.Y)
                start_pt_coords = (start_pt.X, start_pt.Y)
                end_pt_coords = (end_pt.X, end_pt.Y)
                pin_pt_coords = (pin_pt.X, pin_pt.Y)
                short_pt_coords = (short_pt.X, short_pt.Y)
                long_pt_coords = (long_pt.X, long_pt.Y)
                
            eArc = (center_coords, focus1, focus2, start_pt_coords, end_pt_coords, pin_pt_coords, short_pt_coords, long_pt_coords)
        
            #"""
            rs.AddTextDot('center', (center_coords[0], center_coords[1], center_coords[2]))
            rs.AddTextDot('f1', (focus1[0], focus1[1], focus1[2]))
            rs.AddTextDot('f2', (focus2[0], focus2[1], focus2[2]))
            rs.AddTextDot('start', (start_pt_coords[0], start_pt_coords[1], start_pt_coords[2]))
            rs.AddTextDot('end', (end_pt_coords[0], end_pt_coords[1], end_pt_coords[2]))
            rs.AddTextDot('pin', (pin_pt_coords[0], pin_pt_coords[1], pin_pt_coords[2]))
            rs.AddTextDot('short', short_pt)
            rs.AddTextDot('long', long_pt)
            #"""
            
            # Creates attributes dictionary for circle
            att = self.createAtt('eArc', ellArc, eArc, sortName)
            # Extracts predicates/directives from geometry
            self.createTags(att, 'eArc', ellArc, eArc, sortName)
            return (eArc, att)
                
        else:
            warnings.warn('Curve is not an elliptical arc or can be transformed into an ellipse')
        
    def bezCurve(self, id, sortName):
        start = rs.CurveStartPoint(id)
        end = rs.CurveEndPoint(id)
        mid = rs.CurvePoints(id)[1]

        # Computes points relative to frame position/reference point
        start = rs.PointSubtract(start, self.framePosition)
        end = rs.PointSubtract(end, self.framePosition)
        mid = rs.PointSubtract(mid, self.framePosition)
        
        if '3' in self.classification:
            start_coords =(start.X, start.Y, start.Z)
            end_coords =(end.X, end.Y, end.Z)
            mid_coords =(mid.X, mid.Y, mid.Z)
        
        else:
            start_coords =(start.X, start.Y)
            end_coords =(end.X, end.Y)
            mid_coords =(mid.X, mid.Y)
            
        
        bez = (start_coords, mid_coords, end_coords)
        # Creates attributes dictionary for circle
        att = self.createAtt('bez', id, bez, sortName)
        # Extracts predicates/directives from geometry
        self.createTags(att, 'bez', id, bez, sortName)

        return (bez, att)
    
    """
    =====================
    CALCULATION FUNCTIONS
    =====================
    These functions are used to calculate certain properties of the incoming geometry:
        
    1. isEllipse: checks if an arc is an ellipse (primarily used for arcs)
    2. checkArc: checks if a curve is an arc or a closed curve by checking its start and end points against each other
    3. checkEllipse: checks if a closed curve/arc is circular or elliptical
    """
    
    def dist2Pts(self, A, B):
        dist = sqrt((A.X - B.X)**2 + (A.Y - B.Y)**2 + (A.Z - B.Z)**2)
        return dist
    
    def isEllipse(self, crv_id, tol=0.01):
        crv = rs.coercecurve(crv_id)
        success, ellipse = crv.TryGetEllipse(tol)
        return success, ellipse
    
    def checkArc(self, crv_id):
        if rs.CurveStartPoint(crv_id) == rs.CurveEndPoint(crv_id):
            return False
            
        return True
    
    def checkPoint(self, ptA, ptB):
        try:
            ptA_coords = (round(ptA.X, coordinate.precision()), round(ptA.Y, coordinate.precision()), round(ptA.Z, coordinate.precision()))
        except:
            ptA_coords = (round(rs.PointCoordinates(ptA).X, coordinate.precision()), round(rs.PointCoordinates(ptA).Y, coordinate.precision()), round(rs.PointCoordinates(ptA).Z, coordinate.precision()))
            
        try:
            ptB_coords = (round(ptB.X, coordinate.precision()), round(ptB.Y, coordinate.precision()), round(ptB.Z, coordinate.precision()))
        except:
            ptB_coords = (round(rs.PointCoordinates(ptB).X, coordinate.precision()), round(rs.PointCoordinates(ptB).Y, coordinate.precision()), round(rs.PointCoordinates(ptB).Z, coordinate.precision()))
            
        if ptA_coords[0] == ptB_coords[0] and ptA_coords[1] == ptB_coords[1] and ptA_coords[2] == ptB_coords[2]:
            return True
            
        return False
    
    def checkEllipse(self, crv_id):
        if self.isEllipse(crv_id)[0] or type(self.isEllipse(crv_id)[1]) == r.Geometry.Ellipse or rs.IsEllipse(crv_id):
            if rs.IsEllipse(crv_id):
                quad_pts = rs.EllipseQuadPoints(crv_id)
                major = quad_pts[0] - quad_pts[1]
                minor = quad_pts[2] - quad_pts[3]
                dist1 = float(sqrt(major.X**2 + major.Y**2 + major.Z**2))
                dist2 = float(sqrt(minor.X**2 + minor.Y**2 + minor.Z**2))
                
            elif self.isEllipse(crv_id)[0] or type(self.isEllipse(crv_id)[1]) == r.Geometry.Ellipse:
                ell = self.isEllipse(crv_id)[1]
                dist1 = float(ell.Radius1)
                dist2 = float(ell.Radius2)

            if dist1 == dist2:
                return False
                
            return True
        
        else:
            warnings.warn('Curve is neither an ellipse or an elliptical arc')
            return False
            
    # Used to deterime which quadrant point lies on the major/minor axis   
    def determineAxes(self, quad_points):
        long_pt1 = quad_points[0]
        long_pt2 = quad_points[1]
        short_pt1 = quad_points[2]
        short_pt2 = quad_points[3]
        order = [long_pt2] # (x-axis, y-axis)
        if short_pt1.Y == short_pt2.Y:
            order.insert(0, short_pt1)
        elif short_pt1.X == short_pt2.X:
            order.append(short_pt1)
            
        return order
    
    # Used to compute ellipse foci and axes distances
    def fociCompute(self, center, normal, radx, rady, short_pt, long_pt = None):
        if radx > rady:
            cDist = sqrt(radx**2 - rady**2)
            aDist = radx
            bDist = rady
        else:
            cDist = sqrt(rady**2 - radx**2)
            aDist = rady
            bDist = radx

        # Computation of major axis quadrant point 'long_pt'
        # Necessary bits: sortPt, center, plane of ellipse
        short_pt_coords = r.Geometry.Vector3d(short_pt.X, short_pt.Y, short_pt.Z)
        shortVec = center - short_pt_coords
        shortVec = r.Geometry.Vector3d(shortVec.X, shortVec.Y, shortVec.Z)
        shortUnit = deepcopy(shortVec)
        shortUnit.Unitize()
        normUnit = deepcopy(normal)
        normUnit.Unitize()
        longUnit = r.Geometry.Vector3d.CrossProduct(shortUnit, normUnit)
        longUnit.Unitize()
        
        # Computation of foci and major axis quadrant point
        focus1 = center + r.Geometry.Vector3d.Multiply(longUnit, cDist)
        focus2 = center - r.Geometry.Vector3d.Multiply(longUnit, cDist)
        
        if long_pt == None:
            long_pt = center + r.Geometry.Vector3d.Multiply(longUnit, aDist)
            
        return focus1, focus2, long_pt, aDist, bDist
    
    """
    =======================================================
    ADDING DESCRIPTIONS, ATTRIBUTES, PREDICATES, DIRECTIVES
    =======================================================
    """

    # Adds external description agnostic objects to main agnostic object
    def addDescriptions(self):
        newDescs = {}

        if self.descriptions != None and self.descriptions != '':
            # Extracts descriptions and their corresponding description sort type
            descriptions = self.descriptions.split('|')

            for d in descriptions:
                d = d.split('@')
                contents = d[1].split(';')
                
                # Creates new entry in descriptions dictionary if it does not yet exist
                if d[0] not in newDescs:
                    newDescs[d[0]] = []
                
                # Adds new description to dictionary entry list
                for indiv in contents:
                    newDescs[d[0]].append(indiv)
            
            self.agnosticObject['descriptions'] = newDescs
        
        return
    
    # Extracts attribute data
    # Stores each attribute type under corresponding hash key
    # Text attributes (labels, descriptions, enums) are stored as lists
    def createAtt(self, geomType, id, geometry, sortName):
        att = {'tags':[]}
        attributeCheck = self.attributes[sortName]
        weightCheck = ['lineSegment', 'circle', 'ellipse', 'cArc', 'eArc', 'bez']

        if 'label' in attributeCheck:
            att['labels'] = []
            for name in rs.GetUserText(id):
                if '&' in name:
                    att['labels'].append(rs.GetUserText(id, name))
                    
                    
            if (len(att['labels'])) == 0:
                if rs.IsTextDot(id):
                    dotText = rs.TextDotText(id)
                    if 'l(' not in dotText and 'd(' not in dotText and '#(' not in dotText:
                        att['labels'].append(dotText)
                else:
                    del att['labels']
        
        if 'desc' in attributeCheck:
            att['label_descriptions'] = []
            for key_name in rs.GetUserText(id):
                if '#' in key_name:
                    text =  rs.GetUserText(id, key_name)
                    if text != '""':
                        att['label_descriptions'].append(text)
                    
            if len(att['label_descriptions']) == 0:
                del att['label_descriptions']

        if 'enum' in attributeCheck:
            text = rs.GetUserText(id, "Enum")    
            if text != '':
                att['enums'] = [text]
            
        # Weight - black (0,0,0) to white (255, 255, 255)
        if 'weight' in attributeCheck:        
            if geomType not in weightCheck:
                weight = rs.coercecolor(rs.ObjectColor(id))
                if weight.R != 0 and weight.G != 0 and weight.B != 0:
                    rgb = [int(weight.R), int(weight.G), int(weight.B)]
                    att['weight255'] = rgb
                    
            elif geomType in weightCheck:
                # Obtains line segment print width
                # If found to be 0, then given default 0.01 weight
                if rs.ObjectPrintWidth(id) == 0:
                    weight = 0.01
                else:
                    weight = rs.ObjectPrintWidth(id)
                    weight = weight
                att ['weight'] = weight

        # Weight255 - black to white in grayscale
        if 'weight255' in attributeCheck:
            weight255 = rs.coercecolor(rs.ObjectColor(id))  
            rgb = [int(weight255.R), int(weight255.G), int(weight255.B)]
            if rgb[0] != rgb[1] or rgb[1] != rgb[2] or rgb[0] != rgb[2]:
                att['weight255'] = rgb
            
        # Color - normal rgb 0 to 255   
        if 'color' in attributeCheck:
            color = rs.coercecolor(rs.ObjectColor(id))  
            rgb = [int(color.R), int(color.G), int(color.B)]
            if rgb[0] != rgb[1] or rgb[1] != rgb[2] or rgb[0] != rgb[2]:
                att['color'] = rgb
                
        return att
    
    # General function for extracting predicates/directives and their necessary data from geometry
    def createTags(self, att, geomType, id, geometry, sortName):
        predicateNames = None
        tags = ['pointOL', 'length', 'distance', 'normal']
        if 'tag' in rs.GetUserText(id):
            if geomType != 'polyline':
                tagName = geomType + '-' + str(rs.GetUserText(id, 'tag'))
            else:
                tagName = geomType + '-' + str(rs.GetUserText(id, 'tag'))
            classification = sort.register[sortName].baseSort().characteristic.type
            
            if tagName not in att['tags']:
                att['tags'].append(tagName)

        # Points/Planes/Line Segments
        if 'tagText' in rs.GetUserText(id):
            predicateNames = rs.GetUserText(id, 'tagText')
            predicateNames = predicateNames.split('|')
            count = rs.GetUserText(id, 'tag')

            for name in predicateNames:
                # Extracts void information
                if 'void' in name:
                    setName = 'void'
                        
                    if name not in att['tags']:
                        att['tags'].append(name)

                    voidName = name.split('_')
                    groupName = voidName[0] + '_' + voidName[1]
                    
                    # Creates void group inside void dictionary in Predicates
                    if groupName not in self.predicates[setName]:
                        self.predicates[setName][groupName] = ()
                    if len(voidName) > 2:
                        if groupName not in self.voidSorts:
                            self.voidSorts[groupName] = ()
                            sortTypes = voidName[2:]
                            sortList = ()
                            for sortName in sortTypes:
                                if sortName.isdigit() == False:
                                    self.voidSorts[groupName] += (sortName,)
                        
                    if geomType == 'point':
                        self.predicates[setName][groupName] += (geometry,)
                    
                    elif geomType == 'planeSegment':
                        if type(geometry) == type({}):
                            vectors = geometry['outer']
                        
                            for vec in vectors: 
                                self.predicates[setName][groupName] = self.predicates[setName][groupName] +  (vec,)
                                
                        elif type(geometry) == type([]):
                            for vec in geometry:
                                self.predicates[setName][groupName] = self.predicates[setName][groupName] + (vec,)
                    
                    elif geomType == 'lineSegment':
                        if groupName not in self.predicates[setName]:
                            self.predicates[setName][groupName] = ()
                        
                        if len(geometry) == 2:
                            self.predicates[setName][groupName] = self.predicates[setName][groupName] + (geometry[0], geometry[1])
                        
                    elif geomType == 'polyline':
                        if len(geometry) == 2:
                            self.predicates[setName][groupName] = self.predicates[setName][groupName] + (geometry[0], geometry[1])
                        else:
                            voidList = ()
                            
                            for vec in geometry:
                                voidList = voidList + (vec,)
    
                            self.predicates[setName][groupName] = self.predicates[setName][groupName] + voidList
                # Extracts distance information
                elif 'distance' in name:
                    groupName = name.split('_')
                    groupName = groupName[0].split('^')[-1]

                    if 'distance' in name:
                        keyName = 'distance'
                    
                    # Gets key name: either 'ang' or 'dist'
                    retrieve = self.retrieve[keyName]
                    try:
                        val1 = float(rs.GetUserText(id, (retrieve + '1' + '_' + groupName)))
                    except:
                        val1 = rs.GetUserText(id, (retrieve + '1' + '_' + groupName))
                    
                    try:
                        val2 = float(rs.GetUserText(id, (retrieve + '2' + '_' + groupName)))
                    except:
                        val2 = rs.GetUserText(id, retrieve + '2' + '_' + groupName)
                    
                    # Retrieves tag name of RHS individual tag 
                    if keyName == 'distance':
                        otherTag = rs.GetUserText(id, 'pair' + '_' + retrieve + '_' + groupName)
                        otherTag = otherTag.split('-')
                        
                        # Adds in descriptor of to-be-generated geometry tag type
                        if otherTag[0] == '1' or otherTag[0] == '8192':
                            otherTag = 'point' + '-' + otherTag[1]
                        elif otherTag[0] == '4':
                            otherTag = 'lineSegment' + '-' + otherTag[1]
                        elif otherTag[0] == '8' or otherTag[0] == '16' or otherTag[0] == '32':
                            otherTag = 'planeSegment' + '-' + otherTag[1]

                        if groupName not in self.directives[keyName]:
                            if val1 == val2 and val1 != '':
                                self.directives[keyName][str(groupName)] = ['@' + otherTag, '@' + tagName, val1]
                            else:
                                self.directives[keyName][str(groupName)] = ['@' + otherTag, '@' + tagName, val1, val2]
                
                # Extracts bound line information
                elif 'bound' in name:
                    if name not in self.predicates['bound']:
                        self.predicates['bound'][name] = []
                    self.predicates['bound'][name].append('@'+tagName)
                
                # Extracts pointOL/length/normal information
                elif 'pointOL' in name or 'normal' in name:
                    tempSet = []
                    val2 = None
                    groupName = name.split('_')
                    group = groupName[0].split('^')[-1]

                    if 'pointOL' in name or 'length' in name:
                        if 'pointOL' in name:
                            keyName = 'pointOL'
                        elif 'length' in name:
                            keyName = 'length'

                        try:
                            val1 = float(groupName[1])
                        except:
                            val1 = groupName[1]

                        if len(groupName) == 3:
                            try:
                                val2 = float(groupName[2])
                            except:
                                val2 = groupName[2]

                        if val1 == val2 and val1 != '':
                            self.directives[keyName].append([[val1], '@' + tagName])
                        else:
                            self.directives[keyName].append([[val1, val2], '@' + tagName])
                            
                    elif 'normal' in name:
                        keyName = 'normal' 
                        setName = self.retrieve[keyName]
                        # Extracts intended tag name of RHS new normal line
                        rhsTag = rs.GetUserText(id, setName + '_' + 'rhs_' + group)
                        tempSet.append('@' + rhsTag)
                        
                        # Extracts minimum length (and maximum length, if available) from id)
                        lengthMin = rs.GetUserText(id, setName + '1_' + group)
                        lengthMax = rs.GetUserText(id, setName + '2_' + group)
                        
                        # Extracts coordinates of starting point of geometry generation for normal line
                        target = rs.GetUserText(id, 'norm_end_' + group)
                        # Presence of ',' indicates a coordinate set
                        if ',' in target:
                            target = target.split(',')
                        # Otherwise, it contains the tag name of the plane segment the new normal line will be perpendicular to
                        else:
                            target = '@' + target
                        
                        if type(target) == type([]):
                            for coord in target:
                                try:
                                    tempSet.append(float(coord))
                                except:
                                    tempSet.append(coord)
                        elif type(target) == type('a') and target != rhsTag:
                            tempSet.insert(0, target)
                        
                        try:
                            tempSet.append(float(lengthMin))
                        except:
                            tempSet.append(lengthMin)
                        if lengthMin != lengthMax:
                            try:
                                tempSet.append(float(lengthMax))
                            except:
                                tempSet.append(lengthMax)
                        
                        tempSet.insert(0, '@' + tagName)
                        self.directives[keyName][group] = tempSet
                    
                # For maxline/shortest/nolabel
                elif 'max' in name or 'shortest' in name or 'nolabel' in name or 'longest' in name:
                    self.predicates[name].append('@' + tagName)   
                
                elif 'embeds' in name:
                    groupName = name.split('_')
                    group = groupName[0].split('^')[-1]
                    
                    # Creates dictionary element
                    if group not in self.predicates['embeds']:
                        self.predicates['embeds'][group] = []
                    
                    # Determines embedding pairs and their order
                    if len(self.predicates['embeds'][group]) == 0:
                        self.predicates['embeds'][group].append(tagName)
                    elif len(self.predicates['embeds'][group]) == 1:
                        if 'planeSegment' in self.predicates['embeds'][group][0]:
                            if 'lineSegment' in tagName or 'point' in tagName:
                                self.predicates['embeds'][group].append(tagName)
                                
                        elif 'lineSegment' in self.predicates['embeds'][group][0]:
                            if 'planeSegment' in tagName:
                                self.predicates['embeds'][group].insert(0, tagName)
                            elif 'point' in tagName:
                                self.predicates['embeds'][group].append(tagName)
                            
        return
    
    """
    ========================================
    CLEANING DIRECTIVE TAGS FOR OPTIMIZATION
    ========================================
    """
    
    # Clean-up of Point on Line / Length list to clear any redundancies
    # All similar numerical ratio - line tags are placed in the same list
    def cleanTags(self, tagDict, tagName):
        
        if len(tagDict[tagName]) > 0:
            tempGeom = deepcopy(tagDict[tagName])
            cleanGeom = []
            
            for a, geom in enumerate(tempGeom):
                if type(geom) == type([]):
                    first = geom[0][0]
                    tempSet = [geom[1]]
                    second = None
                    
                    if len(geom[0]) == 2:
                        second = geom[0][1]
                        
                    for i in tempGeom[a:]:
                        if first == i[0][0]:
                            if len(i[0]) == 1:
                                if i[1] not in tempSet:
                                    tempSet.append(i[1])
                                    tempGeom.pop(tempGeom.index(i))
                            elif len(i[0]) == 2:
                                if second == i[0][1] and i[1] not in tempSet:
                                    tempSet.append(i[1])
                                    tempGeom.pop(tempGeom.index(i))

                    tempSet.append(first)
                    if second != None:
                        tempSet.append(second)
                    
                    cleanGeom.append(tempSet)

            tagDict[tagName] = cleanGeom
        return
        
