#-*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

# RHINO SPECIFIC IMPORTS | FROM PYTHON 3.5 to 2.7
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

import os

import timeit

# Rhino Scripting
import rhinoscriptsyntax as rs
import Rhino as r
import scriptcontext as sc

# Python built-in imports
from builtins import range
import math
import itertools
import warnings

# Sortal Library Imports
from sortal import (rule, sort, form, flow, sdlParser, coordinate)
                    
# Import for transformations
from mpmath import pi, sin, cos
# Import for duplications
from copy import deepcopy

"""
========================
RHINO SPECIFIC FUNCTIONS
========================
The following functions are meant to be publicly available to the user:
    1	add_bound_line:	Adds bound line predicate tag to line segment/polyline; indicated as 'bound' within predicates text ('#( )')
    2	add_color:	Adds color to Rhino Object (one or several objects may be selected) using RGB values; this color is not visible in Print Preview Mode
    3	add_description:	Adds description to Rhino Object's user text, visualized as text dot with extra 'd( )' enclosing text in visual text dot assigned to geometry
    4	add_distance:	Adds distance directive tag to geometries; first selected geometry must be found in LHS, second selected geometry must be found in RHS; 
        range of values for length is inputted to function call; directive is indicated as 'distance' within directives text ('#( )') in visual text dot assigned to geometry
    5	add_embeds:	Adds embed predicate tag to geometries; first selected geometry must be the larger geometry (either a plane segment or a straight line segment) 
        and the second selected geometry  must be the smaller geometry that will be contained inside the first selected geometry; indicated as 'embeds' within predicates text ('#( )') in visual text dot assigned to geometry
    6	add_enum:	Adds enumerative value to Rhino Object's user text, not visualized
    7	add_label:	Adds label data to Rhino Object's user text, visualized as text dot with extra 'l( )' enclosing text in visual text dot assigned to geometry
    8	add_longest_line:	Adds longest line predicate tag to line segment/polyline; indicated as 'longest' within predicates text ('#( )')  in visual text dot 
        assigned to geometry
    9	add_max_line:	Adds maximum line predicate tag to line segment/polyline; indicated as 'max' within predicates text ('#( )') in visual text dot assigned to geometry
    10	add_no_label:	Adds no-label predicate tag to geometry to constrain matching to not-labelled geometries; indicated as 'no_label' 
        within predicates text ('#( )')  in visual text dot assigned to geometry
    11	add_normal:	Adds normal directive tag to line segments (if 2D), plane segments (if 3D); indicated as 'distance' within directives text ('#( )')  
        in visual text dot assigned to geometry
    12	add_point_on_line:	Adds point-on-line directive tag to line segment/polyline; range of values for point placement is inputted to function call; 
        indicated as 'point_on_line' within directives text ('#( )') followed by coordinates of point in visual text dot assigned to geometry
    13	add_shortest_line:	Adds shortest line predicate tag to line segment/polyline; indicated as 'short' within predicates text ('#( )') in visual text dot 
        assigned to geometry in visual text dot assigned to geometry
    14	add_void:	Adds void predicate tag to collection of points/text dots/line segments/plane segments that form a closed polygon; indicated as 'void' 
        within predicates text ('#( )'), and indicates which void group it belongs to in visual text dot assigned to geometry in visual text dot assigned to geometry
    15	add_weight:	Adds width (weight) to line or grayscale (weight255) to point or text dot in visual text dot assigned to geometry in visual text dot 
        assigned to geometry
    16	analysis:	Analyses a list of Rhino geometry, and moves points/extends lines to intersect directly with each other (for lines to intersect with each other, 
        for points to line on lines) in visual text dot assigned to geometry in visual text dot assigned to geometry
    17	attach_label:	Attaches text dot to a particular line segment; this stores the text dot's GUID in the line segment's user text and moves the text dot 
        in visual text dot assigned to geometry
    18	clear_shape:	Clears a list of Rhino objects; takes as input a list/dictionary of Rhino GUIDs OR single Rhino GUID
    19	delete_description:	Removes description data from Rhino Object, if any are present, and in visual text dot assigned to geometry
    20	delete_label:	Removes label from Rhino Object's user text and in visual text dot assigned to geometry; deletes text dot 'attached' to line segment, 
        or reverts text dot to a point
    21	delete_label:	Removes label data from Rhino Object, if any are present, and in visual text dot assigned to geometry
    22	delete_pred_dir:	Removes predicates/directives from Rhino Object, if any are present, and in visual text dot assigned to geometry
    23	delete_tag:	Removes geometry predicate/directive tag from Rhino Object, if any are present, and in visual text dot assigned to geometry
    24	delete:	Uses Rhino's DeleteObjects to delete all Rhino objects in window, clears sort and rule registers, sets up new grammar frames
    25	tag: 	Adds or changes the tag data stored in a Rhino geometry and changes the visual text dot 'attached' to the geometry accordingly
   
The other functions listed here are meant as support functions for some of the publicly available functions above:
    1	addPredDir:	General method that handles all repetitive tasks in adding predicates/directives
    2	addText:	General method that handles all repetitive tasks in add_label and add_description, and subsequent text dot redraw
    3	deletion: 	General method that handles all repetitive tasks in delete_label, delete_description, delete_tag; 
        it handles label/description/predicate/directive deletion and the subsequent text dot redraw
    4	ptDistance:	Determines distance based on x, y, z differences

"""

# Searches for original Rhino geometry of GUID
rhdoc = r.RhinoDoc.ActiveDoc

# Dictionary of tag retrieval equivalents for each predicate/directive type
tags = {'pointOL':'seg', 'length':'len', 'angle':'ang', 'distance':'dist', 'normal': 'norm'}

"""
=================
SUPPORT FUNCTIONS
=================

1	addPredDir:	General method that handles all repetitive tasks in adding predicates/directives

2	addText:	General method that handles all repetitive tasks in add_label and add_description, and subsequent text dot redraw

3	deletion: 	General method that handles all repetitive tasks in delete_label, delete_description, delete_tag; 
    it handles label/description/predicate/directive deletion and the subsequent text dot redraw
    
4	ptDistance:	Determines distance based on x, y, z differences
"""
    
# General method for adding tags and predicate information to line segments
# For normal, seg1 = direction vector coordinates, seg2 = length (single value that could be a float/integer/text description, or list)
def addPredDir(tag, tagSelf, seg1 = None, seg2 = None, ends = None):
    if tag == 'pointOL':
        message = 'Select line/s to tag with point on line directive'
    elif tag == 'length':
        message = 'Select line/s to tag with length directive'
    elif tag == 'distance':
        message = 'Select two geometries to serve as reference geometry for distance tag.'
    elif tag == 'normal':
        message = 'Select line/s (2D) or plane segments (3D) to tag with normal directive. Please note that when creating the sortal shape later on, there should be a line segment included in the total geometry as basis for the new normal line, and that the a line with the same tag name as the RHS tag should be present'
    elif tag == 'nolabel':
        message = 'Select geometry/ies to tag with no-label predicate'
    elif tag == 'embeds': 
        message = 'Select geometry/ies to tag with embeds predicate. The first selected geometry must be the container geometry (i.e. the bigger object), and the second selected geometry must be the embedded geometry (i.e. the smaller object'
    else:
        message = 'Select line/s to tag as ' + tag +' line'
        
    ids = rs.GetObjects(message)  
    specTags = ['distance', 'normal', 'pointOL', 'max', 'shortest', 'longest', 'nolabel', 'embeds']
    index = 0
    obj_type = ''
    
    if len(ids) == 0:
        warnings.warn('No geometries selected')
        return False
        
    # Bound line
    if tag == 'bound':
        if type(ends) == type(1):
            if ends == 0 or ends == 1 or ends == 2:
                if ends == 0:
                    newTag = '>bound'
                elif ends == 1:
                    newTag = 'bound<'
                elif ends == 2:
                    newTag = '>bound<'    
            else:
                raise TypeError('Input for type of bound line is not a integer (only 0, 1 and 2 are accepted')
                
        # Goes through inputs to check if all inputs are straight line segments, not closed
        for lineSeg in ids:
            if rs.ObjectType(lineSeg) != 4:
                raise TypeError('Input contains an object that is not a line segment. This cannot contain a bound line directive.')
            if rs.IsCurveClosed(lineSeg) or rs.IsCircle(lineSeg) or rs.IsArc(lineSeg) or rs.IsEllipse(lineSeg):
                raise TypeError('Input contains a curved line segment. This cannot contain a bound line directive.')
                try:
                    crv = rs.coercecurve(object)
                    if type(crv) == r.Geometry.LineCurve:
                        pass
                    elif type(crv) == r.Geometry.PolylineCurve:
                        if crv.PointCount != 2:
                             raise TypeError('Input contains a polyline curve. This cannot contain a ' +  message + '. It must be broken into individual pieces first.')
                    elif type(crv) == r.Geometry.PolyCurve:
                        if crv.SegmentCount > 1: 
                            raise TypeError('Input contains a polycurve. This cannot contain a ' + message + '. It must borken into individual pieces first.')
                    elif crv.HasBezierSpans:
                        raise TypeError('Input contains a curved line segment. This cannot contain a ' +  message)
                except:
                    raise TypeError('Input contains an object that is not a line segment. This cannot contain a ' +  message)
                
    # Point on line / Distance / Normal
    elif tag in specTags:
        count = str(deepcopy(sc.sticky['groupCount']))
        sc.sticky['groupCount'] += 1
        if 'distance' in tag:
            if len(ids) < 2:
                print ('You have selected less than 2 geometries. Exiting function.')
                return
            elif len(ids) > 2:
                print ('You have selected more than 2 geometries. Exiting function.')
                return
            
            # Goes through inputs to check if all inputs are either straight line segments or points
            for object in ids:
                if rs.ObjectType(object) != 4 and rs.ObjectType(object) != 1 and rs.ObjectType(object) != 8192:
                    raise TypeError('Input contains an object that is not a line segment or a point. This cannot contain a distance directive.')
                
                if rs.ObjectType(object) == 1 or rs.ObjectType(object) == 8192:
                    if 'visual' in rs.GetUserText(object):
                        raise ValueError('Input contains a visual text dot, which is only used to contain information for other geometries and will not be registered by Rhino GUIDs to sortal individual conversion.')
                if rs.ObjectType(object) == 4:
                    if rs.IsCurveClosed(object) or rs.IsCircle(object) or rs.IsArc(object) or rs.IsEllipse(object):
                        raise TypeError('Input contains a closed curve. This cannot contain a distance directive.')
                        
                        try:
                            crv = rs.coercecurve(object)
                            if type(crv) == r.Geometry.LineCurve:
                                pass
                            elif type(crv) == r.Geometry.PolylineCurve:
                                if crv.PointCount != 2:
                                     raise TypeError('Input contains a polyline curve. This cannot contain a ' +  message + '. It must be broken into individual pieces first.')
                            elif type(crv) == r.Geometry.PolyCurve:
                                if crv.SegmentCount > 1: 
                                    raise TypeError('Input contains a polycurve. This cannot contain a ' + message + '. It must broken into individual pieces first.')
                            elif crv.HasBezierSpans:
                                raise TypeError('Input contains a curved line segment. This cannot contain a ' +  message)
                        except:
                            raise TypeError('Input contains an object that is not a line segment. This cannot contain a ' +  message)
                    
        elif tag in specTags[2:6]:
            if tag == specTags[2]:
                message = 'point on line directive.'
            elif tag == specTags[3]:
                message = 'max line predicate.'
            elif tag == specTags[4]:
                message = 'shortest line predicate.'
            elif tag == specTags[5]:
                message = 'longest line predicate.'
                
            # Goes through inputs to check if all inputs are straight line segments, not closed
            for lineSeg in ids:
                if rs.ObjectType(lineSeg) != 4:
                    raise TypeError('Input contains an object that is not a line segment. This cannot contain a ' +  message)
                if rs.IsCurveClosed(lineSeg) or rs.IsCircle(lineSeg) or rs.IsArc(lineSeg) or rs.IsEllipse(lineSeg):
                    raise TypeError('Input contains a curved line segment. This cannot contain a ' +  message)
                    
                try:
                    crv = rs.coercecurve(lineSeg)
                    if type(crv) == r.Geometry.LineCurve:
                        pass
                    elif type(crv) == r.Geometry.PolylineCurve:
                        if crv.PointCount != 2:
                             raise TypeError('Input contains a polyline curve. This cannot contain a ' +  message + '. It must be broken into individual pieces first.')
                    elif type(crv) == r.Geometry.PolyCurve:
                        if crv.SegmentCount > 1: 
                            raise TypeError('Input contains a polycurve. This cannot contain a ' + message + '. It must borken into individual pieces first.')
                    elif crv.HasBezierSpans:
                        raise TypeError('Input contains a curved line segment. This cannot contain a ' +  message)
                except:
                    raise TypeError('Input contains an object that is not a line segment. This cannot contain a ' +  message)
                    
        elif 'normal' in tag :
            for object in ids:
                # Only rhsTag is necessary, creating a normal to a line segment
                if seg2 != None:
                    if rs.ObjectType(object) != 4:
                        raise TypeError('Input contains an object that is not a line segment. This cannot contain a normal directive.')
                        
                    if rs.ObjectType(object) == 4:
                        if rs.IsCurveClosed(object) or rs.IsCircle(object) or rs.IsArc(object) or rs.IsEllipse(object):
                            raise TypeError('Input contains a curved line segment. This cannot contain a normal directive.')
                        
                        try:
                            crv = rs.coercecurve(object)
                            if type(crv) == r.Geometry.LineCurve:
                                pass
                            elif type(crv) == r.Geometry.PolylineCurve:
                                if crv.PointCount != 2:
                                     raise TypeError('Input contains a polyline curve. This cannot contain a normal directive, as it must be broken into individual pieces first.')
                            elif type(crv) == r.Geometry.PolyCurve:
                                if crv.SegmentCount > 1: 
                                    raise TypeError('Input contains a polycurve. This cannot contain a normal directive, as it must be borken into individual pieces first.')
                            elif crv.HasBezierSpans:
                                raise TypeError('Input contains a curved line segment. This cannot contain a normal directive.')
                        except:
                            raise TypeError('Input contains an object that is not a line segment. This cannot contain a normal directive.')
                            
                # Both rhsTag and planeSegmentTag are necessary, creating a normal to a plane segment
                elif seg2 == None:
                    if rs.ObjectType(object) != 8 and rs.ObjectType(object) != 16 and rs.ObjectType(object) != 32:
                        raise TypeError('Input contains an object that is not a surface, BREP or mesh. This cannot contain a normal directive.')
        # Embeds
        elif 'embeds' in tag:
            if len(ids) > 2:
                print ('You have selected more than two (2) geometries. Only two (2) geometries may be selected for the embed tag. Exiting function.')
                return
            elif len(ids) < 2:
                print ('You have selected less than two (2) geometries. Exactly two (2) geometries must be selected for the embed tag. Exiting function.')
                return
                
            # Goes through inputs to check if all inputs exhibit the correct relationships
            # i.e. [planeSegment, lineSegment or point] OR [lineSegment, point]
            bigObject = ids[0]
            smallObject = ids[1]
            
            if rs.ObjectType(bigObject) != 8 and rs.ObjectType(bigObject) != 16 and rs.ObjectType(bigObject) != 32 and rs.ObjectType(bigObject) != 4:
                raise TypeError('Input for container object of embeds predicate is neither a surface nor a straight line segment. This cannot be the container object of the embeds predicate.')
            
            if rs.ObjectType(smallObject) != 4 and rs.ObjectType(smallObject) != 1 and rs.ObjectType(smallObject) != 8192:
                raise TypeError('Input for container object of embeds predicate is neither a line segment, point or text dot. This cannot the embedded object of the embeds predicate.')
        
        newTag = tag 
        
        if newTag in specTags[0:2]:
            newTag += '^' + count

        # Point on line / Distance
        if 'normal' not in newTag:
            if type(seg1) == type(0) or type(seg1) == type(0.0):
                seg1name = str(seg1)
            else:
                seg1name = seg1
                
            if type(seg2) == type(0) or type(seg2) == type(0.0):
                seg2name = str(seg2)
            else:
                seg2name = seg2
        else:
            # seg1 = length/s
            # seg2 = coordinates (This may have a list of coordinates or None) 
            # end = tags (either [rhsTag, originTag] for plane segments or rhsTag only for line segments)
            if type(seg1) == type(0.0) or type(seg1) == type(0) or type(seg1) == type('a'):
                seg1name = str(seg1)
                seg2name = str(seg1)
            elif type(seg1) == type([]):
                seg1name = str(seg1[0])
                seg2name = str(seg1[1])
            
            # Tag name of new normal line
            # If coordinates are given (seg2 != None), then the assumption is a 2D normal line
            # If no coordinates are given (seg2 == None), then the assumption is a 3D normal line
            if seg2 != None:
                newTag += '_' + ends
            elif seg2 == None:
                newTag += '_' + ends + '_' + 'planeSegment-'

            # Coordinates of the direction vector the new normal line will be perpendicular to
            if seg2 != None:
                if type(seg2) == type('a'):
                    newTag += '_' + seg2
                # Line segment, direction vector as target
                elif type(seg2) == type(()) or type(seg2) == type([]):
                    newTag += '_' + str(seg2[0]) + ',' + str(seg2[1])

        if seg1name != None:
            # Length/distance/normal length/point on line segmentation values for both minimum and maximum values are the same
            if seg1name == seg2name:
                newTag += '_' + seg1name
            # Length/distance/normal length/point on line segmentation values for minimum and maximum values are different
            else:
                # If any of the values are strings that are above twenty characters, then the values are not displayed
                # They are still stored in the UserText of the object
                if tag != 'pointOL':
                    include_length = True
                    if type(seg1name) == type('a'):
                        if len(seg1name) > 20:
                            include_length = False
                    if type(seg2name) == type('a'):
                        if len(seg1name) > 20:
                            include_length = False
                    
                    if include_length:
                        newTag += '_' + seg1name + '_' + seg2name
                else:
                    newTag += '_' + seg1name + '_' + seg2name

    for id in ids: 
        if rs.IsCurve(id) or rs.IsPoint(id) or rs.IsSurface(id) or rs.IsMesh(id) or rs.IsBrep(id) or rs.IsTextDot(id):
            if 'bound' in rs.GetUserText(id, 'tagText') and 'bound' in newTag and tag == 'bound':
                newTag = '>bound<' 
            
            if 'tag' not in rs.GetUserText(id): 
                if tagSelf:
                    origColor = rs.coercecolor(rs.ObjectColor(id))
                    rs.ObjectColor(id, [255, 225, 0])
                    tagName = raw_input('Please enter tag name for the highlighted geometry')
                    rs.SetUserText(id, 'tag', tagName)
                    rs.ObjectColor(id, origColor)
                    
                if tagSelf == False or tagName == '':
                    rs.SetUserText(id, 'tag', str(sc.sticky['keyCount']))
                    sc.sticky['keyCount'] += 1
                    
            # Obtains tag name for plane segment from plane segment that normal directive will be added to
            if tag == 'normal' and seg2 == None:
                if rs.IsSurface(id) or rs.IsMesh(id) or rs.IsBrep(id):
                    if 'planeSegment-' in newTag:   
                        newTag = newTag.split('_')
                        index = newTag.index('planeSegment-')
                        newTag[index] = 'planeSegment-' + rs.GetUserText(id, 'tag')
                        tempTag = newTag[0]
                        
                        for tagBit in newTag[1:]:
                            tempTag += '_' + tagBit
                            
                        newTag = tempTag
               
            if newTag in rs.GetUserText(id, 'tagText'):
                print ('The ' + newTag +' tag already exists in the Rhino object')
                continue
                
            lText0 = ''
            ldText0 = ''
            ptText0 = ''
            fullText = ''

            if 'LabelDot' in rs.GetUserText(id):
                try:    
                    rs.DeleteObject(rs.GetUserText(id, "LabelDot"))
                    textDot = rs.GetUserText(id, "LabelDot")
                except:
                    textDot = None
            
            # Collects pre-existing predicate information
            if 'lText' in rs.GetUserText(id) or 'ldText' in rs.GetUserText(id) or 'tagText' in rs.GetUserText(id):
                if 'lText' in rs.GetUserText(id):
                    lText0 = rs.GetUserText(id, "lText")
                    
                if 'ldText' in rs.GetUserText(id):
                    ldText0 = rs.GetUserText(id, "ldText")
                    
                if 'tagText' in rs.GetUserText(id):
                    ptText0 = rs.GetUserText(id, "tagText")

            if tag == 'pointOL' or tag == 'normal' or tag == 'length':
                if tag == 'pointOL':
                    setName = tags['pointOL']
                elif tag == 'length':
                    setName = tags['pointOL']
                
                # seg1 = length/s
                # seg2 = coordinates (This may have a list of coordinates or None) 
                # end = tags (either [originTag, rhsTag] for plane segments or rhsTag only for line segments)
                elif tag == 'normal':
                    setName = tags['normal']
                    # Length, followed by targetCoords/tag name of new normal line
                    # Indicates either the direction vector that the new normal line will be perpendicular to, or the tag name of the plane segment
                    # that the new normal line will be perpedicular to
                    
                    if rs.ObjectType(id) == 4 and seg2 != None:
                        rs.SetUserText(id, setName + '_end_' + count, (str(seg2[0]) + ',' + str(seg2[1])))
                        rs.SetUserText(id, setName + '_rhs_' + count, ends)
                    elif rs.ObjectType(id) == 8 or rs.ObjectType(id) == 16 or rs.ObjectType(id) == 32:
                        rs.SetUserText(id, setName + '_end_' + count, 'planeSegment-' +rs.GetUserText(id, 'tag'))
                        rs.SetUserText(id, setName + '_rhs_' + count, ends)  

                    # Minimum length
                    rs.SetUserText(id, setName + '1_' + count, seg1name)
                    # Maximum length
                    rs.SetUserText(id, setName + '2_' + count, seg2name)

            elif tag == 'distance':
                if tag == 'distance':
                    setName = tags['distance']

                rs.SetUserText(id, setName + '1_' + count, seg1name)
                rs.SetUserText(id, setName + '2_' + count, seg2name)
                
                # Stores information about LHS tag name in RHS tag name and vice versa
                if ids.index(id) == 1:
                    rs.SetUserText(id, 'pair_' + setName + '_' + count, str(rs.ObjectType(ids[0])) + '-' + rs.GetUserText(ids[0], 'tag'))
                    rs.SetUserText(ids[0], 'pair_' + setName + '_' + count,  str(rs.ObjectType(id)) + '-' + rs.GetUserText(id, 'tag'))

                    if rs.IsCurve(id):
                        geomName = 'lineSegment-'
                    elif rs.IsPoint(id) or rs.IsTextDot(id):
                        geomName = 'point-'
                    elif rs.IsSurface(id) or rs.IsMesh(id) or rs.IsBrep(id):
                        geomName = 'planeSegment-'
                        
                    if (geomName + rs.GetUserText(ids[1], 'tag')) in newTag:
                        newTagText = newTag.split('_')
                        index = newTagText.index(geomName + rs.GetUserText(ids[1], 'tag'))
                        newTagText[index] = geomName + rs.GetUserText(ids[0], 'tag')
                        tempText = newTagText[0]
                        
                        for text in newTagText[1:]:
                            tempText += '_' + text
                            
                        newTag = tempText
                        
                elif ids.index(id) == 0:
                    if rs.IsCurve(id):
                        geomName = 'lineSegment-'
                    elif rs.IsPoint(id) or rs.IsTextDot(id):
                        geomName = 'point-'
                    elif rs.IsSurface(id) or rs.IsMesh(id) or rs.IsBrep(id):
                        geomName = 'planeSegment-'
                        
                    newTag += '_' + geomName + rs.GetUserText(ids[1], 'tag')
            
            elif tag == 'embeds':
                if ids.index(id) == 0:
                    if 'tag' not in rs.GetUserText(ids[1]):
                        origColor = rs.coercecolor(rs.ObjectColor(ids[1]))
                        rs.ObjectColor(ids[1], [255,225,0])
                        tagName = raw_input('Please enter tag name for the highlighted geometry')
                        rs.SetUserText(ids[1], 'tag', tagName)
                        rs.ObjectColor(id, origColor)
                    else:
                        tagName = rs.GetUserText(ids[1], 'tag')
                    
                    # Sets pair tag data inside geometry
                    # Container geometry
                    rs.SetUserText(id, 'embeds_' + count,  str(rs.ObjectType(ids[1])) + '-' + tagName)
                    targetGeom = ids[1]
                    
                elif ids.index(id) == 1:
                    tagName = rs.GetUserText(ids[0], 'tag')                     
                    # Sets pair tag data inside geometry
                    # Embedded geometry
                    rs.SetUserText(id, 'embeds_' + count,  str(rs.ObjectType(ids[0])) + '-' + tagName)
                    targetGeom = ids[0]
                
                # Determines geometry type of target geometry
                if rs.IsCurve(targetGeom):
                    geomName = 'lineSegment-'
                elif rs.IsPoint(targetGeom) or rs.IsTextDot(targetGeom):
                    geomName = 'point-'
                elif rs.IsSurface(targetGeom) or rs.IsMesh(targetGeom) or rs.IsBrep(targetGeom):
                    geomName = 'planeSegment-'
                
                # Creates new text bit to add to visual text dot text of Rhino geometry
                newTag = tag + '^' + str(count) + '_' + geomName + tagName

            # Creating new visual text dot text
            if ptText0 != '':
                if newTag != '':
                    newText = ''
                    if newTag == '>bound<' and 'bound' in ptText0:
                        text = ptText0.split('|')
                        for item in text:
                            if 'bound' not in item:
                                if newText == '':
                                    newText = item
                                else:
                                    newText += '|' + item
                    if newText == '':           
                        ptText0 += '|' +newTag
                    else:
                        ptText0 = newText + '|' + newTag
            else:
                ptText0 = newTag

            rs.SetUserText(id, 'tagText', ptText0)
            
            if lText0 != '':
                fullText = 'l(' + lText0 + ')'
            
            if ldText0 != '':
                if fullText != '':
                    fullText = fullText + '|d(' + ldText0 +')'
                else:
                    fullText = 'd(' + ldText0 +')'

            if ptText0 != '':
                if fullText != '':
                    fullText = fullText + '|#(' + ptText0 + ')'
                else:
                    fullText = '#(' + ptText0 + ')'

            # Changing contents of text dot label
            if rs.IsCurve(id) or rs.IsSurface(id) or rs.IsMesh(id) or rs.IsBrep(id):
                if rs.IsCurve(id):
                    fullText = 'line(' + rs.GetUserText(id, 'tag') + ')|' + fullText
                    textDot = rs.AddTextDot(fullText, rs.CurveMidPoint(id))
                    obj_type = 'line segment'
                else:
                    fullText = 'plane(' + rs.GetUserText(id, 'tag') + ')|' + fullText
                    obj = rhdoc.Objects.Find(id).BrepGeometry
                    obj_type = 'plane segment'
                    for curve in obj.Loops:
                        if curve.LoopType == r.Geometry.BrepLoopType.Outer:
                            mainCurve = curve.To3dCurve().DuplicateSegments()[0]
                            startpt = mainCurve.PointAtStart
                            endpt = mainCurve.PointAtEnd
                            midpt = ((startpt.X + endpt.X)/2,(startpt.Y + endpt.Y)/2,(startpt.Z + endpt.Z)/2)
                            midpt =  r.Geometry.Point3d(midpt[0], midpt[1], midpt[2])
                            textDot = rs.AddTextDot(fullText, midpt)
                            
                            break
                
                rs.SetUserText(id, "LabelDot", textDot)
                rs.SetUserText(textDot, 'visual', True)
                rs.ObjectColor(textDot, (0,0,0))
            else:
                fullText = 'point(' + rs.GetUserText(id, 'tag') + ')|' + fullText
            
                if rs.IsPoint(id):
                    coords = rs.PointCoordinates(id)
                    orig_color = rs.coercecolor(rs.ObjectColor(id))
                    textDot = rs.AddTextDot(fullText, coords)
                    
                    if ids.index(id) != (len(ids) - 1):
                        ids[ids.index(id)] = textDot

                    for item in rs.GetUserText(id):
                        rs.SetUserText(textDot, item, rs.GetUserText(id, item))

                    rs.DeleteObject(id)
                    rs.ObjectColor(textDot, orig_color)
                    obj_type = 'point'
                    
                elif rs.IsTextDot(id): 
                    textDot = rs.TextDotText(id, fullText)
                    obj_type = 'text dot'

        else:
            print ('Geometry object is not a curve, plane segment (BREP/Surface/Mesh), point or text dot')
    
        string = '' 
        
        if obj_type != '':
            string = obj_type + ' was tagged with ' + tag + '.'
        
    if index != 0:
        string = string + str(index) + ' objects was/were not tagged with ' + tag + ', as they are the incorrect geometric input.'
        
        return string
    
    return True

# General function for adding labels/descriptions to geometry
# If several Rhino objects are selected at the same time, a label/description text input for each of them is prompted for 
# If count is set to True (default), then several objects may be selected and have a label/description placed onto them
def addText(tagType, count):
    retrieve = {'l': 'label', 'd': 'description'}
    signs = {'l' : '&', 'd': '#'}
    retrieve_name = retrieve[tagType]
    sign_name = signs[tagType]

    if count:
        ids = rs.GetObjects('Select objects to place ' + retrieve_name + 's')
    else:
        ids = rs.GetObject('Select object to place ' + retrieve_name)
        ids = [ids]
    
    if len(ids) > 0:
        for id in ids:
            orig_color_geom = rs.coercecolor(rs.ObjectColor(id))
            rs.ObjectColor(id, [255,225,0])
            text_obj = raw_input('Enter ' + retrieve_name + ' text for highlighted object')
            obj_type = '' 
            index = 0
            move_on = False
            key_names = rs.GetUserText(id)
            
            for name in key_names:
                if rs.GetUserText(id, name) == text_obj and sign_name in name:
                    print ('The ' + retrieve_name + ' value already exists in the Rhino object')
                    move_on = True
                
            if move_on:
                continue
                
            # Handles adding labels to LINE SEGMENTS AND PLANES
            if rs.ObjectType(id) == 4 or rs.ObjectType(id) == 8 or rs.ObjectType(id) == 16 or rs.ObjectType(id) == 32 or rs.ObjectType(id) == 1 or rs.ObjectType(id) == 8192:
                
                if text_obj != '' and text_obj != None:
                    lText0 = ''
                    ldText0 = ''
                    ptText0 = ''
                    fullText = ''
                    textCount = 0
                    
                    if 'LabelDot' in rs.GetUserText(id):
                        try:    
                            rs.DeleteObject(rs.GetUserText(id, "LabelDot"))
                            textDot = rs.GetUserText(id, "LabelDot")
                        except:
                            textDot = None
                    
                    # Construction of complete text string to be displayed in visual text dot attached to geometry
                    if 'lText' in rs.GetUserText(id) or 'ldText' in rs.GetUserText(id) or 'tagText' in rs.GetUserText(id):
                        if 'lText' in rs.GetUserText(id):
                            lText0 = rs.GetUserText(id, "lText")
                            if tagType == 'l':
                                textCount = len(lText0.split('|'))
                        
                        if 'ldText' in rs.GetUserText(id):
                            ldText0 = rs.GetUserText(id, "ldText")
                            if tagType == 'd':
                                textCount = len(ldText0.split('|'))
                            
                        if 'tagText' in rs.GetUserText(id):
                            ptText0 = rs.GetUserText(id, "tagText")
                    
                    rs.SetUserText(id, (sign_name + str(textCount)), text_obj)
                    
                    if tagType == 'l':
                        if lText0 != '':
                            lText0 = lText0 + '|' + text_obj
                        else:
                            lText0 = text_obj
                        rs.SetUserText(id, 'lText', lText0)
                    elif tagType == 'd':
                        if ldText0 != '':
                            ldText0 = ldText0 + '|' + text_obj
                        else:
                            ldText0 = text_obj
                        
                        rs.SetUserText(id, 'ldText', ldText0)
                    
                    # Creating new visual text dot text
                    if lText0 != '':
                        fullText = 'l(' + lText0 + ')'
                    
                    if ldText0 != '':
                        if fullText != '':
                            fullText = fullText + '|d(' + ldText0 +')'
                        else:
                            fullText = 'd(' + ldText0 +')'
                            
                    if ptText0 != '':
                        if fullText != '':
                            fullText = fullText + '|#(' + ptText0 + ')'
                        else:
                            fullText = '#(' + ptText0 + ')'
                    
                    if 'tag' in rs.GetUserText(id):
                        if rs.ObjectType(id) == 1 or rs.ObjectType(id) == 8192:
                            geom_name = 'point'
                        elif rs.ObjectType(id) == 4:
                            geom_name = 'line'
                        elif rs.ObjectType(id) == 8 or rs.ObjectType(id) == 16 or rs.ObjectType(id) == 32:
                            geom_name = 'plane'
                            
                        fullText = geom_name+'(' + rs.GetUserText(id, 'tag') + ')|' + fullText

                    if rs.ObjectType(id) == 4 or rs.ObjectType(id) == 8 or rs.ObjectType(id) == 16 or rs.ObjectType(id) == 32:
                        if rs.ObjectType(id) == 4:
                            textDot = rs.AddTextDot(fullText, rs.CurveMidPoint(id))
                            obj_type = 'line segment'
                    
                        elif rs.ObjectType(id) == 8 or rs.ObjectType(id) == 16 or rs.ObjectType(id) == 32:
                            obj = rhdoc.Objects.Find(id).BrepGeometry
                            obj_type = 'plane segment'
                            for curve in obj.Loops:
                                if curve.LoopType == r.Geometry.BrepLoopType.Outer:
                                    mainCurve = curve.To3dCurve().DuplicateSegments()[0]
                                    startpt = mainCurve.PointAtStart
                                    endpt = mainCurve.PointAtEnd
                                    midpt = ((startpt.X + endpt.X)/2,(startpt.Y + endpt.Y)/2,(startpt.Z + endpt.Z)/2)
                                    midpt =  r.Geometry.Point3d(midpt[0], midpt[1], midpt[2])
                                    textDot = rs.AddTextDot(fullText, midpt)
                                    break
                                    
                        rs.SetUserText(id, "LabelDot", textDot)
                        rs.SetUserText(textDot, 'visual', True)
                        rs.ObjectColor(textDot, (0,0,0))
                        rs.ObjectColor(id, orig_color_geom)
                        newGeom = id
                    # Handles adding labels to SIMPLE POINTS
                    elif rs.ObjectType(id) == 1:
                        obj_type = 'point'
                        
                        if fullText != '':
                            # Obtains data from original point
                            rgb = rs.coercecolor(rs.ObjectColor(id))
                            coords = rs.PointCoordinates(id)
                            
                            # Creates new text dot point
                            newPt = rs.AddTextDot(fullText, coords)
                            for name in rs.GetUserText(id):
                                rs.SetUserText(newPt, name, rs.GetUserText(id, name))
                                
                            rs.ObjectColor(newPt, rgb)
                            rs.DeleteObject(id)
                            newGeom = newPt
                    # Handles changing text of TEXT DOTS
                    elif rs.ObjectType(id) == 8192:
                        obj_type = 'text dot'
                        
                        if fullText != '':
                            rs.TextDotText(id, fullText)
                    
                        newGeom = id
            string = text_obj + ' has been added to the ' + obj_type
            
            if rs.ObjectType(newGeom) != 1:
                rs.ObjectColor(newGeom, orig_color_geom)
        
        print ('Label/s have been added to ' + str(len(ids)) + ' object/s')
        return True
    else:
        warnings.warn('No objects were selected.')
        return False
        
# Determines distance based on x, y, z differences
def ptDistance (x, y, z):
    return math.sqrt(x**2 + y**2 + z**2)

# Handles all repetitive tasks in delLabel, delDesc, delTag such as label/description/predicate removal 
# from the user text of a Rhino Object, redrawing text dots, etc
def deletion(type, id, input = 0): 
    lText0 = ''
    ldText0 = ''
    ptText0 = ''
    fullText = ''
    tagName = ''
    removal = ''
    orig_color = None
    tagValue = None
    textDot = None

    if 'LabelDot' in rs.GetUserText(id):
        try:   
            orig_color = rs.coercecolor(rs.ObjectColor(id))
        except:
            pass
        
        try:
            rs.DeleteObject(rs.GetUserText(id, "LabelDot"))
        except:
            pass
    
    # For labels and label descriptions: deletes relevant label/label description  and lText/ldText data from Rhino Object
    # For predicates/directives: obtains tagText data for processing later on
    if 'lText' in rs.GetUserText(id) or 'ldText' in rs.GetUserText(id) or 'tagText' in rs.GetUserText(id) or 'tag' in rs.GetUserText(id):
        if 'lText' in rs.GetUserText(id) and type == 'label':
            removal = rs.GetUserText(id, input)
            rs.SetUserText(id, input)
            rs.SetUserText(id, "lText")

        elif 'ldText' in rs.GetUserText(id) and type == 'desc':
            removal = rs.GetUserText(id, input)
            rs.SetUserText(id, input)
            rs.SetUserText(id, "ldText")

        elif 'tagText' in rs.GetUserText(id) and type == 'pd':
            tagValue = rs.GetUserText(id, "tagText")
            
        # All suceeding conditions in this if-else tree obtain the text of the other two
        # Generates new lText, ldText, ptText, fullText for text dots after deletion, if necessary

        if type == 'label':
            for key_name in rs.GetUserText(id):
                if '&' in key_name:
                    if lText0 == '':
                        lText0 = rs.GetUserText(id, key_name)
                    else:
                        lText0 = lText0 + '|' + rs.GetUserText(id, key_name)
                        
            rs.SetUserText(id, 'lText', lText0)
            ldText0 = rs.GetUserText(id, 'ldText')
            ptText0 = rs.GetUserText(id, 'tagText')
            tagName = rs.GetUserText(id, 'tag')
            
        # Generates new ldText for label descriptions after deletion
        elif type == 'desc':
            for key_name in rs.GetUserText(id):
                if '#' in key_name:
                    if ldText0 == '':
                        ldText0 = rs.GetUserText(id, key_name)
                    else:
                        ldText0 = ldText0 + '|' + rs.GetUserText(id, key_name)
            
            rs.SetUserText(id, 'ldText', ldText0)
            lText0 = rs.GetUserText(id, 'lText')
            ptText0 = rs.GetUserText(id, 'tagText')
            tagName = rs.GetUserText(id, 'tag')
            
        # Deletes relevant predicate/directive from tagValue
        # Generates new ptText for predicates/directives after deletion
        elif type == 'pd' and tagValue != None:
            tagValue = tagValue.split('|')
            checkAgainst = tagValue[input]
            removal = tagValue.pop(input)
            tagR = None
            groupName = removal.split('_')[0].split('^')[-1]
            
            if 'pointOL' in removal:
                tagR = tags['pointOL']
            elif 'length' in removal:
                tagR = tags['length']
            elif 'angle' in removal:
                tagR = tags['angle']
            elif 'distance' in removal:
                tagR = tags['distance']
            elif 'normal' in removal:
                tagR = tags['normal']
                rs.SetUserText(id, tagR + '_end_' + groupName)
                rs.SetUserText(id, tagR + '_rhs_' + groupName)
            elif 'embeds' in removal:
                rs.SetUserText(id, 'embeds_' + groupName)
                
            if tagR != None:
                if groupName != None:
                    rs.SetUserText(id, tagR + '1_' + groupName)
                    rs.SetUserText(id, tagR + '2_' + groupName)
                    if tagR == 'dist' or tagR == 'ang':
                        rs.SetUserText(id, 'pair_' + tagR + '_' + groupName)
                    elif tagR == 'norm':
                        rs.SetUserText(id, 'new_line_' + groupName)
                else:
                    rs.SetUserText(id, tagR + '1')
                    rs.SetUserText(id, tagR + '2')
                
            for val in tagValue:
                if ptText0 == '':
                    ptText0 = val
                else:
                    ptText0 = ptText0 + '|' + val
                    
            rs.SetUserText(id, 'tagText', ptText0)
            lText0 = rs.GetUserText(id, 'lText')
            ldText0 = rs.GetUserText(id, 'ldText')
            tagName = rs.GetUserText(id, 'tag')
            
        elif type == 'tag':
            # Sets 'tag' input in User Text Dictionary to empty
            # Also does the same for tagText -> predicates/directives
            removal = rs.GetUserText(id, 'tag')
            rs.SetUserText(id, 'tag')
            #rs.SetUserText(id, 'tagText')
            ptText0 = rs.GetUserText(id, 'tagText')
            lText0 = rs.GetUserText(id, 'lText')
            ldText0 = rs.GetUserText(id, 'ldText')
    
    # Composes fullText for display in textDot
    if lText0 != '':
        fullText = 'l(' + lText0 + ')'
        
    if ldText0 != '':
        if fullText != '':
            fullText = fullText + '|d(' + ldText0 +')'
        else:
            fullText = 'd(' + ldText0 +')'
            
    if ptText0 != '':
        if fullText != '':
            fullText = fullText + '|#(' + ptText0 + ')'
        else:
            fullText = '#(' + ptText0 + ')'

    if tagName != '':
        if rs.ObjectType(id) == 4:
            geomName = 'line'
        elif rs.ObjectType(id) == 1 or rs.ObjectType(id) == 8192:
            geomName = 'point'
        elif rs.ObjectType(id) == 8 or rs.ObjectType(id) == 16 or rs.ObjectType(id) == 32:
            geomName = 'plane'

        if fullText != '':
            fullText = geomName + '(' + rs.GetUserText(id, 'tag') + ')|' + fullText
        else:
            fullText = geomName+ '(' + rs.GetUserText(id, 'tag') + ')'

    # Changes the text dot text if fullText is not empty
    if fullText != '': 
        if rs.ObjectType(id) == 1 or rs.ObjectType(id) == 8192:
            if rs.ObjectType(id) == 1:
                if fullText != '':
                    orig_color = rs.coercecolor(rs.ObjectColor(id))
                    coords = rs.PointCoordinates(id)
                    textDot = rs.AddTextDot(fullText, coords)
                    rs.ObjectColor(textDot, orig_color)
            # Changes text dot directly
            elif rs.ObjectType(id) == 8192:
                if fullText  == '':
                    orig_color = rs.coercecolor(rs.ObjectColor(id))
                    textDot = rs.AddPoint(rs.TextDotPoint(id))
                    rs.ObjectColor(textDot, orig_color)
                    rs.DeleteObject(id)
                else:
                    rs.TextDotText(id, fullText)
                    textDot = id
                    
            return textDot
        
        # Generates new text dot to contain fullText (previous one was deleted)
        elif rs.ObjectType(id) == 4:
            if 'LabelDot' in rs.GetUserText(id):
                rs.DeleteObject(rs.GetUserText(id, 'LabelDot'))
            textDot = rs.AddTextDot(fullText, rs.CurveMidPoint(id))
    
        elif rs.ObjectType(id) == 8 or rs.ObjectType(id) == 16 or rs.ObjectType(id) == 32:
            obj = rhdoc.Objects.Find(id).BrepGeometry
            obj_type = 'plane segment'
            for curve in obj.Loops:
                if curve.LoopType == r.Geometry.BrepLoopType.Outer:
                    mainCurve = curve.To3dCurve().DuplicateSegments()[0]
                    startpt = mainCurve.PointAtStart
                    endpt = mainCurve.PointAtEnd
                    midpt = ((startpt.X + endpt.X)/2,(startpt.Y + endpt.Y)/2,(startpt.Z + endpt.Z)/2)
                    midpt =  r.Geometry.Point3d(midpt[0], midpt[1], midpt[2])
                    textDot = rs.AddTextDot(fullText, midpt)
                    break
        
        if textDot != None:
            rs.SetUserText(id, "LabelDot", textDot)
            rs.SetUserText(textDot, 'visual', True)
            rs.ObjectColor(textDot, (0,0,0))
            rs.ObjectColor(id, orig_color)
    
    # Deletes text dot if fullText is empty
    else:
        if rs.ObjectType(id) == 8192:
            newID = rs.AddPoint(rs.TextDotPoint(id))
            rs.DeleteObject(id)
            id = newID
        else:
            if 'LabelDot' in rs.GetUserText(id):
                rs.DeleteObject(rs.GetUserText(id, 'LabelDot'))
                rs.SetUserText(id, 'LabelDot')
      
    print (removal + ' was removed from geometry')       
    return id
            
class rhino_methods: 
    # Adds label to user text of Rhino object
    # If several Rhino objects are selected at the same time, a label text input for each of them is prompted for 
    # If count is set to True (default), then several objects may be selected and have a label placed onto them
    @staticmethod
    def add_label(count = True):
        return addText('l', count)
    
    # Adds description to user text of Rhino object
    # If several Rhino objects are selected at the same time, a description text input for each of them is prompted for 
    # If count is set to True (default), then several objects may be selected and have a description placed onto them
    @staticmethod
    def add_description(count = True):
        return addText('d', count)
    
    # Adds enumerative value to user text of Rhino Object
    @staticmethod
    def add_enum():
        id = rs.GetObject('Select object to add enumerative value to')    
        
        if id != None:
            print ('Enter single value to be entered as enumerative input. This will be read as a string.')
            enumText = raw_input('Enter enumerative value')
            rs.SetUserText(id, 'Enum', enumText)
            
            print (enumText + ' was entered as an enumerative value.')
            return True
        else:
            warnings.warn ('No object selected.')
            return False
            
    # Adds width to curve object or grayscale to point
    @staticmethod
    def add_weight(several = True):
        # Handles several objects
        if several:
            ids =  rs.GetObjects("Select objects to change weight")
            lines, points = [], []
            rs.ObjectColor(ids, [255,225,0])                             # Highlights objects selected
            
            for id in ids:
                if rs.ObjectType(id) == 1 or rs.ObjectType(id) == 8192:     # Sorts objects into points or lines
                    points.append(id)
                elif rs.ObjectType(id) == 4:
                    lines.append(id)
            
            # Adds width to lines, if any
            if len(lines) !=0:   
                weight = raw_input('Enter weight for selected line/s (RANGE: 0.0 to 2.0mm)')
                try:
                    weight = float(weight)
                except:
                    raise TypeError('Invalid input for weight.  Line widths are only decmal values from 0.0 to 2.0mm.')
                    
                if weight <= 0.0 or weight >= 2.0 or type(weight) != type(0.0):
                    raise TypeError('Invalid input for weight.  Line widths are only decmal values from 0.0 to 2.0mm.')

                else:
                    for line in lines:
                        rs.ObjectPrintWidth(line, weight)
                
            # Adds grayscale to points, if any    
            if len(points) !=0:
                rs.ObjectColor(points, [255,225,0])  
                weight= raw_input('Enter color for highlighted point/s (RANGE: 0 to 255)')
                try:
                    weight = int(weight)
                except:
                   raise TypeError('Invalid input for weight.  Point weights are only integers from 0 (black) to 255 (white).')

                if (rs.ObjectType(id) == 1 or rs.ObjectType(id) == 8192) and (weight <= 0.0 or weight >= 255 or type(weight) != type(0)):
                    raise TypeError('Invalid input for weight.  Point weights are only integers from 0 (black) to 255 (white).')
                else:
                    # Changes color of point
                    rs.ObjectColor(id, (weight, weight, weight))
                    
        # Handles one object       
        elif several == False:
            id = rs.GetObject("Select object to change weight")
            orig_color = rs.coercecolor(rs.ObjectColor(id))
            rs.ObjectColor(id, [255,225,0])
            
            # Adds width to line, if any
            if rs.ObjectType(id) == 4:
                weight= raw_input('Enter weight for selected line/s (RANGE: 0.0 to 2.0mm)')
                try:
                    weight = float(weight)
                except:
                    raise TypeError('Invalid input for weight.  Line widths are only decimals values from 0.0 to 2.0mm.')

            # Adds grayscale to point/text dot, if any        
            elif (rs.ObjectType(id) == 1 or rs.ObjectType(id) == 8192):
                weight= raw_input('Enter color for selected point/s (RANGE: 0 to 255)')
                try:
                    weight = int(weight)
                except:
                    raise TypeError('Invalid input for weight.  Point weights are only integers from 0 (black) to 255 (white).')
                    
            if rs.ObjectType(id) == 4 and (weight < 0.0 or weight > 2.0 or type(weight) != type(0.0)):
                raise TypeError('Invalid input for weight.  Line widths are only decmal values from 0.0 to 2.0mm.')
                
            elif (rs.ObjectType(id) == 1 or rs.ObjectType(id) == 8192) and (weight < 0.0 or weight > 255 or type(weight) != type(0)):
                raise TypeError('Invalid input for weight.  Point weights are only integers from 0 (black) to 255 (white).')
                
            else:    
                if rs.ObjectType(id) == 4:                                  # Changes line width
                    rs.ObjectPrintWidth(id, weight)
                
                elif rs.ObjectType(id) == 1 or rs.ObjectType(id) == 8192:   # Changes point/text dot color
                    rs.ObjectColor(id, (weight, weight, weight))
                    rs.SetUserText(id, "Color", str(weight), True)
        else:
            warnings.warn('Invalid input for number of objects to be handled. Please enter either only True (select several objects) or False (select only one object).')
            return False
        
        return True
        
    # Changes color of an object or several objects by selecting the object/s and inputting the RGB value
    # Note: Color is not visible in Print Preview
    @staticmethod
    def add_color(r, g, b):
        objectIds = rs.GetObjects('Select which objects to change color of')
        color = []
        
        if type(r) == type(0) and type(g) == type(0) and type(b) == type(0):
            if r < 0 or r > 255:
                raise ValueError('r: ' + str(r) + ' is not within the color range of 0 to 255.')
            else:
                color.append(r)
                
            if g < 0 or g > 255:
                raise ValueError('g: ' + str(r) + ' is not within the color range of 0 to 255.')
            else:
                color.append(g)
            
            if b < 0 or b > 255:
                raise ValueError('b: ' + str(b) + ' is not within the color range of 0 to 255.')
            else:
                color.append(b)
                
            if len(color) == 3:
                rs.ObjectColor(objectIds, (color[0], color[1], color[2]))
                return True
            else:
                return False
        else:
            raise TypeError('Incorrect input for color. All three inputs must be integers between 0 to 255.')
            
    # Deletes all Rhino objects 
    # Sets up new frames
    # Cleans up sort and rule registers
    @staticmethod
    def clear_everything(): 
        rs.DeleteObjects(rs.AllObjects())
        sort.cleanup()
        rule.cleanup()
        flow.cleanup()
        form.formRegister = {}
        sdlParser.rules.clear()
        sdlParser.forms.clear()
        sdlParser.flows.clear()
        return True
    
    # Deletes the list of shape GUIDs associated with a form from the Rhino workspace
    # Input can be a list, tuple, dictionary or single item
    @staticmethod
    def clear_shape(guids):
        if type(guids) == type([]) or type(guids) == type({}) or type(guids) == type(()):
            if len(guids) > 0:
                for shape in guids:
                    if type(shape[0]) == type([]):
                        for j in shape:
                            rs.DeleteObjects(j)
                    elif type(shape) == type('a'):
                        rs.DeleteObjects(shapeIds[shape])
                    else:
                        rs.DeleteObject(shape)
                
                return True
            else:
                warnings.warn('No objects were selected. for deletion from Rhino workspace.')
                return False
        else:
            try:
                rs.DeleteObject(guids)
                return True
            except:
                raise TypeError('Invalid input for deletion from Rhino workspace. This must be either a list/tuple/dictionary of Rhino GUIDs or a single Rhino GUID.')
                
    # Adds maximum line tag to user text of line segments/polylines
    # Generates text dot with text '#(max)' 
    @staticmethod
    def add_max_line(tagSelf = True):
        return addPredDir('max', tagSelf)
    
    # Adds shortest line tag to user text of line segments/polylines
    # Generates text dot with text '#(short)' 
    @staticmethod
    def add_shortest_line(tagSelf = True):
        return addPredDir('shortest', tagSelf)
        
    # Adds longest line tag to user text of line segments/polylines
    # Generates text dot with text '#(long)' 
    @staticmethod
    def add_longest_line(tagSelf = True):
        return addPredDir('longest', tagSelf)

    # Adds bound line tag to user text of line segments/polylines
    # Generates text dot with text '#(>bound) or #(bound<) or #(>bound<)' 
    @staticmethod
    def add_bound_line(ends = 2, tagSelf = True):
        return addPredDir('bound', tagSelf, None, None, ends)

    # Adds point on line tag to user text of line segments/polylines
    # Generates text dot with text '#(polyline_<value1>_<value2>' if different values, '#(polyline_<value1>' if two similar values)' 
    @staticmethod
    def add_no_label(tagSelf = True):
        if type(tagSelf) == type(True):
            return addPredDir('nolabel', tagSelf)
    
    # Adds void line tag to user text of line segments/polylines/plane segments/points
    # Generates text dot with text '#(void_<void number>_<order in void predicate>)' 
    # The void number is based on the number of void predicates already pre-existing in the Rhino workspace
    # The order in the void predicate is used to keep track of the order of the vertices
    # Points/line segments should be entered in the order in which they will be connected
    # scriptcontext.sticky['count'] is used to track the set of geometries comprising a void predicate
    # point checker for collinearity: (c-a)*(f-b) - (d - b)*(e - a) = 0
    # Also functions like now-defunct voidOf => directive checks if a certain boundary is clear of objects with specified sort(s0
    # Generates text dot with text '#(polyline_<value1>_<value2>' if different values, '#(polyline_<value1>' if two similar values)' 
    @staticmethod
    def add_void(sortType = [], detectIntersection = False):
        if type(sortType) == type('') or type(sortType) == type([]):
            if type(sortType) == type(''):
                if sortType not in sort.register:
                    raise ValueError('Sort type', sortType, 'does not exist in sort register')
            elif type(sortType) == type([]):
                if len(sortType) > 0:
                    for sortName in sortType:
                        if sortName not in sort.register:
                            raise ValueError('Sort type ' + sortType +' does not exist in sort register')
            message = 'Select object/s to tag as void polygon/s. For points/lines, please select them in the same order as they would be listed in a polygon'
        else:
            raise TypeError('Incorrect input for sort type for voidOf predicate. Input should be a text string or a list of sort type names')

        ids = rs.GetObjects(message)
        obj_type = ''
        geometry = [[], [], []]
        index = 0
        geomCount = 0
        
        for id in ids:
            if rs.ObjectType(id) == 1 or rs.ObjectType(id) == 8192:
                geometry[0].append(id)
            
            # Check for indiv line segments
            elif rs.ObjectType(id) == 4:
                geometry[1].append(id)

            elif rs.ObjectType(id) == 8 or rs.ObjectType(id) == 16 or rs.ObjectType(id) == 32:
                geometry[2].append(id)
                    
        # Checks if any geometry is available to tag with the void predicate
        if geometry[0] == [] and geometry[1] == [] and geometry[2] == []:
            warnings.warn ('No geometry to tag with void predicate; exiting function')
            return False
        
        # Checking validity of accepted geometry
        # points, line segments, plane segments
        
        # Points
        # Checks if geometry fulfills conditions sufficient for void tagging
        if len(geometry[0]) >= 3:
            cullList = []
            
            for pt in geometry[0]:
                if 'visual' not in rs.GetUserText(pt):
                    if rs.ObjectType(pt) == 8192:
                       cullList.append(rs.TextDotPoint(pt))
                    else:
                        cullList.append(pt)
                else:
                    continue
            
            if len(cullList) > 0:
                culledTempList = rs.CullDuplicatePoints(cullList, 0.00001)
                culledList = []
                tempPtList = []
                
                #need to fix this
                for geom in geometry[0]:
                    for obj in culledTempList:
    
                        if rs.ObjectType(geom) == 8192:
                            pt = rs.TextDotPoint(geom)
                        else:
                            pt = rs.PointCoordinates(geom)
    
                        if obj.X == pt.X and obj.Y == pt.Y and obj.Z == pt.Z: 
                            tempPtList.append(geom)
                            culledList.append(culledTempList.pop(culledTempList.index(obj)))
                            
                        if len(culledList) ==  0:
                            break
                
                vectorList = []
                coincidenceList = 0
                tempList = []
       
                tempList = rs.SortPointList(culledList) #  Sorts a list of 3-D points so they will be connected in "reasonable" polyline order
    
                for vector in tempList:
                    vectorList.append(r.Geometry.Vector3d(vector.X, vector.Y, vector.Z))
                    
                # Computes initial line's unit direction vector
                A = vectorList[0]
                B = vectorList[1]
                C = vectorList[2]
                
                dVector =  (B[0]-A[0], B[1]-A[1], B[2]-A[2]) 
                norm = math.sqrt((dVector[0])**2 + (dVector[1])**2 + (dVector[2])**2)
                uVectorABS = (math.fabs(dVector[0]/norm), math.fabs(dVector[1]/norm), math.fabs(dVector[2]/norm))
    
                for c in vectorList[2:]:
                    if c != A and c != B:
                        dVectorC =  (A[0]-c[0], A[1]-c[1], A[2]-c[2]) 
                        normC = math.sqrt(dVectorC[0]**2 + dVectorC[1]**2 + dVectorC[2]**2)
                        uVectorCABS = (math.fabs(dVectorC[0]/normC), math.fabs(dVectorC[1]/normC), math.fabs(dVectorC[2]/normC))
                        
                        if uVectorCABS != uVectorABS:
                            break
                        else:
                            coincidenceList += 1
                    else:
                        continue
                        
                if coincidenceList == (len(vectorList[2:])):
                    warnings.warn ('The input points do not form a polygon, as all the points are collinear. Exiting function.')
                    return False
                
                # Generates text of text dot for each geometry
                ptText0 = 'void' 
                for id in tempPtList:
                    lText0 = ''
                    ldText0 = ''
                    tagName = ''
                    fullText= '' 
                    voidText = ('void_' + str(sc.sticky['voidCount']) + '_' + str (geomCount))
                    
                    if type(sortType) == type('') and sortType != '':
                        rs.SetUserText(id, 'sort_' + str(sc.sticky['voidCount']), sortType)
                        voidText += '_' + sortType
                    elif type(sortType) == type([]) and len(sortType) != 0:
                        sortTag = sortType[0]
                        for sortName in sortType[1:]:
                            sortTag += '_' + sortName
                            
                        rs.SetUserText(id, 'sort_' + str(sc.sticky['voidCount']), sortTag)
                        voidText += '_' + sortTag
    
                    # Simple point
                    # Creates new text dot
                    if rs.ObjectType(id) == 1:
                        textDot =  rs.AddTextDot(voidText, id)
                        rs.SetUserText(textDot, 'tagText', voidText)
                        for userText in rs.GetUserText(id):
                            rs.SetUserText(textDot, userText, rs.GetUserText(id, userText))
                            
                        rs.DeleteObject(id)
                        
                    # Text dot
                    # Alters text of text dot
                    elif rs.ObjectType(id) == 8192:
                        if 'tagText' in rs.GetUserText(id):
                            ptText0 = rs.GetUserText(id, 'tagText')
                            ptText0 = ptText0 + '|' + voidText
                        else:
                            ptText0 = voidText
                            
                        rs.SetUserText(id, 'tagText', ptText0)
                            
                        if 'lText' in rs.GetUserText(id):
                            lText0 = rs.GetUserText(id, "lText")
                    
                        if 'ldText' in rs.GetUserText(id):
                            ldText0 = rs.GetUserText(id, "ldText")
                        
                        if 'tag' in rs.GetUserText(id):
                            tagName = rs.GetUserText(id, 'tag')

                        if lText0 != '':
                            fullText = 'l(' + lText0 + ')'

                        if ldText0 != '':
                            if fullText != '':
                                fullText = fullText + '|d(' + ldText0 +')'
                            else:
                                fullText = 'd(' + ldText0 +')'
                                
                        if ptText0 != '':
                            if fullText != '':
                                fullText = fullText + '|#(' + ptText0 + ')'
                            else:
                                fullText = '#(' + ptText0 + ')'
                        
                        if tagName != '':
                            if fullText != '':
                                fullText = 'point(' + tagName + ')|' + fullText
                            else:
                                fullText = 'point(' + tagName + ')'
                        
                        rs.TextDotText(id, fullText)
                        
                    geomCount += 1
                    
                if len(geometry[0]) > 0:
                    sc.sticky['voidCount'] += 1
            else:
                if len(geometry[0]) > 0 and len(geometry[0]) <= 2:
                    print ('Insufficient points to tag as void polygon')
            
        # Lines            
        if len(geometry[1]) > 0:
            geomCount = 0
            
            if len(geometry[1]) > 1:
                geomStart = rs.CurveStartPoint(geometry[1][0])
                geomEnd = rs.CurveEndPoint(geometry[1][-1])
                """
                if geomStart.X != geomEnd.X or geomStart.Y != geomEnd.Y or geomStart.Z != geomEnd.Z:
                    print ('Selected line segments do not form a closed polygon')
                    return
                """
            else:
                if rs.IsPolyline(geometry[1][0]) == False:
                    raise TypeError('Insufficient line segments/polylines to tag as void predicate object')

            permList = itertools.permutations(geometry[1], 2)
            
            if detectIntersection:
                for perm in permList:
                    if rs.CurveCurveIntersection(perm[0], perm[1]) != None:
                        raise TypeError('Line segment/s cannot be tagged as void objects, as the polygon they form is self-intersecting')
                      
            for id in geometry[1]:
                if rs.CurveCurveIntersection(id) != None and detectIntersection:
                    raise TypeError('Line segment/s cannot be tagged as void objects, as they are self-intersecting')
                    return 
                    
                lText0 = ''
                ldText0 = ''
                ptText0 = ''
                tagName = ''
                fullText = ''
                
                if 'LabelDot' in rs.GetUserText(id):
                    try:    
                        textDot = rs.GetUserText(id, "LabelDot")
                        rs.DeleteObject(rs.GetUserText(id, "LabelDot"))
                    except:
                        textDot = None
                
                # SIMPLIFY SOLUTION for creating visual text strings!!!
                if 'lText' in rs.GetUserText(id) or 'ldText' in rs.GetUserText(id) or 'tagText' in rs.GetUserText(id) or 'tag' in rs.GetUserText(id):
                    if 'lText' in rs.GetUserText(id):
                        lText0 = rs.GetUserText(id, "lText")
                        
                    if 'ldText' in rs.GetUserText(id):
                        ldText0 = rs.GetUserText(id, "ldText")
                        
                    if 'tagText' in rs.GetUserText(id):
                        ptText0 = rs.GetUserText(id, "tagText")
                        
                    if 'tag' in rs.GetUserText(id):
                        tagName = rs.GetUserText(id, 'tag')
                
                voidText = 'void_' + str(sc.sticky['voidCount'])
                
                if type(sortType) == type('') and sortType != '':
                    rs.SetUserText(id, 'sort_' + str(sc.sticky['voidCount']), sortType)
                    voidText += '_' + sortType
                elif type(sortType) == type([]) and len(sortType) != 0:
                    sortTag = sortType[0]
                    for sortName in sortType[1:]:
                        sortTag += '_' + sortName
                        
                    rs.SetUserText(id, 'sort_' + str(sc.sticky['voidCount']), sortTag)
                    voidText += '_' + sortTag
                
                if ptText0 != '':
                    ptText0 = ptText0 + '|' + voidText
                else:
                    ptText0 = voidText
                
                rs.SetUserText(id, ('tagText'), ptText0)
                
                if tagName != '':
                    fullText = 'line(' + tagName + ')'
                    
                if lText0 != '':
                    if fullText == '':
                        fullText = 'l(' + lText0 + ')'
                    else:
                        fullText += 'l(' + lText0 + ')'
                    
                if ldText0 != '':
                    if fullText != '':
                        fullText = fullText + '|d(' + ldText0 +')'
                    else:
                        fullText = 'd(' + ldText0 +')'
                        
                if ptText0 != '':
                    if fullText != '':
                        fullText = fullText + '|#(' + ptText0 + ')'
                    else:
                        fullText = '#(' + ptText0 + ')'
                
                if 'tag' in rs.GetUserText(id, 'tag'):
                    fullText = 'line(' + rs.GetUserText(id, 'tag') + ')|' + fullText

                textDot = rs.AddTextDot(fullText, rs.CurveMidPoint(id))
                
                if rs.IsPolyline(id) or rs.IsPolyCurve(id):
                    obj_type = 'polyline'
                elif rs.IsCurve(id):
                    obj_type = 'line segment'
                
                rs.SetUserText(id, "LabelDot", textDot)
                rs.SetUserText(textDot, 'visual', True)
                rs.ObjectColor(textDot, (0,0,0))
                geomCount += 1
            
            sc.sticky['voidCount'] += 1
            
        if len(geometry[2]) > 0:
            
            for id in geometry[2]:
                obj = rhdoc.Objects.Find(id).BrepGeometry
                obj_type = 'plane segment'
                lText0 = ''
                ldText0 = ''
                ptText0 = ''
                fullText = ''
                
                if 'LabelDot' in rs.GetUserText(id):
                    try:    
                        rs.DeleteObject(rs.GetUserText(id, "LabelDot"))
                        textDot = rs.GetUserText(id, "LabelDot")
                    except:
                        textDot = None
                
                if 'lText' in rs.GetUserText(id) or 'ldText' in rs.GetUserText(id) or 'tagText' in rs.GetUserText(id):
                    if 'lText' in rs.GetUserText(id):
                        lText0 = rs.GetUserText(id, "lText")
                        
                    if 'ldText' in rs.GetUserText(id):
                        ldText0 = rs.GetUserText(id, "ldText")
                        
                    if 'tagText' in rs.GetUserText(id):
                        ptText0 = rs.GetUserText(id, "tagText")
                
                voidText = 'void_' + str(sc.sticky['voidCount'])
                
                if type(sortType) == type('') and sortType != '':
                    rs.SetUserText(id, 'sort_' + str(sc.sticky['voidCount']), sortType)
                    voidText += '_' + sortType
                elif type(sortType) == type([]) and len(sortType) != 0:
                    sortTag = sortType[0]
                    for sortName in sortType[1:]:
                        sortTag += '_' + sortName
                        
                    rs.SetUserText(id, 'sort_' + str(sc.sticky['voidCount']), sortTag)
                    voidText += '_' + sortTag

                if ptText0 != '':
                    ptText0 = ptText0 + '|' + voidText
                else:
                    ptText0 = voidText
                
                rs.SetUserText(id, ('tagText'), ptText0)
                
                if lText0 != '':
                    fullText = 'l(' + lText0 + ')'
                
                if ldText0 != '':
                    if fullText == '':
                        fullText = 'd(' + ldText0 + ')'
                    else:
                        fullText += '|d(' + ldText0 +')'
                    
                if ptText0 != '':
                    if fullText == '':
                        fullText = '#(' + ptText0 + ')'
                    else:
                        fullText = fullText + '|#(' + ptText0 + ')'
                        
                if 'plane' in rs.GetUserText(id, 'tag'):
                    fullText = 'plane(' + rs.GetUserText(id, 'tag') + ')|' + fullText
                
                for curve in obj.Loops:
                    if curve.LoopType == r.Geometry.BrepLoopType.Outer:
                        mainCurve = curve.To3dCurve().DuplicateSegments()[0]
                        startpt = mainCurve.PointAtStart
                        endpt = mainCurve.PointAtEnd
                        midpt = ((startpt.X + endpt.X)/2,(startpt.Y + endpt.Y)/2,(startpt.Z + endpt.Z)/2)
                        midpt =  r.Geometry.Point3d(midpt[0], midpt[1], midpt[2])
                        textDot = rs.AddTextDot(fullText, midpt)
                        break

                rs.SetUserText(textDot, 'visual', True)
                rs.ObjectColor(textDot, (0,0,0))
            sc.sticky['voidCount'] += 1
             
        string = '' 
        
        if obj_type != '':
            string = obj_type[0].upper() + obj_type[1:] + ' was tagged as a void polygon'
            
        if index != 0:
            string = string + str(index) + ' objects was/were not tagged as ' + tag + ' polygons, as they are not line segments/plane segments/points'
        
        print (string)
        return True
    
    # Adds point on line tag to user text of line segments/polylines
    # Generates text dot with text '#(polyline_<value1>_<value2>' if different values, '#(polyline_<value1>' if two similar values)' 
    @staticmethod
    def add_point_on_line(seg1 = None, seg2 = None, tagSelf = True):
        # Checks correctness of inputs for seg1 and seg2
        if type(seg1) == type(0.0) or type(seg1) == type('') or seg1 == None:
            if type(seg1) == type(0.0):
                if seg1 >= 1.0 or seg1 <= 0.0:
                    raise ValueError('Minimum ratio value for Point on Line must be greater than 0 or less than 1.')
        else:
            raise ValueError('Invalid input for minimum segment ratio of Point on Line')
            
        if type(seg2) == type(0.0) or seg2 == None or type(seg2) == type(''):
            if type(seg2) == type(0.0):
                if seg2 >= 1.0 or seg2 <= 0.0:
                    raise ValueError('Maximum ratio value for Point on Line must be greater than 0 or less than 1.')
            if seg2 == None:
                seg2 = seg1
        else:
            raise ValueError('Invalid input for maximum segment ratio of Point on Line')
            
        return addPredDir('pointOL', tagSelf, seg1, seg2)
    
    """
    Adds normal generation directive tag to user text of an existing line
    Generates text dot with text '#(normal-<x>_<y>_<length>)' 
    *******************************************************************************
    Directive: normal2D - creating a new line that is perpendicular to an existing line starting at a certain direction vector
                          syntax : [lineTag, x, y, lengthMin, lengthMax, targetSortType]
                                or [lineTag, x, y, length]
                                or [planeTag, rhsLineTag, lengthMin, lengthMax]
                                   x, y: x and y coordinate of the start point of the new normal line. 
                                   This starting coordinate must exist in the lhs as an end-point/intersection-point/point/vector
                                         
                                   length/lengthMin/lengthMax: length of the new normal line. 
                                   if 2 values are given, a random number between the 2 values will be taken
                                   if length is not given, the new normal will be defaulted to have a length of the lhs object
                                           
                                   lineTag/planeTag: tag of the line/plane in the lhs that the new normal will be perpendicular to  
                                   
                                   rhsLineTag: tag of the new normal line in the rhs shape
    """
    @staticmethod    
    def add_normal(rhsTag, length, targetCoords = None, tagSelf = True):
        coords = None
        
        if type(rhsTag) != type('a'):
            raise TypeError('Input for RHS tag name of new normal line is not a text string.')
        elif type(rhsTag) == type('a') and len(rhsTag) == 0:
            raise TypeError('Input for RHS tag name of new normal line is empty.')
        else:
            if 'lineSegment-' not in rhsTag:
                rhsTag = 'lineSegment-' + rhsTag

        # Checks and prepares length value/s
        if type(length) != type([]) and type(length) != type('a') and type(length) != type(1.0) and type(length) != type(1):
            raise TypeError('Invalid input for length; only a list of two values, or a single integer, float, or description expression is allowed.')
            
        elif type(length) == type([]):
            if len(length) != 2:
                raise ValueError('There are more than two elements in the input for length of the new normal line segment.')
            else:
                if type(length[0]) != type('a') and type(length[0]) != type(1.0) and type(length[0]) != type(1):
                    raise ValueError('Invalid input for minimum length; only a single integer, float, or description expression is allowed.')
                if type(length[1]) != type('a') and type(length[1]) != type(1.0) and type(length[1]) != type(1):
                     raise ValueError('Invalid input for maximum length; only a single integer, float, or description expression is allowed.')
        
        # Checks correctness of value within targetCoords 
        # This can be a set of coordinates (for use with a line segment)
        # Or the tag of the plane segment that the new normal line will be generated from
        if targetCoords != None:
            try:
                if rs.ObjectType(targetCoords) == 1: 
                    coords = (rs.PointCoordinates(targetCoords).X, rs.PointCoordinates(targetCoords).Y, rs.PointCoordinates(targetCoords).Z)
                elif rs.ObjectType(targetCoords) == 8192:
                    coords = rs.TextDotPoint(targetCoords)
                    coords = (targetCoords.X, targetCoords.Y, targetCoords.Z)
                elif rs.ObjectType(targetCoords) != 1 and rs.ObjectType(targetCoords) != 8192:
                    raise TypeError('Invalid coordinate input for the direction vector/RHS tag name of the new normal line; this should be a list, a tuple, a Rhino point/vector or a text string')
            except:
                if type(targetCoords) == r.Geometry.Vector3d or type(targetCoords) == r.Geometry.Point3d:
                    coords = (targetCoords.X, targetCoords.Y, targetCoords.Z)
                elif type(targetCoords) == type('a'):
                    coords = targetCoords
                elif type(targetCoords) != type(()) and type(targetCoords) != type([]):
                    raise TypeError('Invalid coordinate input for the direction vector/RHS tag name of the new normal line; this should be a list, a tuple, a Rhino point/vector or a text string')

            return addPredDir('normal', tagSelf, length, coords, rhsTag)
        
        else:
            return addPredDir('normal', tagSelf, length, None, rhsTag)
            
    # Adds distance tag to user text of line segments/polylines/points/text dots
    # Generates text dot with text '#(distance-<group no>_dist1_dist2)' 
    @staticmethod
    def add_distance(dist1 = None, dist2 = None, tagSelf = True):
        # Checks correctness of inputs for seg1 and seg2
        if type(dist1) == type(0.0) or type(dist1) == type('') or type(dist1) == type(0):
            pass
        else:
            if dist1 == None:
                raise TypeError('Please input at least one value for the distance between the two geometries.')
            else:
                raise ValueError('Invalid input for minimum distance value')
            
        if type(dist2) == type(0.0) or dist2 == None or type(dist2) == type('') or type(dist2) == type(0):
            if dist2 == None:
                dist2 = dist1
        else:
            raise ValueError('Invalid input for maximum distance value')
            
        return addPredDir('distance', tagSelf, dist1, dist2)
    
    # Adds embeds predicate to two geometries
    # First selected geometry is the larger container geometry
    # Second selected geometry is the embedded geometry
    @staticmethod
    def add_embeds(tagSelf = True):
        addPredDir('embeds', tagSelf)
    
    # Removes label tag of chosen label from user text of Rhino Object
    # Redraws text dot accompanying Rhino Object to reflect the change
    @staticmethod
    def delete_label():
        objectIDs = rs.GetObjects('Select objects to remove label from')
        print ('You have selected ' + str(len(objectIDs)) + ' objects to remove labels from') 
        go_ahead = False
        
        if len(objectIDs) == 0:
            warnings.warn('No objects were selected.')
            return False
        
        # Goes through list of selected objects
        for id in objectIDs:
            if rs.ObjectType(id) == 1 or rs.ObjectType(id) == 4 or rs.ObjectType(id) == 8192 or rs.ObjectType(id) == 8 or rs.ObjectType(id) == 16 or rs.ObjectType(id) == 32:
                objText = rs.GetUserText(id)
                for key_name in objText:
                    if '&' in key_name:
                        go_ahead = True
                
                if go_ahead:
                    print ('The following labels are present in this object:')
                    
                    for key_name in objText:
                        if '&' in key_name:
                            print (key_name[1:] + ': ' + rs.GetUserText(id, key_name))
                            
                    input = raw_input('Please enter the number of the label you wish to remove')
                    if input.isdigit():
                        input = '&' + input
                        
                        # Checks if input is present in object's UserText
                        # If no, continues onto other objects
                        if input in objText:
                            deletion('label', id, input)
                        else:
                            raise ValueError('That input number is not present in the Rhino object')
                    else:
                        raise TypeError('Invalid input; please enter an integer to select which predicate/directive to remove from the Rhino object')
                else:
                    warnings.warn('No labels are attached to this Rhino object')
            else:
                raise TypeError('Label removal cannot be performed on this object')
                
        return True
    
    # Removes label description tag of chosen label description from user text of Rhino Object
    # Redraws text dot accompanying Rhino Object to reflect the change
    @staticmethod
    def delete_description():
        objectIDs = rs.GetObjects('Select objects to remove label descriptions from')
        print ('You have selected ' + str(len(objectIDs)) + ' objects to remove label descriptions from') 
        go_ahead = False
        
        if len(objectIDs) == 0:
            warnings.warn('No objects were selected.')
            return False
            
        # Goes through list of selected objects
        for id in objectIDs:
            if rs.ObjectType(id) == 1 or rs.ObjectType(id) == 4 or rs.ObjectType(id) == 8192 or rs.ObjectType(id) == 8 or rs.ObjectType(id) == 16 or rs.ObjectType(id) == 32:
                objText = rs.GetUserText(id)
                for key_name in objText:
                    if '#' in key_name:
                        go_ahead = True
                
                if go_ahead:
                    print ('The following descriptions are present in this object:')
                    
                    for key_name in objText:
                        if '#' in key_name:
                            print (key_name[1:] + ': ' + rs.GetUserText(id, key_name))
                            
                    input = raw_input('Please enter the number of the descriptions you wish to remove')
                    if input.isdigit():
                        input = '#' + input
    
                        # Checks if input is present in object's UserText
                        # If no, continues onto other objects
                        if input in objText:
                            deletion('desc', id, input)
                        else:
                            raise ValueError('That input number is not present in the Rhino object')
                    else:
                        raise TypeError('Invalid input; please enter an integer to select which predicate/directive to remove from the Rhino object')
                else:
                    print ('No label descriptions are attached to this Rhino object')
            else:
                raise TypeError('Label description removal cannot be performed on this object')
                
        return True
    
    # Removes predicate/directive tag of chosen predicate from user text of Rhino Object
    # Redraws text dot accompanying Rhino Object to reflect the change
    @staticmethod
    def delete_pred_dir():    
        objectIDs = rs.GetObjects('Select objects to remove predicates/directives from')
        if len(objectIDs) > 0:
            print ('You have selected ' + str(len(objectIDs)) + ' object/s to remove predicates/directives from') 
        else:
            warnings.warn('No objects were selected.')
            return False

        # Goes through list of selected objects
        for obj in objectIDs:
            if rs.ObjectType(obj) == 1 or rs.ObjectType(obj) == 4 or rs.ObjectType(obj) == 8192 or rs.ObjectType(obj) == 8 or rs.ObjectType(obj) == 16 or rs.ObjectType(objs) == 32:
                
                if 'tagText' in rs.GetUserText(obj):
                    objText = rs.GetUserText(obj)
    
                    print ('The following predicates/directives are present in this object:')
                    tagValue = rs.GetUserText(obj, 'tagText')
                    tagValue = tagValue.split('|')
                    
                    for count, tag in enumerate(tagValue):
                        print (str(count) + ': ' + tag)
                            
                    input = raw_input('Please enter the number of the label you wish to remove')
                    
                    # Checks if input is present in object's UserText
                    # If no, continues onto other objects
    
                    if input.isdigit():
                        if int(input) < len(objText):
                            deletion('pd', obj, int(input))
                        else:
                            raise ValueError('That input number is not present in the Rhino object')
                    else:
                        raise TypeError('Invalid input; please enter an integer to select which predicate/directive to remove from the Rhino object')
                else:
                    print ('There are no predicates/directives to remove from this object')
            else:
                raise TypeError('Label description removal cannot be performed on this object')        
        return True
        
    # Removes predicate/directive name tag from geometry
    @staticmethod
    def delete_tag():
        objectIDs = rs.GetObjects('Select objects to remove tags from')
        print ('You have selected ' + str(len(objectIDs)) + ' object/s to remove tags from') 
        
        if len(objectIDs) > 0:
            # Goes through list of selected objects
            for obj in objectIDs:
                if 'tag' in rs.GetUserText(obj):
                    tagValue = rs.GetUserText(obj, 'tag')
                    print ('The following tag will be removed from this object: ' + tagValue)
                    
                    # Automatically removes tag as there is only one available
                    deletion('tag', obj)
                else:
                    warnings.warn('Tag removal cannot be performed on this object')
                    
            return True
        else:
            warnings.warn('No objects were selected')
            return False
    
    # Need to update this
    # Keep track of planesegment changes esp wrt Normal tagging
    # Adds or changes the tag present in a Rhino geometry
    # Stores tag preceeded by its geometry type as part of user text in Rhino geometry, i.e. 'line(<tag text>)'
    @staticmethod
    def tag(tagSelf = True):
        ids = rs.GetObjects('Select geometry/ies to change tag information') 
        
        for id in ids:   
            if 'visual' in rs.GetUserText(id):
                continue
            
            orig_color_geom = rs.coercecolor(rs.ObjectColor(id))
            rs.ObjectColor(id, [255,225,0])
            oldTag = None
            
            if 'tag' not in rs.GetUserText(id): 
                if tagSelf:
                    origColor = rs.coercecolor(rs.ObjectColor(id))
                    rs.ObjectColor(id, [255, 225, 0])
                    tagName = raw_input('Please enter tag name for the highlighted geometry. If no input is entered, then the tag will be deleted')
                    rs.SetUserText(id, 'tag', tagName)
                    rs.ObjectColor(id, origColor)
                    
                if tagSelf == False or tagName == '':
                    rs.SetUserText(id, 'tag', str(sc.sticky['keyCount']))
                    sc.sticky['keyCount'] += 1
            else:
                message = ('Object currently has the tag ' + rs.GetUserText(id, 'tag') + '. Enter new tag')
                oldTagName = rs.GetUserText(id, 'tag')
                tagName = raw_input(message)
                
                if tagSelf == False or tagName == '':
                    tagName = str(sc.sticky['keyCount'])
                    sc.sticky['keyCount'] += 1
                    
                rs.SetUserText(id, 'tag', tagName)
                print ('Tag changed to ' + tagName)
            
            if 'tagText' in rs.GetUserText(id):
                # Retrieves entirety of text displayed in visual text dot
                oldText = rs.GetUserText(id, 'tagText').split('|')
                
                # Searches for text bits that have 'normal' in them
                for index, oldBit in enumerate(oldText):
                    if 'normal' in oldBit:
                        oldBits = oldBit.split('_')
                        groupName = oldBits[0].split('^')[-1]
                        
                        # Checks if normal statement is a 3D normal statement
                        if 'planeSegment-' in rs.GetUserText(id, 'norm_end_' + groupName):
                            rs.SetUserText(id, 'norm_end_' + groupName, 'planeSegment-' + rs.GetUserText(id, 'tag'))
                            # Obtains old tag name; this will be used to search for things to replace with the new tag name within the tag text
                            oldTag = 'planeSegment-' + oldTagName
                            newText = ''
                            
                            # Searches for plane segment tags to replace
                            for count, bit in enumerate(oldBits):
                                if bit == oldTag:
                                    oldBits[count] = 'planeSegment-' +  rs.GetUserText(id, 'tag')
                            
                            for bit in oldBits:
                                if newText == '':
                                    newText = bit
                                else:
                                    newText+= '_' + bit
    
                            oldText[index] = newText
                
                # Constructs new tag text
                tagText = '' 
                for bit in oldText:
                    if tagText == '':
                        tagText = bit
                    else:
                        tagText+= '_' + bit
        
                rs.SetUserText(id, 'tagText', tagText)
                        
            lText0 = ''
            ldText0 = ''
            ptText0 = ''
            fullText = ''
            
            if 'LabelDot' in rs.GetUserText(id):
                try:    
                    rs.DeleteObject(rs.GetUserText(id, "LabelDot"))
                    textDot = rs.GetUserText(id, "LabelDot")
                except:
                    textDot = None
            
            # Collects pre-existing predicate information
            if 'lText' in rs.GetUserText(id) or 'ldText' in rs.GetUserText(id) or 'tagText' in rs.GetUserText(id):
                if 'lText' in rs.GetUserText(id):
                    lText0 = rs.GetUserText(id, "lText")
                    
                if 'ldText' in rs.GetUserText(id):
                    ldText0 = rs.GetUserText(id, "ldText")
                    
                if 'tagText' in rs.GetUserText(id):
                    ptText0 = rs.GetUserText(id, "tagText")
            
            # Creating new visual text dot text
            if lText0 != '':
                fullText = 'l(' + lText0 + ')'
            
            if ldText0 != '':
                if fullText != '':
                    fullText = fullText + '|d(' + ldText0 +')'
                else:
                    fullText = 'd(' + ldText0 +')'
                    
            if ptText0 != '':
                if fullText != '':
                    fullText = fullText + '|#(' + ptText0 + ')'
                else:
                    fullText = '#(' + ptText0 + ')'
            
            if rs.IsCurve(id) or rs.IsSurface(id) or rs.IsMesh(id) or rs.IsBrep(id):
                if rs.IsCurve(id):
                    if fullText != '':
                        fullText = 'line(' + rs.GetUserText(id, 'tag') + ')|' + fullText
                    else:
                        fullText = 'line(' + rs.GetUserText(id, 'tag') + ')'
                    textDot = rs.AddTextDot(fullText, rs.CurveMidPoint(id))
                else:
                    if fullText != '':
                        fullText = 'plane(' + rs.GetUserText(id, 'tag') + ')|' + fullText
                    else:
                        fullText = 'plane(' + rs.GetUserText(id, 'tag') + ')'
                    obj = rhdoc.Objects.Find(id).BrepGeometry
                    for curve in obj.Loops:
                        if curve.LoopType == r.Geometry.BrepLoopType.Outer:
                            mainCurve = curve.To3dCurve().DuplicateSegments()[0]
                            startpt = mainCurve.PointAtStart
                            endpt = mainCurve.PointAtEnd
                            midpt = ((startpt.X + endpt.X)/2,(startpt.Y + endpt.Y)/2,(startpt.Z + endpt.Z)/2)
                            midpt =  r.Geometry.Point3d(midpt[0], midpt[1], midpt[2])
                            textDot = rs.AddTextDot(fullText, midpt)
                            
                            break
                
                rs.SetUserText(id, "LabelDot", textDot)
                rs.SetUserText(textDot, 'visual', True)
                rs.ObjectColor(textDot, (0,0,0))
                rs.ObjectColor(id, orig_color_geom)
            else:
                if fullText == '':
                    fullText = 'point(' + rs.GetUserText(id, 'tag') + ')'
                else:
                    fullText = 'point(' + rs.GetUserText(id, 'tag') + ')|' + fullText
            
                if rs.IsPoint(id):
                    coords = rs.PointCoordinates(id)
                    textDot = rs.AddTextDot(fullText, coords)
                    
                    for item in rs.GetUserText(id):
                        rs.SetUserText(textDot, rs.GetUserText(id, item))
                        
                    rs.DeleteObject(id)
                    rs.SetUserText(textDot, 'tag', tagName)
                    rs.ObjectColor(textDot, orig_color_geom)
                elif rs.IsTextDot(id): 
                    id = rs.TextDotText(id, fullText)
                    rs.SetUserText(id, 'tag', tagName)
                    rs.ObjectColor(id, orig_color_geom)
                        
        return True
    
    """
    =============================
    GEOMETRIC CORRECTION ANALYSIS
    =============================
    """

    # Analyses a set of Rhino geometry for significant (precision value < x < 1% of the length of the line or Rhino's unit setting)
    # Modifies geometry to fit more accurately with each other (point on line, line segment intersection with other line segments)
    @staticmethod
    def analysis (objects):
        if len(objects) > 1:
            # Obtains power (in tens) of precision in sortal library based on decimal places for use in analysis function
            precision = 10**(-(coordinate.precision()))
            
            for count, change_object in enumerate(objects):
                start = None
                end = None
               
                # Get object to change
                # Line segments/polylines
                if rs.ObjectType(change_object) == 4:
                    start = rs.CurveStartPoint(change_object)
                    end = rs.CurveEndPoint(change_object)
                    compare_length = rs.CurveLength(change_object) * 0.01
                    start_param = rs.CurveClosestPoint(change_object,start)
                    end_param = rs.CurveClosestPoint(change_object,end)
                
                # Points
                elif rs.ObjectType(change_object) == 1 or (rs.ObjectType(change_object) == 8192 and 'visual' not in rs.GetUserText(change_object)):
                    start = rs.PointCoordinates(change_object)
                    compare_length = 10**(-rs.UnitSystem()-1)
                
                # Goes through list of remaining objects after change_object's position in the list
                for compare_object in objects[count:]:
                    #print (rs.ObjectType(change_object), rs.ObjectType(compare_object), start)
                    
                    # Handles line segments
                    if rs.ObjectType(compare_object) == 4:
                        start_compare = rs.CurveStartPoint(compare_object)
                        end_compare = rs.CurveEndPoint(compare_object)
                        dist1 = (rs.EvaluateCurve(compare_object, rs.CurveClosestPoint(compare_object,start)) - start)
                        start_dist = ptDistance(dist1.X, dist1.Y, dist1.Z)
    
                        # Both points and line segments will give start a value, so no checking condition is needed
                        if abs(start_dist) > precision and abs(start_dist) < compare_length:
                            new_start = start + dist1
                            if rs.ObjectType(change_object) == 4:
                                rs.CurveStartPoint(change_object, start_param, new_start)
                            elif rs.ObjectType(change_object) == 1 or (rs.ObjectType(change_object) == 8192 and 'visual' not in rs.GetUserText(change_object)):
                                rs.MoveObject(change_object, dist1)
                        
                        # change_object is a line segment
                        if end != None:
                            dist2 = (rs.EvaluateCurve(compare_object, rs.CurveClosestPoint(compare_object,end)) - end)
                            end_dist = ptDistance(dist2.X, dist2.Y, dist2.Z)
                            
                            if abs(end_dist) > precision and abs(end_dist) < compare_length:
                                new_end = end + dist2
                                v1 = r.Geometry.Point3d(start[0], start[1], start[2])
                                v2 = r.Geometry.Point3d(new_end[0], new_end[1], new_end[2])
                                new_obj = rs.AddLine(v1, v2)
            
                                for text in rs.GetUserText(change_object):
                                    rs.SetUserText(new_obj, text, rs.GetUserText(change_object, text))
                                
                                rs.DeleteObject(objects.pop(count))
                                objects.insert(count, new_obj)
                                change_object = new_obj
                    
                    # Handles points
                    # Points are moved to more accurately lie on lines
                    elif rs.ObjectType(compare_object) == 1 or (rs.ObjectType(compare_object) == 8192 and 'visual' not in rs.GetUserText(compare_object)):
                        start_compare = rs.PointCoordinates(compare_object)
        
                        # Handles line segments and polylines
                        # Moves the point to lie on the line segment
                        if rs.ObjectType(change_object) == 4:
                            #print (rs.EvaluateCurve(change_object, rs.CurveClosestPoint(change_object,start)))
                            dist1 = (rs.EvaluateCurve(change_object, rs.CurveClosestPoint(change_object,start)) - start)
                            
                        # Handles points
                        # Moves the point to intersect with the change_object, also a point
                        elif rs.ObjectType(change_object) == 1:
                            dist1 = start_compare - start
        
                        start_dist = ptDistance(dist1.X, dist1.Y, dist1.Z)    
                        if abs(start_dist) > precision and abs(start_dist) < compare_length:
                            rs.MoveObject(compare_object, dist1)
        
            return (objects)
        else:
            warnings.warn ('More than one object must be selected for this function.')
            return False